interface ReadonlyDict<T = any> {
    readonly [key: string]: T;
}
interface Dict<T = any> {
    [key: string]: T;
}

/**
 * A result returned from a shape after it was applied to an input value.
 *
 * - If `null` then the input value is valid and wasn't transformed.
 * - If an {@link Ok} instance then the input value is valid and was transformed.
 * - If an array of {@link Issue issues} then the input value is invalid.
 *
 * @template Value The output value.
 * @see {@link Shape._apply}
 * @see {@link Shape._applyAsync}
 * @group Other
 */
type Result<Value = any> = Ok<Value> | Issue[] | null;
/**
 * Carries the result of a successful input parsing.
 *
 * @template Value The output value.
 * @group Other
 */
interface Ok<Value = any> {
    ok: true;
    /**
     * The output value.
     */
    value: Value;
}
/**
 * Carries the result of a failed input parsing.
 *
 * @group Other
 */
interface Err {
    ok: false;
    /**
     * The array of issues encountered during parsing.
     */
    issues: Issue[];
}
/**
 * A validation issue raised during input parsing.
 *
 * @group Issues
 */
interface Issue {
    /**
     * A code of the validation issue, which uniquely identifies the reason of an issue.
     *
     * For all built-in shape operations, a code can be used to infer the type of the {@link Issue.param}.
     *
     * @see [Validation errors](https://github.com/smikhalevski/doubter#validation-errors)
     */
    code?: any;
    /**
     * An object path to {@link Issue.input} value starting from the parsed object. Both an empty array and `undefined`
     * mean that an issue is caused by the {@link Issue.input} value itself.
     *
     * @example ['users', 0, 'age']
     * @see {@link Shape.at}
     */
    path?: any[];
    /**
     * The input value that caused a validation issue. Note that if coercion or alteration were applied then this contains
     * a transformed value.
     *
     * @see [Type coercion](https://github.com/smikhalevski/doubter#type-coercion)
     */
    input?: any;
    /**
     * The human-readable issue message.
     *
     * Messages produced by built-in operations are strings but custom messages can have an arbitrary type.
     *
     * @see [Localization](https://github.com/smikhalevski/doubter#localization)
     */
    message?: any;
    /**
     * The parameter value associated with the issue.
     *
     * @see [Validation errors](https://github.com/smikhalevski/doubter#validation-errors)
     */
    param?: any;
    /**
     * The optional metadata associated with the issue.
     *
     * @see [Metadata](https://github.com/smikhalevski/doubter#metadata)
     */
    meta?: any;
}
/**
 * Options used by shapes and built-in operations to create an issue.
 *
 * @see {@link Issue}
 * @group Issues
 */
interface IssueOptions {
    /**
     * The custom issue message.
     *
     * @see {@link Message}
     * @see {@link core!Issue#message Issue.message}
     */
    message?: Message | any;
    /**
     * An arbitrary metadata that is added to an issue.
     *
     * @see {@link core!Issue#meta Issue.meta}
     */
    meta?: any;
}
/**
 * A callback that returns an issue message, or an issue message string.
 *
 * `%s` placeholder in string messages is replaced with the {@link Issue.param}.
 *
 * @group Issues
 */
type Message = MessageCallback | string;
/**
 * Returns a human-readable message that describes an issue.
 *
 * You can assign the {@link Issue.message} property directly inside this callback or return the message.
 *
 * @param issue The issue for which the message should be produced.
 * @param options The parsing options.
 * @returns The value that should be used as an issue message. The returned value is ignored if `issue.message` was
 * assigned a non-`undefined` value inside the callback.
 * @group Issues
 */
type MessageCallback = (issue: Issue, options: ApplyOptions) => any;
/**
 * An operation that a shape applies to its output.
 *
 * @template InputValue The input value to which the shape was applied.
 * @template OutputValue The shape output value to which the operation must be applied.
 * @see {@link Shape.check}
 * @see {@link Shape.alter}
 * @see {@link Shape.refine}
 * @see {@link Shape.use}
 * @group Operations
 */
interface Operation<InputValue = any, OutputValue = any> {
    /**
     * The type of the operation such as {@link StringShape#regex "string.regex"} or
     * {@link ArrayShape#includes "array.includes"}.
     */
    readonly type: any;
    /**
     * The additional param associated with the operation.
     *
     * This param usually contains a {@link Operation.type}-specific data is used in the
     * {@link OperationCallback callback} returned by the {@link Operation.factory} method.
     *
     * Built-in operations use the same param for an operation and an issue that is raised if an operation fails.
     *
     * @see {@link Issue.param}
     */
    readonly param: any;
    /**
     * Creates an {@link OperationCallback} that applies the logic of the operation to the shape output and passes the
     * control to the next operation.
     *
     * @param next The callback that applies the next operation.
     * @returns The callback that applies an operation to the shape output.
     */
    readonly factory: (next: OperationCallback) => OperationCallback<InputValue, OutputValue>;
}
/**
 * A callback that applies an operation to the shape output.
 *
 * @param input The input value to which the shape was applied.
 * @param output The shape output value to which the operation must be applied.
 * @param options Parsing options.
 * @param issues The mutable array of issues captured by a shape, or `null` if there were no issues raised yet.
 * @returns The result of the operation.
 * @template InputValue The input value to which the shape was applied.
 * @template OutputValue The shape output value to which the operation must be applied.
 * @see {@link Operation.factory}
 * @group Operations
 */
type OperationCallback<InputValue = any, OutputValue = any> = (input: InputValue, output: OutputValue, options: ApplyOptions, issues: Issue[] | null) => Result;
/**
 * Options of a generic {@link Operation operation}.
 *
 * @group Operations
 */
interface OperationOptions {
    /**
     * The type of the operation.
     *
     * @see {@link Operation.type}
     */
    type?: any;
    /**
     * The additional param associated with the operation.
     *
     * @see {@link Operation.param}
     */
    param?: any;
}
/**
 * Options of a custom {@link Operation operation}.
 *
 * @see {@link Shape.check}
 * @see {@link Shape.alter}
 * @see {@link Shape.refine}
 * @group Operations
 */
interface CustomOperationOptions extends OperationOptions {
    /**
     * If `true` then the operation is applied even if some of the preceding operations have failed.
     *
     * @default false
     */
    force?: boolean;
}
/**
 * @inheritDoc
 * @template Param The param that is passed to a callback when an operation is applied.
 * @group Operations
 */
interface ParameterizedCustomOperationOptions<Param> extends CustomOperationOptions {
    param: Param;
}
/**
 * Checks that a value satisfies a requirement and returns issues if it doesn't.
 *
 * If a {@link ValidationError} is thrown, its issues are captured and incorporated into a parsing result.
 *
 * @param value The value to check.
 * @param param The additional param that was associated with the operation.
 * @param options Parsing options.
 * @returns `null` or `undefined` if the value satisfies a requirement; an issue or an array of issues if a value
 * doesn't satisfy a requirement.
 * @template Value The value to check.
 * @template Param The additional param that was associated with the check operation.
 * @see {@link Shape.check}
 * @group Operations
 */
type CheckCallback<Value = any, Param = any> = (value: Value, param: Param, options: ApplyOptions) => Issue[] | Issue | null | undefined | void;
/**
 * Checks that a value matches a predicate.
 *
 * If a {@link ValidationError} is thrown, its issues are captured and incorporated into a parsing result. Throw if
 * refinement cannot be performed, and you want to abort the operation.
 *
 * @param value The value to refine.
 * @param param The additional param that was associated with the operation.
 * @param options Parsing options.
 * @return Truthy if value matches the predicate, or falsy if it doesn't.
 * @template Value The value to refine.
 * @template Param The additional param that was associated with the operation.
 * @see {@link Shape.refine}
 * @group Operations
 */
type RefineCallback<Value = any, Param = any> = (value: Value, param: Param, options: ApplyOptions) => any;
/**
 * A [narrowing predicate](https://www.typescriptlang.org/docs/handbook/2/narrowing.html) that refines the value type.
 *
 * If a {@link ValidationError} is thrown, its issues are captured and incorporated into a parsing result. Throw if
 * refinement cannot be performed, and you want to abort the operation.
 *
 * @param value The value to refine.
 * @param param The additional param that was associated with the operation.
 * @param options Parsing options.
 * @return `true` if value matches the predicate, or `false` otherwise.
 * @template Value The value to refine.
 * @template RefinedValue The refined value.
 * @template Param The additional param that was associated with the operation.
 * @see {@link Shape.refine}
 * @group Operations
 */
type RefinePredicate<Value = any, RefinedValue extends Value = Value, Param = any> = (value: Value, param: Param, options: ApplyOptions) => value is RefinedValue;
/**
 * Options of {@link Shape.refine a refinement operation}.
 *
 * @group Operations
 */
interface RefineOptions extends CustomOperationOptions, IssueOptions {
    /**
     * The code of an issue that would be raised if the refinement fails.
     *
     * @default "any.refine"
     * @see {@link Issue.code}
     */
    code?: any;
}
/**
 * @inheritDoc
 * @template Param The param that is passed to a callback when an operation is applied.
 * @group Operations
 */
interface ParameterizedRefineOptions<Param> extends RefineOptions {
    param: Param;
}
/**
 * Alters the value without changing its base type.
 *
 * If you want to change the base type, consider using {@link Shape.convert}.
 *
 * If a {@link ValidationError} is thrown, its issues are captured and incorporated into a parsing result. Throw if
 * alteration cannot be performed, and you want to abort the operation.
 *
 * @param value The value to alter.
 * @param param The additional param that was associated with the operation.
 * @param options Parsing options.
 * @returns The altered value.
 * @template Value The value to alter.
 * @template Param The additional param that was associated with the operation.
 * @see {@link Shape.alter}
 * @see {@link Shape.convert}
 * @group Operations
 */
type AlterCallback<Value = any, Param = any> = (value: Value, param: Param, options: ApplyOptions) => Value;
/**
 * Options used when a shape is applied to an input value.
 *
 * @see {@link Shape._apply}
 * @see {@link Shape._applyAsync}
 * @group Other
 */
interface ApplyOptions {
    /**
     * If `true` then parsing is aborted after the first issue is encountered.
     *
     * @default false
     */
    readonly earlyReturn?: boolean;
    /**
     * If `true` then shapes that support input value type coercion, would try to coerce an input to a required type.
     *
     * @default false
     */
    readonly coerce?: boolean;
    /**
     * The custom context.
     */
    readonly context?: any;
}
/**
 * Returns a human-readable message that describes issues that were raised during input parsing.
 *
 * @param issues The array of issues that were raised.
 * @param input The input value that was parsed.
 * @see {@link ParseOptions.errorMessage}
 * @group Other
 */
type ErrorMessageCallback = (issues: Issue[], input: any) => string;
/**
 * Options used during parsing.
 *
 * @group Other
 */
interface ParseOptions extends ApplyOptions {
    /**
     * A message that is passed to {@link ValidationError} if issues are raised during parsing.
     */
    readonly errorMessage?: ErrorMessageCallback | string;
}
/**
 * A literal value of any type.
 *
 * @group Other
 */
type Any = object | string | number | bigint | boolean | symbol | null | undefined;

/**
 * The marker object that is used to denote an impossible value.
 *
 * For example, {@link NEVER} is returned from {@link CoercibleShape._coerce} method when coercion is not possible.
 *
 * @group Other
 */
declare const NEVER: never;
/**
 * Extracts the shape input type.
 *
 * @template S The shape from which the input type must be inferred.
 * @group Type Inference
 */
type Input<S extends AnyShape> = S[INPUT];
/**
 * Extracts the shape output type.
 *
 * @template S The shape from which the output type must be inferred.
 * @group Type Inference
 */
type Output<S extends AnyShape> = S[OUTPUT];
/**
 * Excludes `U` from `T` only if `U` is a literal type.
 */
type ExcludeLiteral<T, U> = number extends U ? T : string extends U ? T : symbol extends U ? T : bigint extends U ? T : object extends U ? T : boolean extends U ? T : Exclude<T, U>;
/**
 * The ephemeral unique symbol that is used for type branding by {@link Branded}.
 */
declare const BRAND: unique symbol;
/**
 * The branded type.
 *
 * @template Value The value to brand.
 * @template Brand The brand value.
 * @group Other
 */
type Branded<Value, Brand> = Value & {
    [BRAND]: Brand;
};
/**
 * A shape should implement {@link DeepPartialProtocol} to support conversion to a deep partial alternative.
 *
 * @template S The deep partial alternative of the shape.
 * @group Other
 */
interface DeepPartialProtocol<S extends AnyShape> {
    /**
     * Converts the shape and its child shapes to deep partial alternatives.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * @returns The deep partial clone of the shape.
     */
    deepPartial(): S;
}
/**
 * An arbitrary shape.
 *
 * @group Shapes
 */
type AnyShape = Shape | Shape<never>;
/**
 * Returns the shape that refines the shape output.
 *
 * @template S The base shape.
 * @template RefinedValue The refined output value.
 * @group Shapes
 */
type RefineShape<S extends AnyShape, RefinedValue> = Shape<Input<S>, RefinedValue> & Omit<S, keyof Shape>;
/**
 * Returns the deep partial alternative of the shape if it implements {@link DeepPartialProtocol}, or returns shape as
 * is if it doesn't.
 *
 * @template S The shape to convert to a deep partial alternative.
 * @group Shapes
 */
type DeepPartialShape<S extends AnyShape> = S extends DeepPartialProtocol<infer T> ? T : S;
/**
 * Shape that is both optional and deep partial.
 *
 * @template S The shape to convert to an optional deep partial alternative.
 * @group Shapes
 */
type OptionalDeepPartialShape<S extends AnyShape> = AllowShape<DeepPartialShape<S>, undefined>;
/**
 * Shortcut for {@link ReplaceShape} that allows the same value as both an input and an output.
 *
 * @template S The shape that parses the input without the replaced value.
 * @template AllowedValue The value that is allowed as an input and output.
 * @group Shapes
 */
type AllowShape<S extends AnyShape, AllowedValue> = ReplaceShape<S, AllowedValue, AllowedValue>;
/**
 * Shortcut for {@link ExcludeShape} that doesn't impose the exclusion on the type level.
 *
 * @template BaseShape The base shape.
 * @template ExcludedShape The shape to which the output must not conform.
 * @group Shapes
 */
interface NotShape<BaseShape extends AnyShape, ExcludedShape extends AnyShape> extends Shape<Input<BaseShape>, Output<BaseShape>>, DeepPartialProtocol<NotShape<DeepPartialShape<BaseShape>, ExcludedShape>> {
    /**
     * The base shape.
     */
    readonly baseShape: BaseShape;
    /**
     * The shape to which the output must not conform.
     */
    readonly excludedShape: ExcludedShape;
}
/**
 * The baseline shape implementation.
 *
 * @template InputValue The input value.
 * @template OutputValue The output value.
 * @group Shapes
 */
declare class Shape<InputValue = any, OutputValue = InputValue> {
    /**
     * The dictionary of shape annotations.
     *
     * @see {@link Shape.annotate}
     */
    annotations: Dict;
    /**
     * The array of operations that are applied to the shape output.
     *
     * @see {@link Shape.use}
     * @see [Operations](https://github.com/smikhalevski/doubter#operations)
     */
    operations: readonly Operation[];
    /**
     * The callback that applies {@link Shape.operations} to the shape output value.
     */
    protected _applyOperations: OperationCallback;
    /**
     * Returns a sub-shape that describes a value associated with the given property name, or `null` if there's no such
     * sub-shape.
     *
     * @param key The key for which the sub-shape must be retrieved.
     * @returns The sub-shape or `null` if there's no such key in the shape.
     */
    at(key: unknown): AnyShape | null;
    /**
     * Returns `true` if the shape accepts given input type or value, or `false` otherwise.
     *
     * @param input The type or value that must be checked.
     */
    accepts(input: unknown): boolean;
    /**
     * Assigns annotations to the shape.
     *
     * @param annotations Annotations to add.
     * @returns The clone of the shape with the updated annotations.
     * @see {@link Shape.annotations}
     */
    annotate(annotations: ReadonlyDict): this;
    /**
     * Appends an operation to the shape.
     *
     * @param cb The factory that produces the operation callback.
     * @param options The operation operations.
     * @returns The clone of the shape.
     * @see [Operations](https://github.com/smikhalevski/doubter#operations)
     */
    use(
    /**
     * Creates an {@link OperationCallback} that applies the logic of the operation to the shape output and passes the
     * control to the next operation.
     *
     * @param next The callback that applies the next operation.
     * @returns The callback that applies an operation to the shape output.
     */
    cb: (next: OperationCallback) => OperationCallback<InputValue, OutputValue>, options?: OperationOptions): this;
    /**
     * Adds the check {@link use operation} that is applied to the shape output.
     *
     * If check callback returns an empty array, it is considered that no issues have occurred.
     *
     * @param cb The callback that checks the shape output.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @template Param The param that is passed to the {@link CheckCallback} when a check operation is applied.
     * @see {@link Shape.refine}
     */
    check<Param>(cb: CheckCallback<OutputValue, Param>, options: ParameterizedCustomOperationOptions<Param>): this;
    /**
     * Adds the check {@link use operation} that is applied to the shape output.
     *
     * If check callback returns an empty array, it is considered that no issues have occurred.
     *
     * @param cb The callback that checks the shape output.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @see {@link Shape.refine}
     */
    check(cb: CheckCallback<OutputValue>, options?: CustomOperationOptions): this;
    /**
     * Adds an {@link use operation} that refines the shape output type with the
     * [narrowing predicate](https://www.typescriptlang.org/docs/handbook/2/narrowing.html).
     *
     * @param cb The predicate that returns `true` if value conforms the required type, or `false` otherwise.
     * @param options The operation options or the issue message.
     * @returns The shape with the narrowed output.
     * @template RefinedValue The narrowed output value.
     * @template Param The param that is passed to the {@link RefinePredicate} when a refinement operation is applied.
     * @see {@link Shape.check}
     */
    refine<RefinedValue extends OutputValue, Param>(cb: RefinePredicate<OutputValue, RefinedValue, Param>, options: ParameterizedRefineOptions<Param> | Message): RefineShape<this, RefinedValue>;
    /**
     * Adds an {@link use operation} that refines the shape output type with the
     * [narrowing predicate](https://www.typescriptlang.org/docs/handbook/2/narrowing.html).
     *
     * @param cb The predicate that returns `true` if value conforms the required type, or `false` otherwise.
     * @param options The operation options or the issue message.
     * @returns The shape with the narrowed output.
     * @template RefinedValue The narrowed output value.
     * @see {@link Shape.check}
     */
    refine<RefinedValue extends OutputValue>(cb: RefinePredicate<OutputValue, RefinedValue>, options?: RefineOptions | Message): RefineShape<this, RefinedValue>;
    /**
     * Adds an {@link use operation} that checks that the output value conforms the predicate.
     *
     * @param cb The predicate that returns truthy result if value is valid, or returns falsy result otherwise.
     * @param options The operation options or the issue message.
     * @returns The clone of the shape.
     * @template Param The param that is passed to the {@link RefineCallback} when a refinement operation is applied.
     * @see {@link Shape.check}
     */
    refine<Param>(cb: RefineCallback<OutputValue, Param>, options?: ParameterizedRefineOptions<Param> | Message): this;
    /**
     * Adds an {@link use operation} that checks that the output value conforms the predicate.
     *
     * @param cb The predicate that returns truthy result if value is valid, or returns falsy result otherwise.
     * @param options The operation options or the issue message.
     * @returns The clone of the shape.
     * @see {@link Shape.check}
     */
    refine(cb: RefineCallback<OutputValue>, options?: RefineOptions | Message): this;
    /**
     * Adds an {@link use operation} that alters the output value without changing its type.
     *
     * @param cb The callback that alters the shape output.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @template Param The param that is passed to the {@link AlterCallback} when an alteration operation is applied.
     * @see {@link Shape.convert}
     */
    alter<Param>(cb: AlterCallback<OutputValue, Param>, options: ParameterizedCustomOperationOptions<Param>): this;
    /**
     * Adds an {@link use operation} that alters the output value without changing its type.
     *
     * @param cb The callback that alters the shape output.
     * @param options The operation options.
     * @returns The clone of the shape.
     * @see {@link Shape.convert}
     */
    alter(cb: AlterCallback<OutputValue>, options?: CustomOperationOptions): this;
    /**
     * Synchronously converts the output value of the shape.
     *
     * @param cb The callback that converts the input value. Throw a {@link ValidationError} to notify that the conversion
     * cannot be successfully completed.
     * @returns The {@link ConvertShape} instance.
     * @template ConvertedValue The value returned from the callback that converts the output value of this shape.
     * @see {@link Shape.alter}
     */
    convert<ConvertedValue>(
    /**
     * @param value The shape output value.
     * @param options Parsing options.
     */
    cb: (value: OutputValue, options: ApplyOptions) => ConvertedValue): Shape<InputValue, ConvertedValue>;
    /**
     * Asynchronously converts the output value of the shape.
     *
     * @param cb The callback that converts the input value asynchronously. The returned promise can be rejected with a
     * {@link ValidationError} to notify that the conversion cannot be successfully completed.
     * @returns The {@link ConvertShape} instance.
     * @template ConvertedValue The value returned from the callback that converts the output value of this shape.
     * @see {@link Shape.alter}
     */
    convertAsync<ConvertedValue>(
    /**
     * @param value The shape output value.
     * @param options Parsing options.
     */
    cb: (value: OutputValue, options: ApplyOptions) => PromiseLike<ConvertedValue>): Shape<InputValue, ConvertedValue>;
    /**
     * Pipes the output of this shape to the input of another shape.
     *
     * @param shape The shape that validates the output if this shape.
     * @template OutputShape The output value.
     */
    to<OutputShape extends AnyShape>(shape: OutputShape): PipeShape<this, OutputShape>;
    /**
     * Returns a shape that adds a brand to the output type.
     *
     * @returns A shape with the branded output type.
     * @template Brand The brand value.
     */
    brand<Brand extends string | number | symbol>(): RefineShape<this, Branded<OutputValue, Brand>>;
    /**
     * Replaces an input value with an output value.
     *
     * @param inputValue The input value to replace.
     * @param outputValue The output value that is returned if an `inputValue` is received.
     * @template InputValue The input value to replace.
     * @template OutputValue The output value that is used as the replacement for an input value.
     */
    replace<InputValue extends Any, OutputValue extends Any>(inputValue: InputValue, outputValue: OutputValue): ReplaceShape<this, InputValue, OutputValue>;
    /**
     * Allows an input value, so it is passed directly to the output.
     *
     * @param value The allowed value.
     * @template AllowedValue The allowed value.
     */
    allow<AllowedValue extends Any>(value: AllowedValue): AllowShape<this, AllowedValue>;
    /**
     * Excludes value from both input and output.
     *
     * @param value The excluded value.
     * @param options The issue options or the issue message.
     * @template DeniedValue The denied value.
     */
    deny<DeniedValue extends InputValue | OutputValue>(value: DeniedValue, options?: IssueOptions | Message): DenyShape<this, DeniedValue>;
    /**
     * Replaces `undefined` input value with an `undefined` output value.
     */
    optional(): AllowShape<this, undefined>;
    /**
     * Replaces `undefined` input value with a default output value.
     *
     * @param defaultValue The value that should be used if an input value is `undefined`.
     * @template DefaultValue The value that is used as the replacement for `undefined`.
     */
    optional<DefaultValue extends Any>(defaultValue: DefaultValue): ReplaceShape<this, undefined, DefaultValue>;
    /**
     * Replaces `null` input value with an `null` output value.
     */
    nullable(): AllowShape<this, null>;
    /**
     * Replaces `null` input value with a default output value.
     *
     * @param defaultValue The value that should be used if an input value is `null`.
     * @template DefaultValue The value that is used as the replacement for `null`.
     */
    nullable<DefaultValue extends Any>(defaultValue: DefaultValue): ReplaceShape<this, null, DefaultValue>;
    /**
     * Passes `null` and `undefined` input values directly to the output without parsing.
     */
    nullish(): AllowShape<AllowShape<this, null>, undefined>;
    /**
     * Replaces `null` and `undefined` input value with a default output value.
     *
     * @param defaultValue The value that should be used if an input value is `undefined` or `null`.
     * @template DefaultValue The value that is used as the replacement for `undefined` and `null`.
     */
    nullish<DefaultValue extends Any>(defaultValue?: DefaultValue): ReplaceShape<ReplaceShape<this, null, DefaultValue>, undefined, DefaultValue>;
    /**
     * Prevents an input and output from being `undefined`.
     *
     * @param options The issue options or the issue message.
     */
    nonOptional(options?: IssueOptions | Message): DenyShape<this, undefined>;
    /**
     * Returns `undefined` if parsing fails.
     */
    catch(): CatchShape<this, undefined>;
    /**
     * Returns the fallback value if parsing fails.
     *
     * @param fallback The value or a callback that returns a value that is returned if parsing has failed. A callback
     * receives an input value, an array of raised issues, and {@link ApplyOptions parsing options}.
     * @template FallbackValue The fallback value.
     */
    catch<FallbackValue extends Any>(fallback: FallbackValue | ((input: any, issues: Issue[], options: ApplyOptions) => FallbackValue)): CatchShape<this, FallbackValue>;
    /**
     * Checks that the input doesn't match the shape.
     *
     * @param shape The shape to which the output must not conform.
     * @param options The issue options or the issue message.
     * @template ExcludedShape The shape to which the output must not conform.
     */
    exclude<ExcludedShape extends AnyShape>(shape: ExcludedShape, options?: IssueOptions | Message): ExcludeShape<this, ExcludedShape>;
    /**
     * Checks that the input doesn't match the shape.
     *
     * This method works exactly as {@link Shape.exclude} at runtime, but it doesn't perform the exclusion on the type
     * level.
     *
     * @param shape The shape to which the output must not conform.
     * @param options The issue options or the issue message.
     * @template ExcludedShape The shape to which the output must not conform.
     */
    not<ExcludedShape extends AnyShape>(shape: ExcludedShape, options?: IssueOptions | Message): NotShape<this, ExcludedShape>;
    /**
     * Must return `true` if the shape must be used in async context only, otherwise the shape can be used in both sync
     * and async contexts. Override this method to implement a custom shape.
     *
     * @see [Advanced shapes](https://github.com/smikhalevski/doubter#advanced-shapes)
     */
    protected _isAsync(): boolean;
    /**
     * Returns input types and literal values that this shape can accept as an input.
     *
     * @see [Advanced shapes](https://github.com/smikhalevski/doubter#advanced-shapes)
     */
    protected _getInputs(): unknown[];
    /**
     * Clones the shape.
     *
     * @see [Advanced shapes](https://github.com/smikhalevski/doubter#advanced-shapes)
     */
    protected _clone(): this;
    /**
     * Synchronously parses the input.
     *
     * **Note:** Don't store or update returned instances of {@link Ok} since they can be reused.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @param nonce The globally unique number that identifies the parsing process.
     * @returns `null` if input matches the output, {@link Ok} that wraps the output, or an array of captured issues.
     * @see [Advanced shapes](https://github.com/smikhalevski/doubter#advanced-shapes)
     */
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<OutputValue>;
    /**
     * Asynchronously parses the input.
     *
     * **Note:** Don't store or update returned instances of {@link Ok} since they can be reused.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @param nonce The globally unique number that identifies the parsing process.
     * @returns `null` if input matches the output, {@link Ok} that wraps the output, or an array of captured issues.
     * @see [Advanced shapes](https://github.com/smikhalevski/doubter#advanced-shapes)
     */
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<OutputValue>>;
}
interface Shape<InputValue, OutputValue> {
    /**
     * The shape input type.
     *
     * @internal
     */
    readonly [INPUT]: InputValue;
    /**
     * The shape output type.
     *
     * @internal
     */
    readonly [OUTPUT]: OutputValue;
    /**
     * The array of unique input types and values that are accepted by the shape.
     */
    readonly inputs: readonly unknown[];
    /**
     * `true` if the shape allows only {@link Shape.parseAsync} and throws an error if {@link Shape.parse} is called, or
     * `false` if the shape can be used in both sync and async contexts.
     */
    readonly isAsync: boolean;
    /**
     * Synchronously parses the value and returns {@link Ok} or {@link Err} object that wraps the result.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The {@link Ok} instance if parsing has succeeded or {@link Err} if parsing has failed.
     * @throws {@link !Error Error} if the shape doesn't support the sync parsing, see {@link Shape.isAsync}.
     */
    try(input: unknown, options?: ApplyOptions): Ok<OutputValue> | Err;
    /**
     * Asynchronously parses the value and returns {@link Ok} or {@link Err} object that wraps the result.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The {@link Ok} instance if parsing has succeeded or {@link Err} if parsing has failed.
     */
    tryAsync(input: unknown, options?: ApplyOptions): Promise<Ok<OutputValue> | Err>;
    /**
     * Synchronously parses the value.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The value that conforms the output type of the shape.
     * @throws {@link !Error Error} if the shape doesn't support the sync parsing, see {@link Shape.isAsync}.
     * @throws {@link ValidationError} if any issues occur during parsing.
     */
    parse(input: unknown, options?: ParseOptions): OutputValue;
    /**
     * Asynchronously parses the value.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The promise that resolves with the value that conforms the output type of the shape, or rejects with a
     * {@link ValidationError} if any issues occur during parsing.
     */
    parseAsync(input: unknown, options?: ParseOptions): Promisify<OutputValue>;
    /**
     * Synchronously parses the value and returns `undefined` if parsing fails.
     *
     * @param input The value to parse.
     * @returns The value that conforms the output type of the shape.
     * @throws {@link !Error Error} if the shape doesn't support the sync parsing, see {@link Shape.isAsync}.
     */
    parseOrDefault(input: unknown): OutputValue | undefined;
    /**
     * Synchronously parses the value and returns the default value if parsing fails.
     *
     * @param input The value to parse.
     * @param defaultValue The default value that is returned if parsing fails.
     * @param options Parsing options.
     * @template DefaultValue The default value that is returned if parsing fails.
     * @returns The value that conforms the output type of the shape.
     * @throws {@link !Error Error} if the shape doesn't support the sync parsing, see {@link Shape.isAsync}.
     */
    parseOrDefault<DefaultValue>(input: unknown, defaultValue: DefaultValue, options?: ApplyOptions): OutputValue | DefaultValue;
    /**
     * Asynchronously parses the value and returns `undefined` value if parsing fails.
     *
     * @param input The value to parse.
     * @returns The value that conforms the output type of the shape.
     */
    parseOrDefaultAsync(input: unknown): Promisify<OutputValue | undefined>;
    /**
     * Asynchronously parses the value and returns the default value if parsing fails.
     *
     * @param input The value to parse.
     * @param defaultValue The default value that is returned if parsing fails.
     * @param options Parsing options.
     * @template DefaultValue The default value that is returned if parsing fails.
     * @returns The value that conforms the output type of the shape.
     */
    parseOrDefaultAsync<DefaultValue>(input: unknown, defaultValue: DefaultValue, options?: ApplyOptions): Promisify<OutputValue | DefaultValue>;
}
/**
 * The shape that applies a converter to the input.
 *
 * @template ConvertedValue The output value of the callback that converts the input value.
 * @group Shapes
 */
declare class ConvertShape<ConvertedValue> extends Shape<any, ConvertedValue> {
    /**
     * The converter that is applied to the shape output value.
     *
     * @param value The input value.
     * @param options Parsing options.
     */
    readonly converter: (value: any, options: ApplyOptions) => PromiseLike<ConvertedValue> | ConvertedValue;
    /**
     * Creates the new {@link ConvertShape} instance.
     *
     * @param converter The callback that converts the input value. Converters can throw or reject with a
     * {@link ValidationError} to notify that the conversion cannot be successfully completed.
     * @param async If `true` then the convert shape waits for the promise returned from the callback to be fulfilled.
     * Otherwise, the value that is synchronously returned from the callback is used as an output.
     * @template ConvertedValue The output value of the callback that converts the input value.
     */
    constructor(
    /**
     * The converter that is applied to the shape output value.
     *
     * @param value The input value.
     * @param options Parsing options.
     */
    converter: (value: any, options: ApplyOptions) => PromiseLike<ConvertedValue> | ConvertedValue, async?: boolean);
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<ConvertedValue>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<ConvertedValue>>;
}
/**
 * The shape that parses the output of the input shape with the output shape.
 *
 * @template InputShape The input shape.
 * @template OutputShape The output shape.
 * @group Shapes
 */
declare class PipeShape<InputShape extends AnyShape, OutputShape extends AnyShape> extends Shape<Input<InputShape>, Output<OutputShape>> implements DeepPartialProtocol<PipeShape<DeepPartialShape<InputShape>, DeepPartialShape<OutputShape>>> {
    /**
     * The input shape.
     */
    readonly inputShape: InputShape;
    /**
     * The output shape.
     */
    readonly outputShape: OutputShape;
    /**
     * Creates the new {@link PipeShape} instance.
     *
     * @param inputShape The shape that parses the input value.
     * @param outputShape The shape that parses the output of `inputShape`.
     * @template InputShape The input shape.
     * @template OutputShape The output shape.
     */
    constructor(
    /**
     * The input shape.
     */
    inputShape: InputShape, 
    /**
     * The output shape.
     */
    outputShape: OutputShape);
    deepPartial(): PipeShape<DeepPartialShape<InputShape>, DeepPartialShape<OutputShape>>;
    protected _isAsync(): boolean;
    protected _getInputs(): unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<Output<OutputShape>>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<Output<OutputShape>>>;
}
/**
 * The shape that replaces an input value with an output value.
 *
 * @template BaseShape The shape that parses the input without the replaced value.
 * @template InputValue The input value to replace.
 * @template OutputValue The output value that is used as the replacement for an input value.
 * @group Shapes
 */
declare class ReplaceShape<BaseShape extends AnyShape, InputValue, OutputValue> extends Shape<Input<BaseShape> | InputValue, ExcludeLiteral<Output<BaseShape>, InputValue> | OutputValue> implements DeepPartialProtocol<ReplaceShape<DeepPartialShape<BaseShape>, InputValue, OutputValue>> {
    /**
     * The shape that parses the input without the replaced value.
     */
    readonly baseShape: BaseShape;
    /**
     * The input value to replace.
     */
    readonly inputValue: InputValue;
    /**
     * The output value that is returned if an {@link ReplaceShape.inputValue} is received.
     */
    readonly outputValue: OutputValue;
    private _result;
    /**
     * Creates the new {@link ReplaceShape} instance.
     *
     * @param baseShape The shape that parses the input without the replaced value.
     * @param inputValue The input value to replace.
     * @param outputValue The output value that is returned if an `inputValue` is received.
     * @template BaseShape The shape that parses the input without the replaced value.
     * @template InputValue The input value to replace.
     * @template OutputValue The output value that is used as the replacement for an input value.
     */
    constructor(
    /**
     * The shape that parses the input without the replaced value.
     */
    baseShape: BaseShape, 
    /**
     * The input value to replace.
     */
    inputValue: InputValue, 
    /**
     * The output value that is returned if an {@link ReplaceShape.inputValue} is received.
     */
    outputValue: OutputValue);
    deepPartial(): ReplaceShape<DeepPartialShape<BaseShape>, InputValue, OutputValue>;
    protected _isAsync(): boolean;
    protected _getInputs(): unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<ExcludeLiteral<Output<BaseShape>, InputValue> | OutputValue>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<ExcludeLiteral<Output<BaseShape>, InputValue> | OutputValue>>;
    private _handleResult;
}
/**
 * The shape that prevents both input and output from being equal to a denied value.
 *
 * @template BaseShape The shape that parses the input without the denied value.
 * @template DeniedValue The denied value.
 * @group Shapes
 */
declare class DenyShape<BaseShape extends AnyShape, DeniedValue> extends Shape<ExcludeLiteral<Input<BaseShape>, DeniedValue>, ExcludeLiteral<Output<BaseShape>, DeniedValue>> implements DeepPartialProtocol<DenyShape<DeepPartialShape<BaseShape>, DeniedValue>> {
    /**
     * The shape that parses the input without the denied value.
     */
    readonly baseShape: BaseShape;
    /**
     * The dined value.
     */
    readonly deniedValue: DeniedValue;
    /**
     * The constraint options or an issue message.
     */
    protected _options: IssueOptions | Message | undefined;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates the new {@link DenyShape} instance.
     *
     * @param baseShape The shape that parses the input without the denied value.
     * @param deniedValue The dined value.
     * @param options The issue options or the issue message.
     * @template BaseShape The shape that parses the input without the denied value.
     * @template DeniedValue The denied value.
     */
    constructor(
    /**
     * The shape that parses the input without the denied value.
     */
    baseShape: BaseShape, 
    /**
     * The dined value.
     */
    deniedValue: DeniedValue, options?: IssueOptions | Message);
    deepPartial(): DenyShape<DeepPartialShape<BaseShape>, DeniedValue>;
    protected _isAsync(): boolean;
    protected _getInputs(): unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<ExcludeLiteral<Output<BaseShape>, DeniedValue>>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<ExcludeLiteral<Output<BaseShape>, DeniedValue>>>;
    private _handleResult;
}
/**
 * The shape that returns the fallback value if parsing fails.
 *
 * @template BaseShape The shape that parses the input.
 * @template FallbackValue The fallback value.
 * @group Shapes
 */
declare class CatchShape<BaseShape extends AnyShape, FallbackValue> extends Shape<Input<BaseShape>, Output<BaseShape> | FallbackValue> implements DeepPartialProtocol<CatchShape<DeepPartialShape<BaseShape>, FallbackValue>> {
    /**
     * The shape that parses the input.
     */
    readonly baseShape: BaseShape;
    /**
     * The value or a callback that returns a value that is returned if parsing has failed.
     *
     *  A callback receives an input value, an array of raised issues, and {@link ApplyOptions parsing options}.
     */
    readonly fallback: FallbackValue | ((input: any, issues: Issue[], options: ApplyOptions) => FallbackValue);
    private _resultProvider;
    /**
     * Creates the new {@link CatchShape} instance.
     *
     * @param baseShape The shape that parses the input.
     * @param fallback The value or a callback that returns a value that is returned if parsing has failed.
     * @template BaseShape The shape that parses the input.
     * @template FallbackValue The fallback value.
     */
    constructor(
    /**
     * The shape that parses the input.
     */
    baseShape: BaseShape, 
    /**
     * The value or a callback that returns a value that is returned if parsing has failed.
     *
     *  A callback receives an input value, an array of raised issues, and {@link ApplyOptions parsing options}.
     */
    fallback: FallbackValue | ((input: any, issues: Issue[], options: ApplyOptions) => FallbackValue));
    deepPartial(): CatchShape<DeepPartialShape<BaseShape>, FallbackValue>;
    protected _isAsync(): boolean;
    protected _getInputs(): unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<Output<BaseShape> | FallbackValue>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<Output<BaseShape> | FallbackValue>>;
    private _handleResult;
}
/**
 * Checks that the input doesn't match the shape.
 *
 * @template BaseShape The base shape.
 * @template ExcludedShape The shape to which the output must not conform.
 * @group Shapes
 */
declare class ExcludeShape<BaseShape extends AnyShape, ExcludedShape extends AnyShape> extends Shape<Input<BaseShape>, Exclude<Output<BaseShape>, Input<ExcludedShape>>> implements DeepPartialProtocol<ExcludeShape<DeepPartialShape<BaseShape>, ExcludedShape>> {
    /**
     * The base shape.
     */
    readonly baseShape: BaseShape;
    /**
     * The shape to which the output must not conform.
     */
    readonly excludedShape: ExcludedShape;
    /**
     * The constraint options or an issue message.
     */
    protected _options: IssueOptions | Message | undefined;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates the new {@link ExcludeShape} instance.
     *
     * @param baseShape The shape that parses the input.
     * @param excludedShape The shape to which the output must not conform.
     * @param options The issue options or the issue message.
     * @template BaseShape The base shape.
     * @template ExcludedShape The shape to which the output must not conform.
     */
    constructor(
    /**
     * The base shape.
     */
    baseShape: BaseShape, 
    /**
     * The shape to which the output must not conform.
     */
    excludedShape: ExcludedShape, options?: IssueOptions | Message);
    deepPartial(): ExcludeShape<DeepPartialShape<BaseShape>, ExcludedShape>;
    protected _isAsync(): boolean;
    protected _getInputs(): unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<Exclude<Output<BaseShape>, Input<ExcludedShape>>>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<Exclude<Output<BaseShape>, Input<ExcludedShape>>>>;
}

type Awaited<T> = T extends null | undefined ? T : T extends object & {
    then(fn: infer F, ...args: any): any;
} ? F extends (value: infer V, ...args: any) => any ? Awaited<V> : never : T;
type Promisify<T> = Promise<Awaited<T>>;
type Awaitable<T> = Awaited<T> extends T ? Promise<T> | T : T;
declare const INPUT: unique symbol;
type INPUT = typeof INPUT;
declare const OUTPUT: unique symbol;
type OUTPUT = typeof OUTPUT;

/**
 * The name of the value type.
 *
 * @group Type Inference
 */
type TypeName = 'array' | 'bigint' | 'boolean' | 'date' | 'function' | 'map' | 'null' | 'number' | 'object' | 'promise' | 'set' | 'string' | 'symbol' | 'undefined' | 'unknown';
/**
 * The enum-like class that describes a value type.
 *
 * @template T The name of the value type.
 * @see {@link Shape.inputs}
 * @group Type Inference
 */
declare class Type<T extends TypeName = TypeName> {
    /**
     * The name of the type.
     */
    readonly name: T;
    static readonly ARRAY: Type<"array">;
    static readonly BIGINT: Type<"bigint">;
    static readonly BOOLEAN: Type<"boolean">;
    static readonly DATE: Type<"date">;
    static readonly FUNCTION: Type<"function">;
    static readonly MAP: Type<"map">;
    static readonly NULL: Type<"null">;
    static readonly NUMBER: Type<"number">;
    static readonly OBJECT: Type<"object">;
    static readonly PROMISE: Type<"promise">;
    static readonly SET: Type<"set">;
    static readonly STRING: Type<"string">;
    static readonly SYMBOL: Type<"symbol">;
    static readonly UNDEFINED: Type<"undefined">;
    static readonly UNKNOWN: Type<"unknown">;
    private constructor();
    /**
     * Returns the type of the given value. If value is a type itself, it is returned as is.
     */
    static of(value: unknown): Type;
    /**
     * @internal
     */
    toString(): string;
}

/**
 * The shape can coerce input value type during parsing.
 *
 * @template InputValue The input value.
 * @template OutputValue The output value.
 * @template CoercedValue The value to which an input is coerced.
 * @group Shapes
 */
declare abstract class CoercibleShape<InputValue = any, OutputValue = InputValue, CoercedValue = InputValue> extends Shape<InputValue, OutputValue> {
    /**
     * `true` if this shapes coerces input values to the required type during parsing, or `false` otherwise.
     */
    isCoercing: boolean;
    /**
     * Enables an input value coercion.
     *
     * @param cb The callback that overrides the built-in coercion.
     * @returns The clone of the shape.
     */
    coerce(
    /**
     * @param value The value to coerce.
     * @returns The coerced value, or {@link NEVER} if coercion isn't possible.
     */
    cb?: (value: unknown) => CoercedValue): this;
    /**
     * Coerces an input value to another type.
     *
     * @param value The value to coerce.
     * @returns The coerced value, or {@link NEVER} if coercion isn't possible.
     */
    protected abstract _coerce(value: unknown): CoercedValue;
}

type InferArray<HeadShapes extends readonly AnyShape[], RestShape extends AnyShape | null, Leg extends INPUT | OUTPUT> = [
    ...{
        [K in keyof HeadShapes]: HeadShapes[K][Leg];
    },
    ...(RestShape extends null | undefined ? [] : RestShape extends Shape ? RestShape[Leg][] : [])
];
type DeepPartialArrayShape<HeadShapes extends readonly AnyShape[], RestShape extends AnyShape | null> = ArrayShape<{
    [K in keyof HeadShapes]: OptionalDeepPartialShape<HeadShapes[K]>;
}, RestShape extends null | undefined ? null : RestShape extends Shape ? OptionalDeepPartialShape<RestShape> : RestShape>;
/**
 * The shape of an array or a tuple value.
 *
 * | Shape | Type |
 * | :-- | :-- |
 * | `ArrayShape<[], null>` | `[]` |
 * | `ArrayShape<[], Shape<B>>` | `B[]` |
 * | `ArrayShape<[Shape<A>, Shape<B>], null>` | `[A, B]` |
 * | `ArrayShape<[Shape<A>], Shape<B>>` | `[A, ...B[]]` |
 *
 * @template HeadShapes The array of positioned element shapes.
 * @template RestShape The shape of rest elements, or `null` if there are no rest elements.
 * @group Shapes
 */
declare class ArrayShape<HeadShapes extends readonly AnyShape[], RestShape extends AnyShape | null> extends CoercibleShape<InferArray<HeadShapes, RestShape, INPUT>, InferArray<HeadShapes, RestShape, OUTPUT>, unknown[]> implements DeepPartialProtocol<DeepPartialArrayShape<HeadShapes, RestShape>> {
    /**
     * The array of positioned element shapes.
     */
    readonly headShapes: HeadShapes;
    /**
     * The shape of rest elements or `null` if there are no rest elements.
     */
    readonly restShape: RestShape;
    /**
     * The type constraint options or the type issue message.
     */
    protected _options: IssueOptions | Message | undefined;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link ArrayShape} instance.
     *
     * @param headShapes The array of positioned element shapes.
     * @param restShape The shape of rest elements or `null` if there are no rest elements.
     * @param options The issue options or the issue message.
     * @template HeadShapes The array of positioned element shapes.
     * @template RestShape The shape of rest elements, or `null` if there are no rest elements.
     */
    constructor(
    /**
     * The array of positioned element shapes.
     */
    headShapes: HeadShapes, 
    /**
     * The shape of rest elements or `null` if there are no rest elements.
     */
    restShape: RestShape, options?: IssueOptions | Message);
    at(key: unknown): AnyShape | null;
    /**
     * Returns an array shape that has rest elements constrained by the given shape.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * @param restShape The shape of rest elements, or `null` if there are no rest elements.
     * @returns The new array shape.
     * @template S The shape of rest elements.
     */
    rest<S extends AnyShape | null>(restShape: S): ArrayShape<HeadShapes, S>;
    deepPartial(): DeepPartialArrayShape<HeadShapes, RestShape>;
    protected _isAsync(): boolean;
    protected _getInputs(): unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<InferArray<HeadShapes, RestShape, OUTPUT>>;
    protected _applyAsync(input: any, options: ApplyOptions, nonce: number): Promise<Result<InferArray<HeadShapes, RestShape, OUTPUT>>>;
    /**
     * Coerces a value to an array.
     *
     * @param value The non-array value to coerce.
     * @returns An array, or {@link NEVER} if coercion isn't possible.
     */
    protected _coerce(value: unknown): unknown[];
}

/**
 * The shape of a bigint value.
 *
 * @group Shapes
 */
declare class BigIntShape extends CoercibleShape<bigint> {
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link BigIntShape} instance.
     *
     * @param options The issue options or the issue message.
     */
    constructor(options?: IssueOptions | Message);
    protected _getInputs(): unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<bigint>;
    /**
     * Coerces a value to a bigint.
     *
     * @param value The non-bigint value to coerce.
     * @returns A bigint value, or {@link NEVER} if coercion isn't possible.
     */
    protected _coerce(value: any): bigint;
}

/**
 * The shape of a boolean value.
 *
 * @group Shapes
 */
declare class BooleanShape extends CoercibleShape<boolean> {
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link BooleanShape} instance.
     *
     * @param options The issue options or the issue message.
     */
    constructor(options?: IssueOptions | Message);
    protected _getInputs(): unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<boolean>;
    /**
     * Coerces a value to a boolean.
     *
     * @param value The non-boolean value to coerce.
     * @returns A boolean value, or {@link NEVER} if coercion isn't possible.
     */
    protected _coerce(value: unknown): boolean;
}

/**
 * The shape of a constant value.
 *
 * @template Value The expected constant value.
 * @group Shapes
 */
declare class ConstShape<Value> extends Shape<Value> {
    /**
     * The expected constant value.
     */
    readonly value: Value;
    /**
     * Returns `true` if an input is equal to the const value, or `false` otherwise.
     */
    protected _typePredicate: (input: unknown) => boolean;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link ConstShape} instance.
     *
     * @param value The expected value.
     * @param options The issue options or the issue message.
     * @template Value The expected value.
     */
    constructor(
    /**
     * The expected constant value.
     */
    value: Value, options?: IssueOptions | Message);
    protected _getInputs(): unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<Value>;
}

/**
 * The shape of the {@link !Date Date} object.
 * @group Shapes
 */
declare class DateShape extends CoercibleShape<Date> {
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link DateShape} instance.
     *
     * @param options The issue options or the issue message.
     */
    constructor(options?: IssueOptions | Message);
    protected _getInputs(): unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<Date>;
    /**
     * Coerces a value to a {@link !Date Date}.
     *
     * @param value The non-{@link !Date Date} value to coerce.
     * @returns A {@link !Date Date} value, or {@link NEVER} if coercion isn't possible.
     */
    protected _coerce(value: unknown): Date;
}

/**
 * The shape of a value enumeration.
 *
 * @template Value The union of allowed enum values.
 * @group Shapes
 */
declare class EnumShape<Value> extends CoercibleShape<Value> {
    /**
     * The array of allowed values, a const key-value mapping, or an TypeScript enum object.
     */
    readonly source: readonly Value[] | ReadonlyDict<Value>;
    /**
     * The array of unique enum values.
     */
    readonly values: readonly Value[];
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link EnumShape} instance.
     *
     * @param source The array of allowed values, a const key-value mapping, or an enum object.
     * @param options The issue options or the issue message.
     * @template Value The union of allowed enum values.
     */
    constructor(
    /**
     * The array of allowed values, a const key-value mapping, or an TypeScript enum object.
     */
    source: readonly Value[] | ReadonlyDict<Value>, options?: IssueOptions | Message);
    protected _getInputs(): unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<Value>;
    /**
     * Coerces a value to an enum value.
     *
     * @param value The non-enum value to coerce.
     * @returns An enum value, or {@link NEVER} if coercion isn't possible.
     */
    protected _coerce(value: any): Value;
}

type InferOrDefault<Shape extends AnyShape | null, Leg extends INPUT | OUTPUT, DefaultValue = any> = Shape extends null | undefined ? DefaultValue : Shape extends AnyShape ? Shape[Leg] : DefaultValue;
type ThisType<F> = F extends (this: infer T, ...args: any[]) => any ? T : any;
/**
 * The shape of a function.
 *
 * @template ArgsShape The shape of the array of arguments.
 * @template ReturnShape The return value shape, or `null` if unconstrained.
 * @template ThisShape The shape of `this` argument, or `null` if unconstrained.
 * @group Shapes
 */
declare class FunctionShape<ArgsShape extends Shape<readonly any[], readonly any[]>, ReturnShape extends AnyShape | null, ThisShape extends AnyShape | null> extends Shape<(this: InferOrDefault<ThisShape, OUTPUT>, ...args: Output<ArgsShape>) => InferOrDefault<ReturnShape, INPUT>, (this: InferOrDefault<ThisShape, INPUT>, ...args: Input<ArgsShape>) => InferOrDefault<ReturnShape, OUTPUT>> {
    /**
     * The shape of the array of arguments.
     */
    readonly argsShape: ArgsShape;
    /**
     * The return value shape, or `null` if unconstrained.
     */
    readonly returnShape: ReturnShape;
    /**
     * The shape of `this` value, or `null` if unconstrained.
     */
    readonly thisShape: ThisShape;
    /**
     * `true` if input functions are wrapped during parsing to ensure runtime signature type-safety, or `false` otherwise.
     */
    isStrict: boolean;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Parsing options that are used by a wrapper.
     */
    protected _parseOptions: ParseOptions | undefined;
    /**
     * Creates a new {@link FunctionShape} instance.
     *
     * @param argsShape The shape of the array of arguments.
     * @param returnShape The return value shape, or `null` if unconstrained.
     * @param thisShape The shape of `this` argument, or `null` if unconstrained.
     * @param options The issue options or the issue message.
     * @template ArgsShape The shape of the array of arguments.
     * @template ReturnShape The return value shape, or `null` if unconstrained.
     * @template ThisShape The shape of `this` argument, or `null` if unconstrained.
     */
    constructor(
    /**
     * The shape of the array of arguments.
     */
    argsShape: ArgsShape, 
    /**
     * The return value shape, or `null` if unconstrained.
     */
    returnShape: ReturnShape, 
    /**
     * The shape of `this` value, or `null` if unconstrained.
     */
    thisShape: ThisShape, options?: IssueOptions | Message);
    /**
     * `true` if some shapes that describe the function signature are {@link Shape#isAsync async}, or `false` otherwise.
     */
    get isAsyncFunction(): boolean;
    /**
     * Constrains the function return value with the given shape.
     *
     * @param shape The return value shape, or `null` if unconstrained.
     * @returns The new function shape.
     * @template S The return value shape.
     */
    return<S extends AnyShape | null>(shape: S): FunctionShape<ArgsShape, S, ThisShape>;
    /**
     * Constrains the type of `this` inside the function.
     *
     * @param shape The shape of `this` argument, or `null` if unconstrained.
     * @returns The new function shape.
     * @template S The shape of `this` argument.
     */
    this<S extends AnyShape | null>(shape: S): FunctionShape<ArgsShape, ReturnShape, S>;
    /**
     * Enables input function wrapping during parsing to ensure runtime signature type-safety. Wrapper ensures that input
     * function receives arguments and `this` values that conform {@link FunctionShape.argsShape} and
     * {@link FunctionShape.thisShape} respectively, and returns the value that conforms
     * {@link FunctionShape.returnShape}.
     *
     * @param options Options that are used by the wrapper. If omitted then default options are applied: no early-return,
     * no type coercion.
     * @returns The new function shape.
     */
    strict(options?: ParseOptions): this;
    /**
     * Creates a function that ensures that `fn` receives arguments and `this` value that conform the
     * {@linkcode FunctionShape.argsShape} and the {@linkcode FunctionShape.thisShape} respectively, and _synchronously_
     * returns the value that conforms the {@linkcode FunctionShape.returnShape}.
     *
     * @param fn The underlying function.
     * @param options Parsing options. By default, options provided to {@linkcode FunctionShape.strict} are used.
     * @returns The wrapper function.
     * @template F The function to which signature must be ensured.
     */
    ensure<F extends (this: InferOrDefault<ThisShape, OUTPUT>, ...args: Output<ArgsShape>) => InferOrDefault<ReturnShape, INPUT>>(fn: F, options?: ParseOptions): (this: InferOrDefault<ThisShape, INPUT, ThisType<F>>, ...args: Input<ArgsShape>) => InferOrDefault<ReturnShape, OUTPUT, ReturnType<F>>;
    /**
     * Creates a function that ensures that `fn` receives arguments and `this` value that conform the
     * {@linkcode FunctionShape.argsShape} and the {@linkcode FunctionShape.thisShape} respectively, and _asynchronously_
     * returns the value that conforms the {@linkcode FunctionShape.returnShape}.
     *
     * Use this method if {@link FunctionShape#isAsyncFunction some shapes that describe the function signature} are
     * {@link Shape#isAsync async}.
     *
     * @param fn The underlying function.
     * @param options Parsing options. By default, options provided to {@linkcode FunctionShape.strict} are used.
     * @returns The wrapper function.
     * @template F The function to which signature must be ensured.
     */
    ensureAsync<F extends (this: InferOrDefault<ThisShape, OUTPUT>, ...args: Output<ArgsShape>) => Awaitable<InferOrDefault<ReturnShape, INPUT>>>(fn: F, options?: ParseOptions): (this: InferOrDefault<ThisShape, INPUT, ThisType<F>>, ...args: Input<ArgsShape>) => Promisify<InferOrDefault<ReturnShape, OUTPUT, ReturnType<F>>>;
    protected _getInputs(): unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<(this: InferOrDefault<ThisShape, INPUT>, ...args: Input<ArgsShape>) => InferOrDefault<ReturnShape, OUTPUT>>;
}

/**
 * The shape of the class instance.
 *
 * @template Ctor The class constructor.
 * @group Shapes
 */
declare class InstanceShape<Ctor extends new (...args: any) => any> extends Shape<InstanceType<Ctor>> {
    readonly ctor: Ctor;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link InstanceShape} instance.
     *
     * @param ctor The class constructor.
     * @param options The issue options or the issue message.
     * @template Ctor The class constructor.
     */
    constructor(ctor: Ctor, options?: IssueOptions | Message);
    protected _getInputs(): unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<InstanceType<Ctor>>;
}

/**
 * Converts union to intersection.
 */
type Intersect<U extends AnyShape> = (U extends any ? (k: U) => void : never) extends (k: infer I) => void ? I extends Shape ? I : never : never;
type DeepPartialIntersectionShape<Shapes extends readonly AnyShape[]> = IntersectionShape<{
    [K in keyof Shapes]: DeepPartialShape<Shapes[K]>;
}>;
/**
 * The shape that requires an input to conform all given shapes.
 *
 * @template Shapes The array of shapes that comprise an intersection.
 * @group Shapes
 */
declare class IntersectionShape<Shapes extends readonly AnyShape[]> extends Shape<Input<Intersect<Shapes[number]>>, Output<Intersect<Shapes[number]>>> implements DeepPartialProtocol<DeepPartialIntersectionShape<Shapes>> {
    /**
     * The array of shapes that comprise an intersection.
     */
    readonly shapes: Shapes;
    /**
     * The intersection constraint options or an issue message.
     */
    protected _options: IssueOptions | Message | undefined;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link IntersectionShape} instance.
     *
     * @param shapes The array of shapes that comprise an intersection.
     * @param options The intersection constraint options or an issue message.
     * @template Shapes The array of shapes that comprise an intersection.
     */
    constructor(
    /**
     * The array of shapes that comprise an intersection.
     */
    shapes: Shapes, options?: IssueOptions | Message);
    at(key: unknown): AnyShape | null;
    deepPartial(): DeepPartialIntersectionShape<Shapes>;
    protected _isAsync(): boolean;
    protected _getInputs(): unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<Output<Intersect<Shapes[number]>>>;
    protected _applyAsync(input: any, options: ApplyOptions, nonce: number): Promise<Result<Output<Intersect<Shapes[number]>>>>;
    /**
     * Merge values and apply checks.
     */
    private _applyIntersection;
}

/**
 * Lazily loads a shape using the provider callback.
 *
 * @template ProvidedShape The lazy-loaded shape.
 * @template Pointer The value returned when a cyclic reference is detected.
 * @group Shapes
 */
declare class LazyShape<ProvidedShape extends AnyShape, Pointer> extends Shape<Input<ProvidedShape>, Output<ProvidedShape> | Pointer> implements DeepPartialProtocol<LazyShape<DeepPartialShape<ProvidedShape>, Pointer>> {
    /**
     * The provider callback that returns the shape to which {@link LazyShape} delegates input handling.
     */
    readonly shapeProvider: () => ProvidedShape;
    /**
     * The provider callback that returns the value that is used instead of a circular reference.
     */
    readonly pointerProvider: (value: Input<ProvidedShape>, options: ApplyOptions) => Pointer;
    /**
     * The provider that caches the returned shape.
     */
    private _cachingShapeProvider;
    /**
     * The map from nonce to an array of inputs seen during parsing.
     */
    private _stackMap;
    /**
     * Creates a new {@link LazyShape} instance.
     *
     * @param shapeProvider The provider callback that returns the shape to which {@link LazyShape} delegates input
     * handling. The provider is called only once.
     * @param pointerProvider The provider callback that returns the value that is used instead of a circular
     * reference.
     * @template ProvidedShape The lazy-loaded shape.
     * @template Pointer The value returned when a cyclic reference is detected.
     */
    constructor(
    /**
     * The provider callback that returns the shape to which {@link LazyShape} delegates input handling.
     */
    shapeProvider: () => ProvidedShape, 
    /**
     * The provider callback that returns the value that is used instead of a circular reference.
     */
    pointerProvider: (value: Input<ProvidedShape>, options: ApplyOptions) => Pointer);
    /**
     * The lazy-loaded shape.
     */
    get providedShape(): ProvidedShape;
    at(key: unknown): AnyShape | null;
    deepPartial(): LazyShape<DeepPartialShape<ProvidedShape>, Input<DeepPartialShape<ProvidedShape>>>;
    /**
     * Replace circular references with a placeholder.
     *
     * @param pointer The value that would be used instead of circular references, or a callback that returns such a
     * value.
     * @template Pointer The value returned when a cyclic reference is detected.
     * @returns The clone of the shape.
     */
    circular<Pointer extends Any>(pointer: Pointer | ((value: Input<ProvidedShape>, options: ApplyOptions) => Pointer)): LazyShape<ProvidedShape, Pointer>;
    protected _isAsync(): boolean;
    protected _getInputs(): unknown[];
    protected _clone(): this;
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<Output<ProvidedShape> | Pointer>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<Output<ProvidedShape> | Pointer>>;
    private _handleResult;
}

/**
 * The shape of a {@link !Map Map} instance.
 *
 * @template KeyShape The key shape.
 * @template ValueShape The value shape.
 * @group Shapes
 */
declare class MapShape<KeyShape extends AnyShape, ValueShape extends AnyShape> extends CoercibleShape<Map<Input<KeyShape>, Input<ValueShape>>, Map<Output<KeyShape>, Output<ValueShape>>, [
    unknown,
    unknown
][]> implements DeepPartialProtocol<MapShape<DeepPartialShape<KeyShape>, OptionalDeepPartialShape<ValueShape>>> {
    /**
     * The key shape.
     */
    readonly keyShape: KeyShape;
    /**
     * The value shape.
     */
    readonly valueShape: ValueShape;
    /**
     * The issue options or the issue message.
     */
    protected _options: IssueOptions | Message | undefined;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link MapShape} instance.
     *
     * @param keyShape The key shape.
     * @param valueShape The value shape.
     * @param options The issue options or the issue message.
     * @template KeyShape The key shape.
     * @template ValueShape The value shape.
     */
    constructor(
    /**
     * The key shape.
     */
    keyShape: KeyShape, 
    /**
     * The value shape.
     */
    valueShape: ValueShape, options?: IssueOptions | Message);
    at(key: unknown): AnyShape | null;
    deepPartial(): MapShape<DeepPartialShape<KeyShape>, OptionalDeepPartialShape<ValueShape>>;
    protected _isAsync(): boolean;
    protected _getInputs(): unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<Map<Output<KeyShape>, Output<ValueShape>>>;
    protected _applyAsync(input: any, options: ApplyOptions, nonce: number): Promise<Result<Map<Output<KeyShape>, Output<ValueShape>>>>;
    /**
     * Coerces a value to an array of {@link !Map Map} entries.
     *
     * @param value A non-{@link !Map Map} value to coerce.
     * @returns An array of entries, or {@link NEVER} if coercion isn't possible.
     */
    protected _coerce(value: any): [unknown, unknown][];
}

/**
 * The shape that doesn't match any input.
 *
 * @group Shapes
 */
declare class NeverShape extends Shape<never> {
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link NeverShape} instance.
     *
     * @param options The issue options or the issue message.
     */
    constructor(options?: IssueOptions | Message);
    protected _getInputs(): unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<never>;
}

/**
 * The shape of a number value.
 *
 * @group Shapes
 */
declare class NumberShape extends CoercibleShape<number> {
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link NumberShape} instance.
     *
     * @param options The issue options or the issue message.
     */
    constructor(options?: IssueOptions | Message);
    /**
     * Allows `NaN` as an input and output value.
     */
    nan(): AllowShape<this, number>;
    /**
     * Replaces an input `NaN` value with a default output value.
     *
     * @param defaultValue The value that is used instead of `NaN` in the output.
     */
    nan<T extends Any>(defaultValue: T): ReplaceShape<this, number, T>;
    protected _getInputs(): unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<number>;
    /**
     * Coerces a value to a number (not `NaN`).
     *
     * @param value The non-number value to coerce.
     * @returns A number value, or {@link NEVER} if coercion isn't possible.
     */
    protected _coerce(value: any): number;
}

type InferObject<PropShapes extends ReadonlyDict<AnyShape>, RestShape extends AnyShape | null, Leg extends INPUT | OUTPUT> = Prettify<UndefinedToOptional<{
    [K in keyof PropShapes]: PropShapes[K][Leg];
}> & (RestShape extends null | undefined ? {} : RestShape extends Shape ? {
    [key: string]: RestShape[Leg];
} : {})>;
type UndefinedToOptional<T> = Omit<T, OptionalKeys<T>> & {
    [K in OptionalKeys<T>]?: T[K];
};
type OptionalKeys<T> = {
    [K in keyof T]: undefined extends Extract<T[K], undefined> ? K : never;
}[keyof T];
type Prettify<T> = {
    [K in keyof T]: T[K];
} & {};
type OptionalPropShapes<PropShapes extends ReadonlyDict<AnyShape>> = {
    [K in keyof PropShapes]: AllowShape<PropShapes[K], undefined>;
};
type RequiredPropShapes<PropShapes extends ReadonlyDict<AnyShape>> = {
    [K in keyof PropShapes]: DenyShape<PropShapes[K], undefined>;
};
type DeepPartialObjectShape<PropShapes extends ReadonlyDict<AnyShape>, RestShape extends AnyShape | null> = ObjectShape<{
    [K in keyof PropShapes]: OptionalDeepPartialShape<PropShapes[K]>;
}, RestShape extends null | undefined ? null : RestShape extends Shape ? OptionalDeepPartialShape<RestShape> : RestShape>;
/**
 * Defines how object keys are handled.
 *
 * @see {@link ObjectShape.keysMode}
 * @group Other
 */
type ObjectKeysMode = 'preserved' | 'stripped' | 'exact';
/**
 * The shape of an object.
 *
 * @template PropShapes The mapping from a string object key to a corresponding value shape.
 * @template RestShape The shape that constrains values of
 * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures), or `null`
 * if there's no index signature.
 * @group Shapes
 */
declare class ObjectShape<PropShapes extends ReadonlyDict<AnyShape>, RestShape extends AnyShape | null> extends Shape<InferObject<PropShapes, RestShape, INPUT>, InferObject<PropShapes, RestShape, OUTPUT>> implements DeepPartialProtocol<DeepPartialObjectShape<PropShapes, RestShape>> {
    /**
     * The mapping from an object key to a corresponding value shape.
     */
    readonly propShapes: PropShapes;
    /**
     * The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    readonly restShape: RestShape;
    /**
     * The mode of keys handling.
     */
    readonly keysMode: ObjectKeysMode;
    /**
     * The array of known object keys.
     */
    readonly keys: readonly string[];
    /**
     * The array of property shapes, parallel to {@link ObjectShape.keys}.
     */
    readonly valueShapes: readonly Shape[];
    /**
     * The issue options or the issue message.
     */
    protected _options: IssueOptions | Message | undefined;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Returns issues which describe that an object has unknown properties.
     */
    protected _exactIssueFactory?: (input: unknown, options: ApplyOptions, param: unknown) => Issue;
    /**
     * Creates a new {@link ObjectShape} instance.
     *
     * @param propShapes The mapping from an object key to a corresponding value shape.
     * @param restShape The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures). If `null`
     * then values thea fall under the index signature are unconstrained.
     * @param options The issue options or the issue message.
     * @param keysMode The mode of keys handling.
     * @template PropShapes The mapping from an object key to a corresponding value shape.
     * @template RestShape The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    constructor(
    /**
     * The mapping from an object key to a corresponding value shape.
     */
    propShapes: PropShapes, 
    /**
     * The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    restShape: RestShape, options?: IssueOptions | Message, 
    /**
     * The mode of keys handling.
     */
    keysMode?: ObjectKeysMode);
    /**
     * The enum shape that describes object keys.
     */
    get keysShape(): EnumShape<keyof PropShapes>;
    at(key: any): AnyShape | null;
    /**
     * Merge properties from the other object shape.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * If a property with the same key already exists on this object shape then it is overwritten. The index signature of
     * this shape and its {@link ObjectShape.keysMode} is preserved intact.
     *
     * @param shape The object shape which properties must be added to this object shape.
     * @returns The new object shape.
     * @template T Properties to add.
     */
    extend<T extends ReadonlyDict<AnyShape>>(shape: ObjectShape<T, any>): ObjectShape<Omit<PropShapes, keyof T> & T, RestShape>;
    /**
     * Add properties to an object shape.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * If a property with the same key already exists on this object shape then it is overwritten. The index signature of
     * this shape and its {@link ObjectShape.keysMode} is preserved intact.
     *
     * @param shapes The properties to add.
     * @returns The new object shape.
     * @template T The shapes of properties to add.
     */
    extend<T extends ReadonlyDict<AnyShape>>(shapes: T): ObjectShape<Omit<PropShapes, keyof T> & T, RestShape>;
    /**
     * Returns an object shape that only has properties with listed keys.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * @param keys The array of property keys to pick.
     * @returns The new object shape.
     * @template K The tuple of keys to pick.
     */
    pick<K extends ReadonlyArray<keyof PropShapes>>(keys: K): ObjectShape<Pick<PropShapes, K[number]>, RestShape>;
    /**
     * Returns an object shape that doesn't have the listed keys.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * @param keys The array of property keys to omit.
     * @returns The new object shape.
     * @template K The tuple of keys to omit.
     */
    omit<K extends ReadonlyArray<keyof PropShapes>>(keys: K): ObjectShape<Omit<PropShapes, K[number]>, RestShape>;
    /**
     * Returns an object shape with all properties marked as optional.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * @returns The new object shape.
     */
    partial(): ObjectShape<OptionalPropShapes<PropShapes>, RestShape>;
    /**
     * Returns an object shape with keys marked as optional.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * @param keys The array of property keys to make optional.
     * @returns The new object shape.
     * @template K The array of string keys.
     */
    partial<K extends ReadonlyArray<keyof PropShapes>>(keys: K): ObjectShape<Omit<PropShapes, K[number]> & OptionalPropShapes<Pick<PropShapes, K[number]>>, RestShape>;
    deepPartial(): DeepPartialObjectShape<PropShapes, RestShape>;
    /**
     * Returns an object shape with all properties marked as required.
     *
     * @returns The new object shape.
     */
    required(): ObjectShape<RequiredPropShapes<PropShapes>, RestShape>;
    /**
     * Returns an object shape with keys marked as required.
     *
     * @param keys The array of property keys to make required.
     * @returns The new object shape.
     * @template K The array of string keys.
     */
    required<K extends ReadonlyArray<keyof PropShapes>>(keys: K): ObjectShape<Omit<PropShapes, K[number]> & RequiredPropShapes<Pick<PropShapes, K[number]>>, RestShape>;
    /**
     * Returns an object shape that allows only known keys and has no index signature.
     *
     * @param options The issue options or the issue message.
     * @returns The new object shape.
     */
    exact(options?: IssueOptions | Message): ObjectShape<PropShapes, null>;
    /**
     * Returns an object shape that doesn't have an index signature and all unknown keys are stripped.
     *
     * @returns The new object shape.
     */
    strip(): ObjectShape<PropShapes, null>;
    /**
     * Returns an object shape that has an index signature that doesn't constrain values.
     *
     * @returns The new object shape.
     */
    preserve(): ObjectShape<PropShapes, null>;
    /**
     * Returns an object shape that has an index signature that is constrained by the given shape.
     *
     * @param restShape The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures), or `null`
     * if there's no index signature.
     * @returns The new object shape.
     * @template S The index signature shape.
     */
    rest<S extends AnyShape | null>(restShape: S): ObjectShape<PropShapes, S>;
    protected _isAsync(): boolean;
    protected _getInputs(): unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<InferObject<PropShapes, RestShape, OUTPUT>>;
    protected _applyAsync(input: any, options: ApplyOptions, nonce: number): Promise<Result<InferObject<PropShapes, RestShape, OUTPUT>>>;
    /**
     * Unknown keys are preserved as is and aren't checked.
     */
    private _applyRestUnchecked;
    /**
     * Unknown keys are either parsed with a {@link ObjectShape.restShape}, stripped, or cause an issue.
     */
    private _applyRestChecked;
}

type InferPromise<ValueShape extends AnyShape | null, Leg extends INPUT | OUTPUT> = Promisify<ValueShape extends null | undefined ? any : ValueShape extends AnyShape ? ValueShape[Leg] : any>;
type DeepPartialPromiseShape<ValueShape extends AnyShape | null> = PromiseShape<ValueShape extends null | undefined ? null : ValueShape extends AnyShape ? OptionalDeepPartialShape<ValueShape> : null>;
/**
 * The shape of a {@link !Promise Promise} value.
 *
 * @template ValueShape The shape of the resolved value, or `null` if resolved value shouldn't be parsed.
 * @group Shapes
 */
declare class PromiseShape<ValueShape extends AnyShape | null> extends CoercibleShape<InferPromise<ValueShape, INPUT>, InferPromise<ValueShape, OUTPUT>, Promise<any>> implements DeepPartialProtocol<DeepPartialPromiseShape<ValueShape>> {
    readonly valueShape: ValueShape;
    /**
     * The type constraint options or the type issue message.
     */
    protected _options: IssueOptions | Message | undefined;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link PromiseShape} instance.
     *
     * @param valueShape The shape of the resolved value, or `null` if resolved value shouldn't be parsed.
     * @param options The issue options or the issue message.
     * @template ValueShape The shape of the resolved value.
     */
    constructor(valueShape: ValueShape, options?: IssueOptions | Message);
    deepPartial(): DeepPartialPromiseShape<ValueShape>;
    protected _isAsync(): boolean;
    protected _getInputs(): unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<InferPromise<ValueShape, OUTPUT>>;
    protected _applyAsync(input: any, options: ApplyOptions, nonce: number): Promise<Result<InferPromise<ValueShape, OUTPUT>>>;
    protected _coerce(value: unknown): Promise<any>;
}

type InferRecord<KeyShape extends Shape<string, PropertyKey> | null, ValueShape extends AnyShape, Leg extends INPUT | OUTPUT> = Record<KeyShape extends null | undefined ? string : KeyShape extends Shape ? KeyShape[Leg] : string, ValueShape[Leg]>;
/**
 * The shape that describes an object with string keys and values that conform the given shape.
 *
 * @template KeyShape The key shape.
 * @template ValueShape The value shape.
 * @group Shapes
 */
declare class RecordShape<KeyShape extends Shape<string, PropertyKey> | null, ValueShape extends AnyShape> extends Shape<InferRecord<KeyShape, ValueShape, INPUT>, InferRecord<KeyShape, ValueShape, OUTPUT>> implements DeepPartialProtocol<RecordShape<KeyShape, OptionalDeepPartialShape<ValueShape>>> {
    /**
     * The key shape or `null` if keys are preserved intact.
     */
    readonly keyShape: KeyShape;
    /**
     * The value shape.
     */
    readonly valueShape: ValueShape;
    /**
     * The issue options or the issue message.
     */
    protected _options: IssueOptions | Message | undefined;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link RecordShape} instance.
     *
     * @param keyShape The key shape, or `null` if keys should be preserved intact.
     * @param valueShape The value shape.
     * @param options The issue options or the issue message.
     * @template KeyShape The key shape.
     * @template ValueShape The value shape.
     */
    constructor(
    /**
     * The key shape or `null` if keys are preserved intact.
     */
    keyShape: KeyShape, 
    /**
     * The value shape.
     */
    valueShape: ValueShape, options?: IssueOptions | Message);
    at(key: unknown): AnyShape | null;
    deepPartial(): RecordShape<KeyShape, OptionalDeepPartialShape<ValueShape>>;
    protected _isAsync(): boolean;
    protected _getInputs(): unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<InferRecord<KeyShape, ValueShape, OUTPUT>>;
    protected _applyAsync(input: any, options: ApplyOptions, nonce: number): Promise<Result<InferRecord<KeyShape, ValueShape, OUTPUT>>>;
}

/**
 * The shape of a {@link !Set Set} instance.
 *
 * @template ValueShape The value shape.
 * @group Shapes
 */
declare class SetShape<ValueShape extends AnyShape> extends CoercibleShape<Set<Input<ValueShape>>, Set<Output<ValueShape>>, unknown[]> implements DeepPartialProtocol<SetShape<OptionalDeepPartialShape<ValueShape>>> {
    /**
     * The value shape.
     */
    readonly valueShape: ValueShape;
    /**
     * The type constraint options or the type issue message.
     */
    protected _options: IssueOptions | Message | undefined;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link SetShape} instance.
     *
     * @param valueShape The value shape.
     * @param options The issue options or the issue message.
     * @template ValueShape The value shape.
     */
    constructor(
    /**
     * The value shape.
     */
    valueShape: ValueShape, options?: IssueOptions | Message);
    at(key: unknown): AnyShape | null;
    deepPartial(): SetShape<OptionalDeepPartialShape<ValueShape>>;
    protected _isAsync(): boolean;
    protected _getInputs(): unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<Set<Output<ValueShape>>>;
    protected _applyAsync(input: any, options: ApplyOptions, nonce: number): Promise<Result<Set<Output<ValueShape>>>>;
    /**
     * Coerces a value to an array of {@link !Set Set} values.
     *
     * @param value The non-{@link !Set Set} value to coerce.
     * @returns An array, or {@link NEVER} if coercion isn't possible.
     */
    protected _coerce(value: unknown): unknown[];
}

/**
 * The shape of a string value.
 *
 * @group Shapes
 */
declare class StringShape extends CoercibleShape<string> {
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link StringShape} instance.
     *
     * @param options The issue options or the issue message.
     */
    constructor(options?: IssueOptions | Message);
    protected _getInputs(): unknown[];
    protected _apply(input: any, options: ApplyOptions, nonce: number): Result<string>;
    /**
     * Coerces a value to a string.
     *
     * @param value The non-string value to coerce.
     * @returns A string value, or {@link NEVER} if coercion isn't possible.
     */
    protected _coerce(value: unknown): string;
}

/**
 * The shape of an arbitrary symbol value.
 *
 * @group Shapes
 */
declare class SymbolShape extends Shape<symbol> {
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions) => Issue;
    /**
     * Creates a new {@link SymbolShape} instance.
     *
     * @param options The issue options or the issue message.
     */
    constructor(options?: IssueOptions | Message);
    protected _getInputs(): unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<symbol>;
}

/**
 * Returns the array of shapes that are applicable to the input.
 */
type LookupCallback = (input: any) => readonly AnyShape[];
type DeepPartialUnionShape<Shapes extends readonly AnyShape[]> = UnionShape<{
    [K in keyof Shapes]: DeepPartialShape<Shapes[K]>;
}>;
/**
 * The shape that requires an input to conform at least one of shapes.
 *
 * @template Shapes The array of shapes that comprise a union.
 * @group Shapes
 */
declare class UnionShape<Shapes extends readonly AnyShape[]> extends Shape<Input<Shapes[number]>, Output<Shapes[number]>> implements DeepPartialProtocol<DeepPartialUnionShape<Shapes>> {
    /**
     * The array of shapes that comprise a union.
     */
    readonly shapes: Shapes;
    /**
     * The union constraint options or an issue message.
     */
    protected _options: IssueOptions | Message | undefined;
    /**
     * Returns issues associated with an invalid input value type.
     */
    protected _typeIssueFactory: (input: unknown, options: ApplyOptions, param: unknown) => Issue;
    /**
     * Creates a new {@link UnionShape} instance.
     *
     * @param shapes The array of shapes that comprise a union.
     * @param options The union constraint options or an issue message.
     * @template Shapes The array of shapes that comprise a union.
     */
    constructor(
    /**
     * The array of shapes that comprise a union.
     */
    shapes: Shapes, options?: IssueOptions | Message);
    protected get _lookup(): LookupCallback;
    at(key: unknown): AnyShape | null;
    deepPartial(): DeepPartialUnionShape<Shapes>;
    protected _isAsync(): boolean;
    protected _getInputs(): unknown[];
    protected _apply(input: unknown, options: ApplyOptions, nonce: number): Result<Output<Shapes[number]>>;
    protected _applyAsync(input: unknown, options: ApplyOptions, nonce: number): Promise<Result<Output<Shapes[number]>>>;
}

/**
 * Creates the unconstrained shape.
 *
 * You can specify compile-time type to enhance type inference.
 *
 * This provides _no runtime type-safety_!
 *
 * @template Value The input and the output value.
 * @group DSL
 */
declare function any<Value = any>(): Shape<Value>;
/**
 * Creates the shape that is constrained with a
 * [narrowing predicate](https://www.typescriptlang.org/docs/handbook/2/narrowing.html).
 *
 * @param cb The type predicate that returns `true` if value conforms the required type, or `false` otherwise.
 * @param options The issue options or the issue message.
 * @returns The shape that has the narrowed output.
 * @template Value The input and the output value.
 * @group DSL
 */
declare function any<Value>(
/**
 * @param value The input value.
 * @param options Parsing options.
 */
cb: (value: any, options: ApplyOptions) => value is Value, options?: RefineOptions | Message): Shape<Value>;
/**
 * Creates the shape that is constrained with a predicate.
 *
 * @param cb The predicate that returns truthy result if value is valid, or returns falsy result otherwise.
 * @param options The issue options or the issue message.
 * @template Value The input and the output value.
 * @group DSL
 */
declare function any<Value = any>(
/**
 * @param value The input value.
 * @param options Parsing options.
 */
cb: (value: any, options: ApplyOptions) => boolean, options?: RefineOptions | Message): Shape<Value>;

/**
 * Creates the unconstrained array shape.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
declare function array(options?: IssueOptions | Message): ArrayShape<[], Shape>;
/**
 * Creates the array shape with elements that conform the element shape.
 *
 * @param shape The shape of an array element.
 * @param options The issue options or the issue message.
 * @template ValueShape The shape of an array element.
 * @group DSL
 */
declare function array<ValueShape extends AnyShape>(shape: ValueShape, options?: IssueOptions | Message): ArrayShape<[], ValueShape>;

/**
 * Creates the bigint shape.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
declare function bigint(options?: IssueOptions | Message): BigIntShape;

/**
 * Creates the boolean shape.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
declare function boolean(options?: IssueOptions | Message): BooleanShape;

/**
 * Creates the constant value shape.
 *
 * @param value The value to which the input must be strictly equal.
 * @param options The issue options or the issue message.
 * @template Value The expected value.
 * @group DSL
 */
declare function const_<Value extends Any>(value: Value, options?: IssueOptions | Message): ConstShape<Value>;

/**
 * Creates the {@link !Date Date} shape.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
declare function date(options?: IssueOptions | Message): DateShape;

/**
 * Creates the shape that constrains input with the array of values.
 *
 * @param values The array of values allowed for the input.
 * @param options The issue options or the issue message.
 * @template Value The union of allowed enum values.
 * @template ValuesArray The array of allowed values.
 * @group DSL
 */
declare function enum_<Value extends Any, ValuesArray extends readonly [Value, ...Value[]]>(values: ValuesArray, options?: IssueOptions | Message): EnumShape<ValuesArray[number]>;
/**
 * Creates the shape that constrains input with values of
 * [the enum-like object](https://www.typescriptlang.org/docs/handbook/enums.html).
 *
 * @param values The native enum or a mapping object.
 * @param options The issue options or the issue message.
 * @template Value The union of allowed enum values.
 * @template ValuesDict The object that maps from the key to an enum value.
 * @group DSL
 */
declare function enum_<Value extends Any, ValuesDict extends ReadonlyDict<Value>>(values: ValuesDict, options?: IssueOptions | Message): EnumShape<ValuesDict[keyof ValuesDict]>;

/**
 * Creates a shape of a function that has no arguments.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
declare function function_(options?: IssueOptions | Message): FunctionShape<ArrayShape<[], null>, null, null>;
/**
 * Creates a shape of a function with arguments parsed by corresponding shapes in the `argShapes` array.
 *
 * @param argShapes The array of argument shapes.
 * @param options The issue options or the issue message.
 * @template ArgShapes The array of argument shapes.
 * @group DSL
 */
declare function function_<ArgShapes extends readonly [AnyShape, ...AnyShape[]] | []>(argShapes: ArgShapes, options?: IssueOptions | Message): FunctionShape<ArrayShape<ArgShapes, null>, null, null>;
/**
 * Creates a shape of a function with arguments parsed by an array shape.
 *
 * @param argsShape The shape of the array of arguments.
 * @param options The issue options or the issue message.
 * @template InputArgs The array of input arguments.
 * @template OutputArgs The array of input arguments.
 * @template ArgsShape The shape of the array of arguments.
 * @group DSL
 */
declare function function_<InputArgs extends readonly any[], OutputArgs extends readonly any[], ArgsShape extends Shape<InputArgs, OutputArgs>>(argsShape: ArgsShape, options?: IssueOptions | Message): FunctionShape<ArgsShape, null, null>;

/**
 * Creates the class instance shape.
 *
 * @param ctor The instance constructor.
 * @param options The issue options or the issue message.
 * @template Ctor The instance constructor.
 * @group DSL
 */
declare function instanceOf<Ctor extends new (...args: any[]) => any>(ctor: Ctor, options?: IssueOptions | Message): InstanceShape<Ctor>;

/**
 * Creates an intersection shape that tries to parse the input with all provided shapes and merge parsing results.
 *
 * @param shapes The array of shapes.
 * @param options The issue options or the issue message.
 * @template Shapes The tuple of intersected shapes.
 * @group DSL
 */
declare function intersection<Shapes extends [AnyShape, ...AnyShape[]]>(shapes: Shapes, options?: IssueOptions | Message): IntersectionShape<Shapes>;

/**
 * Creates the shape that resolves the underlying shape on-demand.
 *
 * @param shapeProvider The provider that returns the resolved shape.
 * @template ProvidedShape The provided shape.
 * @group DSL
 */
declare function lazy<ProvidedShape extends AnyShape>(shapeProvider: () => ProvidedShape): LazyShape<ProvidedShape, Input<ProvidedShape>>;

/**
 * Creates the {@link !Map Map} instance shape.
 *
 * @param keyShape The key shape.
 * @param valueShape The value shape.
 * @param options The issue options or the issue message.
 * @template KeyShape The key shape.
 * @template ValueShape The value shape.
 * @group DSL
 */
declare function map<KeyShape extends AnyShape, ValueShape extends AnyShape>(keyShape: KeyShape, valueShape: ValueShape, options?: IssueOptions | Message): MapShape<KeyShape, ValueShape>;

/**
 * Creates the shape that requires an input to be equal to `NaN`.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
declare function nan(options?: IssueOptions | Message): ConstShape<number>;

/**
 * Creates the shape that always raises an issue.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
declare function never(options?: IssueOptions | Message): NeverShape;

/**
 * Creates the shape that only allows values that don't conform the shape.
 *
 * @param shape The shape to which the output must not conform.
 * @param options The issue options or the issue message.
 * @template ExcludedShape The shape to which the output must not conform.
 * @group DSL
 */
declare function not<ExcludedShape extends AnyShape>(shape: ExcludedShape, options?: IssueOptions | Message): NotShape<Shape, ExcludedShape>;

/**
 * Creates the shape that requires an input to be equal to `null`.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
declare function null_(options?: IssueOptions | Message): ConstShape<null>;

/**
 * Creates the number shape.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
declare function number(options?: IssueOptions | Message): NumberShape;

/**
 * Creates the object shape.
 *
 * @param shapes The mapping from an object key to a corresponding shape.
 * @param options The issue options or the issue message.
 * @template PropShapes The mapping from a string object key to a corresponding value shape.
 * @group DSL
 */
declare function object<PropShapes extends ReadonlyDict<AnyShape>>(shapes: PropShapes, options?: IssueOptions | Message): ObjectShape<PropShapes, null>;

/**
 * Creates the {@link !Promise Promise} instance shape.
 *
 * @param options The issue options or the issue message.
 * @template ValueShape The shape of the resolved value.
 * @group DSL
 */
declare function promise(options?: IssueOptions | Message): PromiseShape<null>;
/**
 * Creates the {@link !Promise Promise} instance shape that validates the fulfillment value.
 *
 * @param shape The shape of the resolved value.
 * @param options The issue options or the issue message.
 * @template ValueShape The shape of the resolved value.
 * @group DSL
 */
declare function promise<ValueShape extends AnyShape>(shape: ValueShape, options?: IssueOptions | Message): PromiseShape<ValueShape>;

/**
 * Creates the shape that describes an object with string keys and values that conform the given shape.
 *
 * @param valueShape The shape of the record values.
 * @param options The issue options or the issue message.
 * @template ValueShape The value shape.
 * @group DSL
 */
declare function record<ValueShape extends AnyShape>(valueShape: ValueShape, options?: IssueOptions | Message): RecordShape<null, ValueShape>;
/**
 * Creates the shape that describes an object with string keys and values that conform the given shape.
 *
 * @param keyShape The shape of record keys.
 * @param valueShape The shape of the record values.
 * @param options The issue options or the issue message.
 * @template KeyShape The key shape.
 * @template ValueShape The value shape.
 * @group DSL
 */
declare function record<KeyShape extends Shape<string, PropertyKey>, ValueShape extends AnyShape>(keyShape: KeyShape, valueShape: ValueShape, options?: IssueOptions | Message): RecordShape<KeyShape, ValueShape>;

/**
 * Creates the {@link !Set Set} instance shape.
 *
 * @param shape The value shape
 * @param options The issue options or the issue message.
 * @template ValueShape The value shape.
 * @group DSL
 */
declare function set<ValueShape extends AnyShape>(shape: ValueShape, options?: IssueOptions | Message): SetShape<ValueShape>;

/**
 * Creates the string shape.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
declare function string(options?: IssueOptions | Message): StringShape;

/**
 * Creates the symbol shape.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
declare function symbol(options?: IssueOptions | Message): SymbolShape;

/**
 * Creates the shape that synchronously converts the input value.
 *
 * @param cb The callback that converts the input value. Throw a {@link ValidationError} to notify that the conversion
 * cannot be successfully completed.
 * @template Value The input value.
 * @template ConvertedValue The output value.
 * @group DSL
 */
declare function convert<Value = any, ConvertedValue = Value>(
/**
 * @param value The input value.
 * @param options Parsing options.
 */
cb: (value: any, options: ApplyOptions) => ConvertedValue): Shape<Value, ConvertedValue>;
/**
 * Creates the shape that asynchronously converts the input value.
 *
 * @param cb The callback that converts the input value asynchronously. The returned promise can be rejected with a
 * {@link ValidationError} to notify that the conversion cannot be successfully completed.
 * @template Value The input value.
 * @template ConvertedValue The output value.
 * @group DSL
 */
declare function convertAsync<Value = any, ConvertedValue = Value>(
/**
 * @param value The input value.
 * @param options Parsing options.
 */
cb: (value: any, options: ApplyOptions) => PromiseLike<ConvertedValue>): Shape<Value, ConvertedValue>;

/**
 * Creates the tuple shape.
 *
 * @param shapes The array of tuple element shapes.
 * @param options The issue options or the issue message.
 * @template HeadShapes The head tuple elements.
 * @group DSL
 */
declare function tuple<HeadShapes extends readonly [AnyShape, ...AnyShape[]]>(shapes: HeadShapes, options?: IssueOptions | Message): ArrayShape<HeadShapes, null>;
/**
 * Creates the tuple shape with rest elements.
 *
 * @param headShapes The array of tuple element shapes.
 * @param restShape The shape of rest elements.
 * @param options The issue options or the issue message.
 * @template HeadShapes The head tuple elements.
 * @template RestShape The rest tuple elements.
 * @group DSL
 */
declare function tuple<HeadShapes extends readonly [AnyShape, ...AnyShape[]], RestShape extends AnyShape | null>(headShapes: HeadShapes, restShape: RestShape, options?: IssueOptions | Message): ArrayShape<HeadShapes, RestShape>;

/**
 * Creates the shape that requires an input to be equal to `undefined`.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
declare function undefined_(options?: IssueOptions | Message): ConstShape<undefined>;

/**
 * Creates a union shape that tries to parse the input with one of the provided shapes.
 *
 * @param shapes The array of shapes to try.
 * @param options The issue options or the issue message.
 * @template Shapes The tuple of united shapes.
 * @group DSL
 */
declare function union<Shapes extends [AnyShape, ...AnyShape[]]>(shapes: Shapes, options?: IssueOptions | Message): UnionShape<Shapes>;

/**
 * Creates the unconstrained shape with unknown value.
 *
 * @group DSL
 */
declare function unknown(): Shape<unknown>;

/**
 * Creates a shape that requires an input to be `undefined` at runtime and typed as `void`.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
declare function void_(options?: IssueOptions | Message): ConstShape<void>;

/**
 * An error thrown if parsing failed. Custom check callbacks, refinement predicates, converters, and fallback
 * functions can throw this error to notify that the operation has failed.
 *
 * @group Other
 */
declare class ValidationError extends Error {
    /**
     * The array of issues that caused the error.
     */
    issues: Issue[];
    /**
     * Creates a new {@link ValidationError} instance.
     *
     * @param issues The array of issues that caused the validation error.
     * @param message The error message. If omitted then `issues` are converted to a string using
     * {@link ValidationError.formatIssues} and used as a message.
     */
    constructor(
    /**
     * The array of issues that caused the error.
     */
    issues: Issue[], message?: string);
    /**
     * The global function that stringifies issues as an error message, if a message is omitted when
     * {@link ValidationError} is instantiated.
     */
    static formatIssues: (issues: Issue[]) => string;
}

export { type AllowShape, type AlterCallback, type Any, type AnyShape, type ApplyOptions, ArrayShape, BigIntShape, BooleanShape, type Branded, CatchShape, type CheckCallback, CoercibleShape, ConstShape, ConvertShape, type CustomOperationOptions, DateShape, type DeepPartialProtocol, type DeepPartialShape, DenyShape, EnumShape, type Err, type ErrorMessageCallback, ExcludeShape, FunctionShape, type Input, InstanceShape, IntersectionShape, type Issue, type IssueOptions, LazyShape, MapShape, type Message, type MessageCallback, NEVER, NeverShape, type NotShape, NumberShape, type ObjectKeysMode, ObjectShape, type Ok, type Operation, type OperationCallback, type OperationOptions, type Output, type ParameterizedCustomOperationOptions, type ParameterizedRefineOptions, type ParseOptions, PipeShape, PromiseShape, RecordShape, type RefineCallback, type RefineOptions, type RefinePredicate, type RefineShape, ReplaceShape, type Result, SetShape, Shape, StringShape, SymbolShape, Type, type TypeName, UnionShape, ValidationError, intersection as and, any, array, bigint, boolean as bool, boolean, const_ as const, convert, convertAsync, date, enum_ as enum, function_ as fn, function_ as function, instanceOf, intersection, lazy, map, nan, never, not, null_ as null, number, object, union as or, promise, record, set, string, symbol, tuple, undefined_ as undefined, union, unknown, void_ as void };
