import { extractOptions, createIssueFactory } from './utils.mjs';

function __extends(constructor, superConstructor) {
    var Super = /** @class */ (function () {
        function Super() {
            this.constructor = constructor;
        }
        return Super;
    }());
    Super.prototype = superConstructor.prototype;
    Object.setPrototypeOf(constructor, superConstructor);
    constructor.prototype = new Super();
}

var ERR_SYNC_UNSUPPORTED = 'Shape is async, use tryAsync, parseAsync, or parseOrDefaultAsync.';
var ERR_SHAPE_EXPECTED = 'Provider must return a shape. Are you accessing a lazy shape prematurely?';
var ERR_ASYNC_FUNCTION = 'The function signature is constrained by async shapes, use ensureAsync.';
var CODE_ANY_DENY = 'any.deny';
var CODE_ANY_EXCLUDE = 'any.exclude';
var CODE_ANY_REFINE = 'any.refine';
var CODE_OBJECT_EXACT = 'object.exact';
var CODE_TYPE = 'type';
var CODE_TYPE_CONST = 'type.const';
var CODE_TYPE_ENUM = 'type.enum';
var CODE_TYPE_INSTANCE_OF = 'type.instanceOf';
var CODE_TYPE_INTERSECTION = 'type.intersection';
var CODE_TYPE_NEVER = 'type.never';
var CODE_TYPE_TUPLE = 'type.tuple';
var CODE_TYPE_UNION = 'type.union';
var MESSAGE_ANY_DENY = 'Must not be equal to %s';
var MESSAGE_ANY_EXCLUDE = 'Must not conform the excluded shape';
var MESSAGE_ANY_REFINE = 'Must conform the predicate';
var MESSAGE_OBJECT_EXACT = 'Must not have unknown keys: %s';
var MESSAGE_TYPE_ARRAY = 'Must be an array';
var MESSAGE_TYPE_BIGINT = 'Must be a bigint';
var MESSAGE_TYPE_BOOLEAN = 'Must be a boolean';
var MESSAGE_TYPE_CONST = 'Must be equal to %s';
var MESSAGE_TYPE_DATE = 'Must be a Date';
var MESSAGE_TYPE_ENUM = 'Must be equal to one of %s';
var MESSAGE_TYPE_FUNCTION = 'Must be a function';
var MESSAGE_TYPE_INSTANCE_OF = 'Must be a class instance';
var MESSAGE_TYPE_INTERSECTION = 'Intersection results are incompatible';
var MESSAGE_TYPE_MAP = 'Must be a Map';
var MESSAGE_TYPE_NEVER = 'Must not be used';
var MESSAGE_TYPE_NUMBER = 'Must be a number';
var MESSAGE_TYPE_OBJECT = 'Must be an object';
var MESSAGE_TYPE_PROMISE = 'Must be a Promise';
var MESSAGE_TYPE_TUPLE = 'Must be a tuple of length %s';
var MESSAGE_TYPE_SET = 'Must be a Set';
var MESSAGE_TYPE_STRING = 'Must be a string';
var MESSAGE_TYPE_SYMBOL = 'Must be a symbol';
var MESSAGE_TYPE_UNION = 'Must conform the union';

/**
 * Array predicate that filters unique elements.
 */
function unique(value, index, arr) {
    return !arr.includes(value, index + 1);
}
/**
 * Converts `k` to a number if it represents a valid array index, or returns -1 if `k` isn't an index.
 */
function toArrayIndex(k) {
    return (typeof k === 'number' || (typeof k === 'string' && k === '' + (k = +k))) && k >>> 0 === k ? k : -1;
}

/**
 * Toggles the bit in the bitmask at given position.
 *
 * @param bitmask The mutable mask to update.
 * @param position The index at which the bit must be toggled.
 * @returns The updated mask.
 */
function toggleBit(bitmask, position) {
    if (typeof bitmask === 'number') {
        if (position < 32) {
            return bitmask ^ (1 << position);
        }
        bitmask = [bitmask, 0, 0];
    }
    var bucketIndex = position >>> 5;
    bitmask[bucketIndex] ^= 1 << (position - (bucketIndex << 5));
    return bitmask;
}
/**
 * Returns the bit value at given position.
 */
function getBit(bitmask, position) {
    if (typeof bitmask === 'number') {
        return (bitmask >>> position) & 1;
    }
    var bucketIndex = position >>> 5;
    return (bitmask[bucketIndex] >>> (position - (bucketIndex << 5))) & 1;
}

var isArray = Array.isArray;
function identity(value) {
    return value;
}
/**
 * [SameValueZero](https://tc39.es/ecma262/multipage/abstract-operations.html#sec-samevaluezero) comparison.
 */
function isEqual(a, b) {
    return a === b || (a !== a && b !== b);
}
function isObjectLike(value) {
    return value !== null && typeof value === 'object';
}
function isObject(value) {
    return isObjectLike(value) && !isArray(value);
}
function isIterableObject(value) {
    return isObjectLike(value) && ((typeof Symbol !== 'undefined' && Symbol.iterator in value) || !isNaN(value.length));
}
function isEqualOrSubclass(ctor, superCtor) {
    return ctor === superCtor || superCtor.prototype.isPrototypeOf(ctor.prototype);
}
function isValidDate(value) {
    return value instanceof Date && (value = value.getTime()) === value;
}
function isMapEntry(value) {
    return isArray(value) && value.length === 2;
}
/**
 * Returns primitive if an object is a wrapper, or returns value as is.
 */
function getCanonicalValueOf(value) {
    if (isObjectLike(value) && (value instanceof String || value instanceof Number || value instanceof Boolean)) {
        return value.valueOf();
    }
    return value;
}
function returnTrue() {
    return true;
}

/**
 * Updates object property value, prevents prototype pollution.
 */
function setObjectProperty(obj, key, value) {
    if (key === '__proto__') {
        Object.defineProperty(obj, key, { value: value, writable: true, enumerable: true, configurable: true });
    }
    else {
        obj[key] = value;
    }
    return value;
}
/**
 * Clones a dictionary-like object.
 */
function cloneDict(dict) {
    var obj = {};
    for (var key in dict) {
        setObjectProperty(obj, key, dict[key]);
    }
    return obj;
}
/**
 * Clones the first `count` keys of a dictionary-like object.
 */
function cloneDictHead(dict, count) {
    var obj = {};
    var index = 0;
    for (var key in dict) {
        if (index >= count) {
            break;
        }
        setObjectProperty(obj, key, dict[key]);
        ++index;
    }
    return obj;
}
/**
 * Clones known keys of a dictionary-like object.
 */
function cloneDictKeys(dict, keys) {
    var obj = {};
    for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
        var key = keys_1[_i];
        if (key in dict) {
            setObjectProperty(obj, key, dict[key]);
        }
    }
    return obj;
}

/**
 * An error thrown if parsing failed. Custom check callbacks, refinement predicates, converters, and fallback
 * functions can throw this error to notify that the operation has failed.
 *
 * @group Other
 */
var ValidationError = /** @class */ (function (_super) {
    __extends(ValidationError, _super);
    /**
     * Creates a new {@link ValidationError} instance.
     *
     * @param issues The array of issues that caused the validation error.
     * @param message The error message. If omitted then `issues` are converted to a string using
     * {@link ValidationError.formatIssues} and used as a message.
     */
    function ValidationError(
    /**
     * The array of issues that caused the error.
     */
    issues, message) {
        if (message === void 0) { message = ValidationError.formatIssues(issues); }
        var _newTarget = this.constructor;
        var _this = _super.call(this, message) || this;
        _this.issues = issues;
        Object.setPrototypeOf(_this, _newTarget.prototype);
        _this.name = 'ValidationError';
        return _this;
    }
    /**
     * The global function that stringifies issues as an error message, if a message is omitted when
     * {@link ValidationError} is instantiated.
     */
    ValidationError.formatIssues = function (issues) { return JSON.stringify(issues, stringifyES6Values, 2); };
    return ValidationError;
}(Error));
function stringifyES6Values(key, value) {
    return typeof value === 'symbol' || typeof value === 'bigint' ? String(value) : value;
}

var defaultApplyOptions = Object.freeze({ earlyReturn: false, coerce: false });
var nonce = -1;
function nextNonce() {
    return ++nonce;
}
function ok(value) {
    return { ok: true, value: value };
}
var universalApplyOperations = function (input, output, options, issues) {
    if (issues !== null) {
        return issues;
    }
    if (isEqual(input, output)) {
        return null;
    }
    return ok(output);
};
function isAsyncShapes(shapes) {
    if (shapes !== null && shapes !== undefined) {
        for (var _i = 0, shapes_1 = shapes; _i < shapes_1.length; _i++) {
            var shape = shapes_1[_i];
            if (shape.isAsync) {
                return true;
            }
        }
    }
    return false;
}
/**
 * Converts the shape to its deep partial alternative if shape implements {@link DeepPartialProtocol}, or returns
 * the shape as is.
 */
function toDeepPartialShape(shape) {
    return typeof shape.deepPartial === 'function' ? shape.deepPartial() : shape;
}
/**
 * Copies checks from `baseShape` to `shape`.
 */
function copyOperations(baseShape, shape) {
    shape.operations = baseShape.operations;
    return shape;
}
/**
 * Calls `Shape._apply` or `Shape._applyAsync` depending on `Shape._isAsync`, and passes the result to `cb` after it
 * becomes available.
 */
function applyShape(shape, input, options, nonce, cb) {
    if (shape.isAsync) {
        return shape['_applyAsync'](input, options, nonce).then(cb);
    }
    else {
        return cb(shape['_apply'](input, options, nonce));
    }
}
/**
 * Prepends a key to a path of each issue.
 */
function unshiftIssuesPath(issues, key) {
    for (var _i = 0, issues_1 = issues; _i < issues_1.length; _i++) {
        var issue = issues_1[_i];
        if (isArray(issue.path)) {
            issue.path.unshift(key);
        }
        else {
            issue.path = [key];
        }
    }
}
function concatIssues(issues, otherIssues) {
    if (issues === null) {
        return otherIssues;
    }
    issues.push.apply(issues, otherIssues);
    return issues;
}
function captureIssues(error) {
    if (error instanceof ValidationError) {
        return error.issues;
    }
    throw error;
}
/**
 * Returns an error message that is composed of the captured issues and parsing options.
 */
function getMessage(issues, input, options) {
    var message = options === null || options === void 0 ? void 0 : options.errorMessage;
    if (typeof message === 'function') {
        return message(issues, input);
    }
    if (message !== undefined) {
        return String(message);
    }
}

/**
 * The enum-like class that describes a value type.
 *
 * @template T The name of the value type.
 * @see {@link Shape.inputs}
 * @group Type Inference
 */
var Type = /** @class */ (function () {
    function Type(
    /**
     * The name of the type.
     */
    name) {
        this.name = name;
        Object.freeze(this);
    }
    /**
     * Returns the type of the given value. If value is a type itself, it is returned as is.
     */
    Type.of = function (value) {
        return getTypeOf(value);
    };
    /**
     * @internal
     */
    Type.prototype.toString = function () {
        return this.name;
    };
    Type.ARRAY = new Type('array');
    Type.BIGINT = new Type('bigint');
    Type.BOOLEAN = new Type('boolean');
    Type.DATE = new Type('date');
    Type.FUNCTION = new Type('function');
    Type.MAP = new Type('map');
    Type.NULL = new Type('null');
    Type.NUMBER = new Type('number');
    Type.OBJECT = new Type('object');
    Type.PROMISE = new Type('promise');
    Type.SET = new Type('set');
    Type.STRING = new Type('string');
    Type.SYMBOL = new Type('symbol');
    Type.UNDEFINED = new Type('undefined');
    Type.UNKNOWN = new Type('unknown');
    return Type;
}());
Object.freeze(Type);
var TYPE_ARRAY = Type.ARRAY;
var TYPE_BIGINT = Type.BIGINT;
var TYPE_BOOLEAN = Type.BOOLEAN;
var TYPE_DATE = Type.DATE;
var TYPE_FUNCTION = Type.FUNCTION;
var TYPE_MAP = Type.MAP;
var TYPE_NULL = Type.NULL;
var TYPE_NUMBER = Type.NUMBER;
var TYPE_OBJECT = Type.OBJECT;
var TYPE_PROMISE = Type.PROMISE;
var TYPE_SET = Type.SET;
var TYPE_STRING = Type.STRING;
var TYPE_SYMBOL = Type.SYMBOL;
var TYPE_UNDEFINED = Type.UNDEFINED;
var TYPE_UNKNOWN = Type.UNKNOWN;
function getTypeOf(value) {
    var type = typeof value;
    if (type === 'undefined') {
        return TYPE_UNDEFINED;
    }
    if (type === 'boolean') {
        return TYPE_BOOLEAN;
    }
    if (type === 'number') {
        return TYPE_NUMBER;
    }
    if (type === 'string') {
        return TYPE_STRING;
    }
    if (type === 'function') {
        return TYPE_FUNCTION;
    }
    if (type === 'symbol') {
        return TYPE_SYMBOL;
    }
    if (type === 'bigint') {
        return TYPE_BIGINT;
    }
    if (value === null) {
        return TYPE_NULL;
    }
    if (isArray(value)) {
        return TYPE_ARRAY;
    }
    if (value instanceof Type) {
        return value;
    }
    if (value instanceof Date) {
        return TYPE_DATE;
    }
    if (value instanceof Promise) {
        return TYPE_PROMISE;
    }
    if (value instanceof Set) {
        return TYPE_SET;
    }
    if (value instanceof Map) {
        return TYPE_MAP;
    }
    return TYPE_OBJECT;
}

function isType(value) {
    return value instanceof Type;
}
/**
 * Returns `true` if type or literal `a` is assignable to the type or literal `b`.
 */
function isAssignable(a, b) {
    return b === TYPE_UNKNOWN || isEqual(a, b) || getTypeOf(a) === b;
}
/**
 * Returns an array of unique types and literals that comprise a union.
 */
function unionTypes(types) {
    var t = types;
    next: for (var i = 0; i < t.length; ++i) {
        var ti = t[i];
        for (var j = i + 1; j < t.length; ++j) {
            var tj = t[j];
            if (isAssignable(ti, tj)) {
                if (t === types) {
                    t = types.slice(0);
                }
                t.splice(i--, 1);
                continue next;
            }
            if (isAssignable(tj, ti)) {
                if (t === types) {
                    t = types.slice(0);
                }
                t.splice(j--, 1);
            }
        }
    }
    return t;
}
/**
 * Takes an array of arrays of types that are treated as an intersection of unions, and applies the distribution rule
 * that produces a union.
 *
 * ```
 * (a | b) & (a | b | c) → a | b
 * ```
 */
function distributeTypes(types) {
    if (types.length === 0) {
        return [];
    }
    var t = [];
    var t0 = types[0];
    for (var i = 0; i < t0.length; ++i) {
        var t0i = t0[i];
        for (var k = 1; k < types.length; ++k) {
            var tk = types[k];
            for (var j = 0; j < tk.length; ++j) {
                var tkj = tk[j];
                if (isAssignable(tkj, t0i)) {
                    t.push(tkj);
                }
                else if (isAssignable(t0i, tkj)) {
                    t.push(t0i);
                }
            }
        }
    }
    return t;
}

/**
 * The marker object that is used to denote an impossible value.
 *
 * For example, {@link NEVER} is returned from {@link CoercibleShape._coerce} method when coercion is not possible.
 *
 * @group Other
 */
var NEVER = Object.freeze({ never: true });
/**
 * The baseline shape implementation.
 *
 * @template InputValue The input value.
 * @template OutputValue The output value.
 * @group Shapes
 */
var Shape = /** @class */ (function () {
    function Shape() {
        /**
         * The dictionary of shape annotations.
         *
         * @see {@link Shape.annotate}
         */
        this.annotations = {};
        /**
         * The array of operations that are applied to the shape output.
         *
         * @see {@link Shape.use}
         * @see [Operations](https://github.com/smikhalevski/doubter#operations)
         */
        this.operations = [];
    }
    /**
     * Returns a sub-shape that describes a value associated with the given property name, or `null` if there's no such
     * sub-shape.
     *
     * @param key The key for which the sub-shape must be retrieved.
     * @returns The sub-shape or `null` if there's no such key in the shape.
     */
    Shape.prototype.at = function (key) {
        return null;
    };
    /**
     * Returns `true` if the shape accepts given input type or value, or `false` otherwise.
     *
     * @param input The type or value that must be checked.
     */
    Shape.prototype.accepts = function (input) {
        var inputs = this.inputs;
        return inputs.includes(TYPE_UNKNOWN) || inputs.includes(input) || inputs.includes(getTypeOf(input));
    };
    /**
     * Assigns annotations to the shape.
     *
     * @param annotations Annotations to add.
     * @returns The clone of the shape with the updated annotations.
     * @see {@link Shape.annotations}
     */
    Shape.prototype.annotate = function (annotations) {
        var shape = this._clone();
        shape.annotations = Object.assign({}, this.annotations, annotations);
        return shape;
    };
    /**
     * Appends an operation to the shape.
     *
     * @param cb The factory that produces the operation callback.
     * @param options The operation operations.
     * @returns The clone of the shape.
     * @see [Operations](https://github.com/smikhalevski/doubter#operations)
     */
    Shape.prototype.use = function (
    /**
     * Creates an {@link OperationCallback} that applies the logic of the operation to the shape output and passes the
     * control to the next operation.
     *
     * @param next The callback that applies the next operation.
     * @returns The callback that applies an operation to the shape output.
     */
    cb, options) {
        if (options === void 0) { options = {}; }
        var _a = options.type, type = _a === void 0 ? cb : _a, param = options.param;
        var shape = this._clone();
        shape.operations = this.operations.concat({ type: type, param: param, factory: cb });
        return shape;
    };
    Shape.prototype.check = function (cb, options) {
        if (options === void 0) { options = {}; }
        var _a = options.type, type = _a === void 0 ? cb : _a, param = options.param, _b = options.force, force = _b === void 0 ? false : _b;
        return this.use(function (next) { return function (input, output, options, issues) {
            if (issues === null || force) {
                var result = void 0;
                try {
                    result = cb(output, param, options);
                }
                catch (error) {
                    issues = concatIssues(issues, captureIssues(error));
                    if (options.earlyReturn) {
                        return issues;
                    }
                }
                if (isObjectLike(result) &&
                    (isArray(result)
                        ? result.length !== 0 && (issues = concatIssues(issues, result)) !== null
                        : (issues || (issues = [])).push(result) !== 0) &&
                    options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: type, param: param });
    };
    Shape.prototype.refine = function (cb, options) {
        var _a = extractOptions(options), _b = _a.type, type = _b === void 0 ? cb : _b, param = _a.param, _c = _a.force, force = _c === void 0 ? false : _c, _d = _a.code, code = _d === void 0 ? CODE_ANY_REFINE : _d;
        var issueFactory = createIssueFactory(code, MESSAGE_ANY_REFINE, options, cb);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (issues === null || force) {
                var result = true;
                try {
                    result = cb(output, param, options);
                }
                catch (error) {
                    issues = concatIssues(issues, captureIssues(error));
                    if (options.earlyReturn) {
                        return issues;
                    }
                }
                if (!result) {
                    (issues || (issues = [])).push(issueFactory(output, options));
                    if (options.earlyReturn) {
                        return issues;
                    }
                }
            }
            return next(input, output, options, issues);
        }; }, { type: type, param: param });
    };
    Shape.prototype.alter = function (cb, options) {
        if (options === void 0) { options = {}; }
        var _a = options.type, type = _a === void 0 ? cb : _a, param = options.param, _b = options.force, force = _b === void 0 ? false : _b;
        return this.use(function (next) { return function (input, output, options, issues) {
            if (issues === null || force) {
                try {
                    output = cb(output, param, options);
                }
                catch (error) {
                    issues = concatIssues(issues, captureIssues(error));
                    if (options.earlyReturn) {
                        return issues;
                    }
                }
            }
            return next(input, output, options, issues);
        }; }, { type: type, param: param });
    };
    /**
     * Synchronously converts the output value of the shape.
     *
     * @param cb The callback that converts the input value. Throw a {@link ValidationError} to notify that the conversion
     * cannot be successfully completed.
     * @returns The {@link ConvertShape} instance.
     * @template ConvertedValue The value returned from the callback that converts the output value of this shape.
     * @see {@link Shape.alter}
     */
    Shape.prototype.convert = function (
    /**
     * @param value The shape output value.
     * @param options Parsing options.
     */
    cb) {
        return this.to(new ConvertShape(cb));
    };
    /**
     * Asynchronously converts the output value of the shape.
     *
     * @param cb The callback that converts the input value asynchronously. The returned promise can be rejected with a
     * {@link ValidationError} to notify that the conversion cannot be successfully completed.
     * @returns The {@link ConvertShape} instance.
     * @template ConvertedValue The value returned from the callback that converts the output value of this shape.
     * @see {@link Shape.alter}
     */
    Shape.prototype.convertAsync = function (
    /**
     * @param value The shape output value.
     * @param options Parsing options.
     */
    cb) {
        return this.to(new ConvertShape(cb, true));
    };
    /**
     * Pipes the output of this shape to the input of another shape.
     *
     * @param shape The shape that validates the output if this shape.
     * @template OutputShape The output value.
     */
    Shape.prototype.to = function (shape) {
        return new PipeShape(this, shape);
    };
    /**
     * Returns a shape that adds a brand to the output type.
     *
     * @returns A shape with the branded output type.
     * @template Brand The brand value.
     */
    Shape.prototype.brand = function () {
        return this;
    };
    /**
     * Replaces an input value with an output value.
     *
     * @param inputValue The input value to replace.
     * @param outputValue The output value that is returned if an `inputValue` is received.
     * @template InputValue The input value to replace.
     * @template OutputValue The output value that is used as the replacement for an input value.
     */
    Shape.prototype.replace = function (inputValue, outputValue) {
        return new ReplaceShape(this, inputValue, outputValue);
    };
    /**
     * Allows an input value, so it is passed directly to the output.
     *
     * @param value The allowed value.
     * @template AllowedValue The allowed value.
     */
    Shape.prototype.allow = function (value) {
        return this.replace(value, value);
    };
    /**
     * Excludes value from both input and output.
     *
     * @param value The excluded value.
     * @param options The issue options or the issue message.
     * @template DeniedValue The denied value.
     */
    Shape.prototype.deny = function (value, options) {
        return new DenyShape(this, value, options);
    };
    Shape.prototype.optional = function (defaultValue) {
        return this.replace(undefined, defaultValue);
    };
    Shape.prototype.nullable = function (defaultValue) {
        return this.replace(null, arguments.length === 0 ? null : defaultValue);
    };
    Shape.prototype.nullish = function (defaultValue) {
        return this.nullable(arguments.length === 0 ? null : defaultValue).replace(undefined, defaultValue);
    };
    /**
     * Prevents an input and output from being `undefined`.
     *
     * @param options The issue options or the issue message.
     */
    Shape.prototype.nonOptional = function (options) {
        return new DenyShape(this, undefined, options);
    };
    Shape.prototype.catch = function (fallback) {
        return new CatchShape(this, fallback);
    };
    /**
     * Checks that the input doesn't match the shape.
     *
     * @param shape The shape to which the output must not conform.
     * @param options The issue options or the issue message.
     * @template ExcludedShape The shape to which the output must not conform.
     */
    Shape.prototype.exclude = function (shape, options) {
        return new ExcludeShape(this, shape, options);
    };
    /**
     * Checks that the input doesn't match the shape.
     *
     * This method works exactly as {@link Shape.exclude} at runtime, but it doesn't perform the exclusion on the type
     * level.
     *
     * @param shape The shape to which the output must not conform.
     * @param options The issue options or the issue message.
     * @template ExcludedShape The shape to which the output must not conform.
     */
    Shape.prototype.not = function (shape, options) {
        return this.exclude(shape, options);
    };
    /**
     * Must return `true` if the shape must be used in async context only, otherwise the shape can be used in both sync
     * and async contexts. Override this method to implement a custom shape.
     *
     * @see [Advanced shapes](https://github.com/smikhalevski/doubter#advanced-shapes)
     */
    Shape.prototype._isAsync = function () {
        return false;
    };
    /**
     * Returns input types and literal values that this shape can accept as an input.
     *
     * @see [Advanced shapes](https://github.com/smikhalevski/doubter#advanced-shapes)
     */
    Shape.prototype._getInputs = function () {
        return [TYPE_UNKNOWN];
    };
    /**
     * Clones the shape.
     *
     * @see [Advanced shapes](https://github.com/smikhalevski/doubter#advanced-shapes)
     */
    Shape.prototype._clone = function () {
        return Object.assign(Object.create(Object.getPrototypeOf(this)), this);
    };
    /**
     * Synchronously parses the input.
     *
     * **Note:** Don't store or update returned instances of {@link Ok} since they can be reused.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @param nonce The globally unique number that identifies the parsing process.
     * @returns `null` if input matches the output, {@link Ok} that wraps the output, or an array of captured issues.
     * @see [Advanced shapes](https://github.com/smikhalevski/doubter#advanced-shapes)
     */
    Shape.prototype._apply = function (input, options, nonce) {
        return this._applyOperations(input, input, options, null);
    };
    /**
     * Asynchronously parses the input.
     *
     * **Note:** Don't store or update returned instances of {@link Ok} since they can be reused.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @param nonce The globally unique number that identifies the parsing process.
     * @returns `null` if input matches the output, {@link Ok} that wraps the output, or an array of captured issues.
     * @see [Advanced shapes](https://github.com/smikhalevski/doubter#advanced-shapes)
     */
    Shape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        return new Promise(function (resolve) {
            resolve(_this._apply(input, options, nonce));
        });
    };
    return Shape;
}());
Object.defineProperties(Shape.prototype, {
    inputs: {
        configurable: true,
        get: function () {
            Object.defineProperty(this, 'inputs', { configurable: true, value: [] });
            var inputs = Object.freeze(unionTypes(this._getInputs()));
            Object.defineProperty(this, 'inputs', { configurable: true, value: inputs });
            return inputs;
        },
    },
    isAsync: {
        configurable: true,
        get: function () {
            Object.defineProperty(this, 'isAsync', { configurable: true, value: false });
            var async = this._isAsync();
            var universalApplyAsync = Shape.prototype._applyAsync;
            if (async) {
                this._apply = function () {
                    throw new Error(ERR_SYNC_UNSUPPORTED);
                };
            }
            else if (this._applyAsync !== universalApplyAsync) {
                this._applyAsync = universalApplyAsync;
            }
            Object.defineProperty(this, 'isAsync', { configurable: true, value: async });
            return async;
        },
    },
    try: {
        configurable: true,
        get: function () {
            var _this = this;
            this.isAsync;
            var cb = function (input, options) {
                var result = _this._apply(input, options || defaultApplyOptions, nextNonce());
                if (result === null) {
                    return { ok: true, value: input };
                }
                if (isArray(result)) {
                    return { ok: false, issues: result };
                }
                return { ok: true, value: result.value };
            };
            Object.defineProperty(this, 'try', { writable: true, value: cb });
            return cb;
        },
    },
    tryAsync: {
        configurable: true,
        get: function () {
            var _this = this;
            this.isAsync;
            var cb = function (input, options) {
                return _this._applyAsync(input, options || defaultApplyOptions, nextNonce()).then(function (result) {
                    if (result === null) {
                        return { ok: true, value: input };
                    }
                    if (isArray(result)) {
                        return { ok: false, issues: result };
                    }
                    return { ok: true, value: result.value };
                });
            };
            Object.defineProperty(this, 'tryAsync', { writable: true, value: cb });
            return cb;
        },
    },
    parse: {
        configurable: true,
        get: function () {
            var _this = this;
            this.isAsync;
            var cb = function (input, options) {
                var result = _this._apply(input, options || defaultApplyOptions, nextNonce());
                if (result === null) {
                    return input;
                }
                if (isArray(result)) {
                    throw new ValidationError(result, getMessage(result, input, options));
                }
                return result.value;
            };
            Object.defineProperty(this, 'parse', { writable: true, value: cb });
            return cb;
        },
    },
    parseAsync: {
        configurable: true,
        get: function () {
            var _this = this;
            this.isAsync;
            var cb = function (input, options) {
                return _this._applyAsync(input, options || defaultApplyOptions, nextNonce()).then(function (result) {
                    if (result === null) {
                        return input;
                    }
                    if (isArray(result)) {
                        throw new ValidationError(result, getMessage(result, input, options));
                    }
                    return result.value;
                });
            };
            Object.defineProperty(this, 'parseAsync', { writable: true, value: cb });
            return cb;
        },
    },
    parseOrDefault: {
        configurable: true,
        get: function () {
            var _this = this;
            this.isAsync;
            var cb = function (input, defaultValue, options) {
                var result = _this._apply(input, options || defaultApplyOptions, nextNonce());
                if (result === null) {
                    return input;
                }
                if (isArray(result)) {
                    return defaultValue;
                }
                return result.value;
            };
            Object.defineProperty(this, 'parseOrDefault', { writable: true, value: cb });
            return cb;
        },
    },
    parseOrDefaultAsync: {
        configurable: true,
        get: function () {
            var _this = this;
            this.isAsync;
            var cb = function (input, defaultValue, options) {
                return _this._applyAsync(input, options || defaultApplyOptions, nextNonce()).then(function (result) {
                    if (result === null) {
                        return input;
                    }
                    if (isArray(result)) {
                        return defaultValue;
                    }
                    return result.value;
                });
            };
            Object.defineProperty(this, 'parseOrDefaultAsync', { writable: true, value: cb });
            return cb;
        },
    },
    _applyOperations: {
        configurable: true,
        get: function () {
            var cb = universalApplyOperations;
            for (var i = this.operations.length - 1; i >= 0; --i) {
                cb = this.operations[i].factory(cb);
            }
            Object.defineProperty(this, '_applyOperations', { writable: true, value: cb });
            return cb;
        },
    },
});
/**
 * The shape that applies a converter to the input.
 *
 * @template ConvertedValue The output value of the callback that converts the input value.
 * @group Shapes
 */
var ConvertShape = /** @class */ (function (_super) {
    __extends(ConvertShape, _super);
    /**
     * Creates the new {@link ConvertShape} instance.
     *
     * @param converter The callback that converts the input value. Converters can throw or reject with a
     * {@link ValidationError} to notify that the conversion cannot be successfully completed.
     * @param async If `true` then the convert shape waits for the promise returned from the callback to be fulfilled.
     * Otherwise, the value that is synchronously returned from the callback is used as an output.
     * @template ConvertedValue The output value of the callback that converts the input value.
     */
    function ConvertShape(
    /**
     * The converter that is applied to the shape output value.
     *
     * @param value The input value.
     * @param options Parsing options.
     */
    converter, async) {
        var _this = _super.call(this) || this;
        _this.converter = converter;
        if (async) {
            _this._isAsync = returnTrue;
        }
        return _this;
    }
    ConvertShape.prototype._apply = function (input, options, nonce) {
        var output;
        try {
            output = this.converter(input, options);
        }
        catch (error) {
            return captureIssues(error);
        }
        return this._applyOperations(input, output, options, null);
    };
    ConvertShape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        return new Promise(function (resolve) {
            resolve(_this.converter(input, options));
        }).then(function (output) { return _this._applyOperations(input, output, options, null); }, captureIssues);
    };
    return ConvertShape;
}(Shape));
/**
 * The shape that parses the output of the input shape with the output shape.
 *
 * @template InputShape The input shape.
 * @template OutputShape The output shape.
 * @group Shapes
 */
var PipeShape = /** @class */ (function (_super) {
    __extends(PipeShape, _super);
    /**
     * Creates the new {@link PipeShape} instance.
     *
     * @param inputShape The shape that parses the input value.
     * @param outputShape The shape that parses the output of `inputShape`.
     * @template InputShape The input shape.
     * @template OutputShape The output shape.
     */
    function PipeShape(
    /**
     * The input shape.
     */
    inputShape, 
    /**
     * The output shape.
     */
    outputShape) {
        var _this = _super.call(this) || this;
        _this.inputShape = inputShape;
        _this.outputShape = outputShape;
        return _this;
    }
    PipeShape.prototype.deepPartial = function () {
        return new PipeShape(toDeepPartialShape(this.inputShape), toDeepPartialShape(this.outputShape));
    };
    PipeShape.prototype._isAsync = function () {
        return this.inputShape.isAsync || this.outputShape.isAsync;
    };
    PipeShape.prototype._getInputs = function () {
        return this.inputShape.inputs.slice(0);
    };
    PipeShape.prototype._apply = function (input, options, nonce) {
        var _a = this, inputShape = _a.inputShape, outputShape = _a.outputShape;
        var output = input;
        var inputResult = inputShape['_apply'](input, options, nonce);
        if (inputResult !== null) {
            if (isArray(inputResult)) {
                return inputResult;
            }
            output = inputResult.value;
        }
        var outputResult = outputShape['_apply'](output, options, nonce);
        if (outputResult !== null) {
            if (isArray(outputResult)) {
                return outputResult;
            }
            output = outputResult.value;
        }
        return this._applyOperations(input, output, options, null);
    };
    PipeShape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        return this.inputShape['_applyAsync'](input, options, nonce).then(function (inputResult) {
            var output = input;
            if (inputResult !== null) {
                if (isArray(inputResult)) {
                    return inputResult;
                }
                output = inputResult.value;
            }
            return applyShape(_this.outputShape, output, options, nonce, function (outputResult) {
                if (outputResult !== null) {
                    if (isArray(outputResult)) {
                        return outputResult;
                    }
                    output = outputResult.value;
                }
                return _this._applyOperations(input, output, options, null);
            });
        });
    };
    return PipeShape;
}(Shape));
/**
 * The shape that replaces an input value with an output value.
 *
 * @template BaseShape The shape that parses the input without the replaced value.
 * @template InputValue The input value to replace.
 * @template OutputValue The output value that is used as the replacement for an input value.
 * @group Shapes
 */
var ReplaceShape = /** @class */ (function (_super) {
    __extends(ReplaceShape, _super);
    /**
     * Creates the new {@link ReplaceShape} instance.
     *
     * @param baseShape The shape that parses the input without the replaced value.
     * @param inputValue The input value to replace.
     * @param outputValue The output value that is returned if an `inputValue` is received.
     * @template BaseShape The shape that parses the input without the replaced value.
     * @template InputValue The input value to replace.
     * @template OutputValue The output value that is used as the replacement for an input value.
     */
    function ReplaceShape(
    /**
     * The shape that parses the input without the replaced value.
     */
    baseShape, 
    /**
     * The input value to replace.
     */
    inputValue, 
    /**
     * The output value that is returned if an {@link ReplaceShape.inputValue} is received.
     */
    outputValue) {
        var _this = _super.call(this) || this;
        _this.baseShape = baseShape;
        _this.inputValue = inputValue;
        _this.outputValue = outputValue;
        _this._result = isEqual(inputValue, outputValue) ? null : { ok: true, value: outputValue };
        return _this;
    }
    ReplaceShape.prototype.deepPartial = function () {
        return new ReplaceShape(toDeepPartialShape(this.baseShape), this.inputValue, this.outputValue);
    };
    ReplaceShape.prototype._isAsync = function () {
        return this.baseShape.isAsync;
    };
    ReplaceShape.prototype._getInputs = function () {
        return this.baseShape.inputs.concat(this.inputValue);
    };
    ReplaceShape.prototype._apply = function (input, options, nonce) {
        var result = isEqual(input, this.inputValue) ? this._result : this.baseShape['_apply'](input, options, nonce);
        return this._handleResult(result, input, options);
    };
    ReplaceShape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        if (isEqual(input, this.inputValue)) {
            return Promise.resolve(this._handleResult(this._result, input, options));
        }
        return this.baseShape['_applyAsync'](input, options, nonce).then(function (result) {
            return _this._handleResult(result, input, options);
        });
    };
    ReplaceShape.prototype._handleResult = function (result, input, options) {
        var output = input;
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
        }
        return this._applyOperations(input, output, options, null);
    };
    return ReplaceShape;
}(Shape));
/**
 * The shape that prevents both input and output from being equal to a denied value.
 *
 * @template BaseShape The shape that parses the input without the denied value.
 * @template DeniedValue The denied value.
 * @group Shapes
 */
var DenyShape = /** @class */ (function (_super) {
    __extends(DenyShape, _super);
    /**
     * Creates the new {@link DenyShape} instance.
     *
     * @param baseShape The shape that parses the input without the denied value.
     * @param deniedValue The dined value.
     * @param options The issue options or the issue message.
     * @template BaseShape The shape that parses the input without the denied value.
     * @template DeniedValue The denied value.
     */
    function DenyShape(
    /**
     * The shape that parses the input without the denied value.
     */
    baseShape, 
    /**
     * The dined value.
     */
    deniedValue, options) {
        var _this = _super.call(this) || this;
        _this.baseShape = baseShape;
        _this.deniedValue = deniedValue;
        _this._options = options;
        _this._typeIssueFactory = createIssueFactory(CODE_ANY_DENY, MESSAGE_ANY_DENY, options, deniedValue);
        return _this;
    }
    DenyShape.prototype.deepPartial = function () {
        return new DenyShape(toDeepPartialShape(this.baseShape), this.deniedValue, this._options);
    };
    DenyShape.prototype._isAsync = function () {
        return this.baseShape.isAsync;
    };
    DenyShape.prototype._getInputs = function () {
        var _this = this;
        return this.baseShape.inputs.filter(function (input) { return !isEqual(_this.deniedValue, input); });
    };
    DenyShape.prototype._apply = function (input, options, nonce) {
        if (isEqual(input, this.deniedValue)) {
            return [this._typeIssueFactory(input, options)];
        }
        return this._handleResult(this.baseShape['_apply'](input, options, nonce), input, options);
    };
    DenyShape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        if (isEqual(input, this.deniedValue)) {
            return Promise.resolve([this._typeIssueFactory(input, options)]);
        }
        return this.baseShape['_applyAsync'](input, options, nonce).then(function (result) {
            return _this._handleResult(result, input, options);
        });
    };
    DenyShape.prototype._handleResult = function (result, input, options) {
        var output = input;
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
            if (isEqual(output, this.deniedValue)) {
                return [this._typeIssueFactory(input, options)];
            }
        }
        return this._applyOperations(input, output, options, null);
    };
    return DenyShape;
}(Shape));
/**
 * The shape that returns the fallback value if parsing fails.
 *
 * @template BaseShape The shape that parses the input.
 * @template FallbackValue The fallback value.
 * @group Shapes
 */
var CatchShape = /** @class */ (function (_super) {
    __extends(CatchShape, _super);
    /**
     * Creates the new {@link CatchShape} instance.
     *
     * @param baseShape The shape that parses the input.
     * @param fallback The value or a callback that returns a value that is returned if parsing has failed.
     * @template BaseShape The shape that parses the input.
     * @template FallbackValue The fallback value.
     */
    function CatchShape(
    /**
     * The shape that parses the input.
     */
    baseShape, 
    /**
     * The value or a callback that returns a value that is returned if parsing has failed.
     *
     *  A callback receives an input value, an array of raised issues, and {@link ApplyOptions parsing options}.
     */
    fallback) {
        var _this = _super.call(this) || this;
        _this.baseShape = baseShape;
        _this.fallback = fallback;
        if (typeof fallback === 'function') {
            _this._resultProvider = function (input, issues, options) { return ok(fallback(input, issues, options)); };
        }
        else {
            var result_1 = { ok: true, value: fallback };
            _this._resultProvider = function () { return result_1; };
        }
        return _this;
    }
    CatchShape.prototype.deepPartial = function () {
        return new CatchShape(toDeepPartialShape(this.baseShape), this.fallback);
    };
    CatchShape.prototype._isAsync = function () {
        return this.baseShape.isAsync;
    };
    CatchShape.prototype._getInputs = function () {
        return this.baseShape.inputs.slice(0);
    };
    CatchShape.prototype._apply = function (input, options, nonce) {
        return this._handleResult(this.baseShape['_apply'](input, options, nonce), input, options);
    };
    CatchShape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        return this.baseShape['_applyAsync'](input, options, nonce).then(function (result) {
            return _this._handleResult(result, input, options);
        });
    };
    CatchShape.prototype._handleResult = function (result, input, options) {
        var output = input;
        if (result !== null) {
            if (isArray(result)) {
                try {
                    result = this._resultProvider(input, result, options);
                }
                catch (error) {
                    return captureIssues(error);
                }
            }
            output = result.value;
        }
        return this._applyOperations(input, output, options, null);
    };
    return CatchShape;
}(Shape));
/**
 * Checks that the input doesn't match the shape.
 *
 * @template BaseShape The base shape.
 * @template ExcludedShape The shape to which the output must not conform.
 * @group Shapes
 */
var ExcludeShape = /** @class */ (function (_super) {
    __extends(ExcludeShape, _super);
    /**
     * Creates the new {@link ExcludeShape} instance.
     *
     * @param baseShape The shape that parses the input.
     * @param excludedShape The shape to which the output must not conform.
     * @param options The issue options or the issue message.
     * @template BaseShape The base shape.
     * @template ExcludedShape The shape to which the output must not conform.
     */
    function ExcludeShape(
    /**
     * The base shape.
     */
    baseShape, 
    /**
     * The shape to which the output must not conform.
     */
    excludedShape, options) {
        var _this = _super.call(this) || this;
        _this.baseShape = baseShape;
        _this.excludedShape = excludedShape;
        _this._options = options;
        _this._typeIssueFactory = createIssueFactory(CODE_ANY_EXCLUDE, MESSAGE_ANY_EXCLUDE, options, excludedShape);
        return _this;
    }
    ExcludeShape.prototype.deepPartial = function () {
        return new ExcludeShape(toDeepPartialShape(this.baseShape), this.excludedShape, this._options);
    };
    ExcludeShape.prototype._isAsync = function () {
        return this.baseShape.isAsync || this.excludedShape.isAsync;
    };
    ExcludeShape.prototype._getInputs = function () {
        var _this = this;
        return this.baseShape.inputs.filter(function (input) { return isType(input) || !_this.excludedShape.inputs.includes(input); });
    };
    ExcludeShape.prototype._apply = function (input, options, nonce) {
        var _a = this, baseShape = _a.baseShape, excludedShape = _a.excludedShape;
        var output = input;
        var result = baseShape['_apply'](input, options, nonce);
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
        }
        if (!isArray(excludedShape['_apply'](output, options, nonce))) {
            return [this._typeIssueFactory(input, options)];
        }
        return this._applyOperations(input, output, options, null);
    };
    ExcludeShape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        var _a = this, baseShape = _a.baseShape, excludedShape = _a.excludedShape;
        return baseShape['_applyAsync'](input, options, nonce).then(function (result) {
            var output = input;
            if (result !== null) {
                if (isArray(result)) {
                    return result;
                }
                output = result.value;
            }
            return applyShape(excludedShape, output, options, nonce, function (outputResult) {
                if (!isArray(outputResult)) {
                    return [_this._typeIssueFactory(input, options)];
                }
                return _this._applyOperations(input, output, options, null);
            });
        });
    };
    return ExcludeShape;
}(Shape));

/**
 * The shape can coerce input value type during parsing.
 *
 * @template InputValue The input value.
 * @template OutputValue The output value.
 * @template CoercedValue The value to which an input is coerced.
 * @group Shapes
 */
var CoercibleShape = /** @class */ (function (_super) {
    __extends(CoercibleShape, _super);
    function CoercibleShape() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        /**
         * `true` if this shapes coerces input values to the required type during parsing, or `false` otherwise.
         */
        _this.isCoercing = false;
        return _this;
    }
    /**
     * Enables an input value coercion.
     *
     * @param cb The callback that overrides the built-in coercion.
     * @returns The clone of the shape.
     */
    CoercibleShape.prototype.coerce = function (
    /**
     * @param value The value to coerce.
     * @returns The coerced value, or {@link NEVER} if coercion isn't possible.
     */
    cb) {
        var shape = this._clone();
        shape.isCoercing = true;
        if (cb !== undefined) {
            shape._coerce = cb;
        }
        return shape;
    };
    return CoercibleShape;
}(Shape));

/**
 * The shape of an array or a tuple value.
 *
 * | Shape | Type |
 * | :-- | :-- |
 * | `ArrayShape<[], null>` | `[]` |
 * | `ArrayShape<[], Shape<B>>` | `B[]` |
 * | `ArrayShape<[Shape<A>, Shape<B>], null>` | `[A, B]` |
 * | `ArrayShape<[Shape<A>], Shape<B>>` | `[A, ...B[]]` |
 *
 * @template HeadShapes The array of positioned element shapes.
 * @template RestShape The shape of rest elements, or `null` if there are no rest elements.
 * @group Shapes
 */
var ArrayShape = /** @class */ (function (_super) {
    __extends(ArrayShape, _super);
    /**
     * Creates a new {@link ArrayShape} instance.
     *
     * @param headShapes The array of positioned element shapes.
     * @param restShape The shape of rest elements or `null` if there are no rest elements.
     * @param options The issue options or the issue message.
     * @template HeadShapes The array of positioned element shapes.
     * @template RestShape The shape of rest elements, or `null` if there are no rest elements.
     */
    function ArrayShape(
    /**
     * The array of positioned element shapes.
     */
    headShapes, 
    /**
     * The shape of rest elements or `null` if there are no rest elements.
     */
    restShape, options) {
        var _this = _super.call(this) || this;
        _this.headShapes = headShapes;
        _this.restShape = restShape;
        _this._options = options;
        if (headShapes.length !== 0 || restShape === null) {
            _this._typeIssueFactory = createIssueFactory(CODE_TYPE_TUPLE, MESSAGE_TYPE_TUPLE, options, headShapes.length);
        }
        else {
            _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_TYPE_ARRAY, options, TYPE_ARRAY);
        }
        return _this;
    }
    ArrayShape.prototype.at = function (key) {
        var index = toArrayIndex(key);
        if (index === -1) {
            return null;
        }
        if (index < this.headShapes.length) {
            return this.headShapes[index];
        }
        return this.restShape;
    };
    /**
     * Returns an array shape that has rest elements constrained by the given shape.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * @param restShape The shape of rest elements, or `null` if there are no rest elements.
     * @returns The new array shape.
     * @template S The shape of rest elements.
     */
    ArrayShape.prototype.rest = function (restShape) {
        return new ArrayShape(this.headShapes, restShape, this._options);
    };
    ArrayShape.prototype.deepPartial = function () {
        var headShapes = this.headShapes.map(function (shape) { return toDeepPartialShape(shape).optional(); });
        var restShape = this.restShape !== null ? toDeepPartialShape(this.restShape).optional() : null;
        return new ArrayShape(headShapes, restShape, this._options);
    };
    ArrayShape.prototype._isAsync = function () {
        var _a;
        return isAsyncShapes(this.headShapes) || ((_a = this.restShape) === null || _a === void 0 ? void 0 : _a.isAsync) || false;
    };
    ArrayShape.prototype._getInputs = function () {
        var _a = this, headShapes = _a.headShapes, restShape = _a.restShape;
        if (!this.isCoercing) {
            return [TYPE_ARRAY];
        }
        if (headShapes.length > 1) {
            return [TYPE_OBJECT, TYPE_ARRAY];
        }
        if (headShapes.length === 1) {
            return headShapes[0].inputs.concat(TYPE_OBJECT, TYPE_ARRAY);
        }
        if (restShape !== null) {
            return restShape.inputs.concat(TYPE_OBJECT, TYPE_ARRAY);
        }
        return [TYPE_UNKNOWN];
    };
    ArrayShape.prototype._apply = function (input, options, nonce) {
        var _a = this, headShapes = _a.headShapes, restShape = _a.restShape, operations = _a.operations;
        var output = input;
        var outputLength;
        var headShapesLength = headShapes.length;
        var issues = null;
        if (
        // Not an array or not coercible
        (!isArray(output) && (!(options.coerce || this.isCoercing) || (output = this._coerce(input)) === NEVER)) ||
            // Invalid tuple length
            (outputLength = output.length) < headShapesLength ||
            (restShape === null && outputLength !== headShapesLength)) {
            return [this._typeIssueFactory(input, options)];
        }
        if (headShapesLength !== 0 || restShape !== null) {
            for (var i = 0; i < outputLength; ++i) {
                var value = output[i];
                var valueShape = i < headShapesLength ? headShapes[i] : restShape;
                var result = valueShape['_apply'](value, options, nonce);
                if (result === null) {
                    continue;
                }
                if (isArray(result)) {
                    unshiftIssuesPath(result, i);
                    if (options.earlyReturn) {
                        return result;
                    }
                    issues = concatIssues(issues, result);
                    continue;
                }
                if (issues === null || operations.length !== 0) {
                    if (input === output) {
                        output = input.slice(0);
                    }
                    output[i] = result.value;
                }
            }
        }
        return this._applyOperations(input, output, options, issues);
    };
    ArrayShape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        return new Promise(function (resolve) {
            var _a = _this, headShapes = _a.headShapes, restShape = _a.restShape, operations = _a.operations;
            var output = input;
            var outputLength;
            var headShapesLength = headShapes.length;
            if (
            // Not an array or not coercible
            (!isArray(output) && (!(options.coerce || _this.isCoercing) || (output = _this._coerce(input)) === NEVER)) ||
                // Invalid tuple length
                (outputLength = output.length) < headShapesLength ||
                (restShape === null && outputLength !== headShapesLength)) {
                resolve([_this._typeIssueFactory(input, options)]);
                return;
            }
            var issues = null;
            var index = -1;
            var handleResult = function (result) {
                if (result !== null) {
                    if (isArray(result)) {
                        unshiftIssuesPath(result, index);
                        if (options.earlyReturn) {
                            return result;
                        }
                        issues = concatIssues(issues, result);
                    }
                    else if (issues === null || operations.length !== 0) {
                        if (input === output) {
                            output = input.slice(0);
                        }
                        output[index] = result.value;
                    }
                }
                return next();
            };
            var next = function () {
                index++;
                if (index !== outputLength && (headShapesLength !== 0 || restShape !== null)) {
                    return applyShape(index < headShapesLength ? headShapes[index] : restShape, output[index], options, nonce, handleResult);
                }
                return _this._applyOperations(input, output, options, issues);
            };
            resolve(next());
        });
    };
    /**
     * Coerces a value to an array.
     *
     * @param value The non-array value to coerce.
     * @returns An array, or {@link NEVER} if coercion isn't possible.
     */
    ArrayShape.prototype._coerce = function (value) {
        value = getCanonicalValueOf(value);
        if (isIterableObject(value)) {
            return Array.from(value);
        }
        return [value];
    };
    return ArrayShape;
}(CoercibleShape));

/**
 * The shape of a bigint value.
 *
 * @group Shapes
 */
var BigIntShape = /** @class */ (function (_super) {
    __extends(BigIntShape, _super);
    /**
     * Creates a new {@link BigIntShape} instance.
     *
     * @param options The issue options or the issue message.
     */
    function BigIntShape(options) {
        var _this = _super.call(this) || this;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_TYPE_BIGINT, options, TYPE_BIGINT);
        return _this;
    }
    BigIntShape.prototype._getInputs = function () {
        if (this.isCoercing) {
            return [TYPE_BIGINT, TYPE_OBJECT, TYPE_STRING, TYPE_NUMBER, TYPE_BOOLEAN, TYPE_ARRAY, null, undefined];
        }
        else {
            return [TYPE_BIGINT];
        }
    };
    BigIntShape.prototype._apply = function (input, options, nonce) {
        var output = input;
        if (typeof output !== 'bigint' &&
            (!(options.coerce || this.isCoercing) || (output = this._coerce(input)) === NEVER)) {
            return [this._typeIssueFactory(input, options)];
        }
        return this._applyOperations(input, output, options, null);
    };
    /**
     * Coerces a value to a bigint.
     *
     * @param value The non-bigint value to coerce.
     * @returns A bigint value, or {@link NEVER} if coercion isn't possible.
     */
    BigIntShape.prototype._coerce = function (value) {
        if (isArray(value) && value.length === 1 && typeof (value = value[0]) === 'bigint') {
            return value;
        }
        if (value === null || value === undefined) {
            return BigInt(0);
        }
        value = getCanonicalValueOf(value);
        if (typeof value === 'number' || typeof value === 'string' || typeof value === 'boolean') {
            try {
                return BigInt(value);
            }
            catch (_a) { }
        }
        return NEVER;
    };
    return BigIntShape;
}(CoercibleShape));

/**
 * The shape of a boolean value.
 *
 * @group Shapes
 */
var BooleanShape = /** @class */ (function (_super) {
    __extends(BooleanShape, _super);
    /**
     * Creates a new {@link BooleanShape} instance.
     *
     * @param options The issue options or the issue message.
     */
    function BooleanShape(options) {
        var _this = _super.call(this) || this;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_TYPE_BOOLEAN, options, TYPE_BOOLEAN);
        return _this;
    }
    BooleanShape.prototype._getInputs = function () {
        if (this.isCoercing) {
            return [TYPE_BOOLEAN, TYPE_OBJECT, TYPE_STRING, TYPE_NUMBER, TYPE_ARRAY, null, undefined];
        }
        else {
            return [TYPE_BOOLEAN];
        }
    };
    BooleanShape.prototype._apply = function (input, options, nonce) {
        var output = input;
        if (typeof output !== 'boolean' &&
            (!(options.coerce || this.isCoercing) || (output = this._coerce(input)) === NEVER)) {
            return [this._typeIssueFactory(input, options)];
        }
        return this._applyOperations(input, output, options, null);
    };
    /**
     * Coerces a value to a boolean.
     *
     * @param value The non-boolean value to coerce.
     * @returns A boolean value, or {@link NEVER} if coercion isn't possible.
     */
    BooleanShape.prototype._coerce = function (value) {
        if (isArray(value) && value.length === 1 && typeof (value = value[0]) === 'boolean') {
            return value;
        }
        value = getCanonicalValueOf(value);
        if (typeof value === 'boolean') {
            return value;
        }
        if (value === null || value === undefined || value === false || value === 0 || value === 'false') {
            return false;
        }
        if (value === true || value === 1 || value === 'true') {
            return true;
        }
        return NEVER;
    };
    return BooleanShape;
}(CoercibleShape));

/**
 * The shape of a constant value.
 *
 * @template Value The expected constant value.
 * @group Shapes
 */
var ConstShape = /** @class */ (function (_super) {
    __extends(ConstShape, _super);
    /**
     * Creates a new {@link ConstShape} instance.
     *
     * @param value The expected value.
     * @param options The issue options or the issue message.
     * @template Value The expected value.
     */
    function ConstShape(
    /**
     * The expected constant value.
     */
    value, options) {
        var _this = _super.call(this) || this;
        _this.value = value;
        _this._typePredicate = value !== value ? Number.isNaN : function (input) { return value === input; };
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE_CONST, MESSAGE_TYPE_CONST, options, value);
        return _this;
    }
    ConstShape.prototype._getInputs = function () {
        return [this.value];
    };
    ConstShape.prototype._apply = function (input, options, nonce) {
        if (!this._typePredicate(input)) {
            return [this._typeIssueFactory(input, options)];
        }
        return this._applyOperations(input, input, options, null);
    };
    return ConstShape;
}(Shape));

/**
 * The shape of the {@link !Date Date} object.
 * @group Shapes
 */
var DateShape = /** @class */ (function (_super) {
    __extends(DateShape, _super);
    /**
     * Creates a new {@link DateShape} instance.
     *
     * @param options The issue options or the issue message.
     */
    function DateShape(options) {
        var _this = _super.call(this) || this;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_TYPE_DATE, options, TYPE_DATE);
        return _this;
    }
    DateShape.prototype._getInputs = function () {
        if (this.isCoercing) {
            return [TYPE_DATE, TYPE_OBJECT, TYPE_STRING, TYPE_NUMBER, TYPE_ARRAY];
        }
        else {
            return [TYPE_DATE];
        }
    };
    DateShape.prototype._apply = function (input, options, nonce) {
        var output = input;
        if (!isValidDate(input) && (!(options.coerce || this.isCoercing) || (output = this._coerce(input)) === NEVER)) {
            return [this._typeIssueFactory(input, options)];
        }
        return this._applyOperations(input, output, options, null);
    };
    /**
     * Coerces a value to a {@link !Date Date}.
     *
     * @param value The non-{@link !Date Date} value to coerce.
     * @returns A {@link !Date Date} value, or {@link NEVER} if coercion isn't possible.
     */
    DateShape.prototype._coerce = function (value) {
        if (isArray(value) && value.length === 1 && isValidDate((value = value[0]))) {
            return value;
        }
        value = getCanonicalValueOf(value);
        if ((typeof value === 'string' || typeof value === 'number') && isValidDate((value = new Date(value)))) {
            return value;
        }
        return NEVER;
    };
    return DateShape;
}(CoercibleShape));

/**
 * The shape of a value enumeration.
 *
 * @template Value The union of allowed enum values.
 * @group Shapes
 */
var EnumShape = /** @class */ (function (_super) {
    __extends(EnumShape, _super);
    /**
     * Creates a new {@link EnumShape} instance.
     *
     * @param source The array of allowed values, a const key-value mapping, or an enum object.
     * @param options The issue options or the issue message.
     * @template Value The union of allowed enum values.
     */
    function EnumShape(
    /**
     * The array of allowed values, a const key-value mapping, or an TypeScript enum object.
     */
    source, options) {
        var _this = _super.call(this) || this;
        _this.source = source;
        _this.values = (isArray(source) ? source : getEnumValues(source)).filter(unique);
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE_ENUM, MESSAGE_TYPE_ENUM, options, _this.values);
        return _this;
    }
    EnumShape.prototype._getInputs = function () {
        var inputs = this.values.slice(0);
        if (!this.isCoercing || inputs.length === 0) {
            return inputs;
        }
        if (!isArray(this.source)) {
            inputs.push.apply(inputs, Object.keys(this.source));
        }
        return inputs.concat(TYPE_ARRAY, TYPE_OBJECT);
    };
    EnumShape.prototype._apply = function (input, options, nonce) {
        var output = input;
        if (!this.values.includes(output) &&
            (!(options.coerce || this.isCoercing) || (output = this._coerce(input)) === NEVER)) {
            return [this._typeIssueFactory(input, options)];
        }
        return this._applyOperations(input, output, options, null);
    };
    /**
     * Coerces a value to an enum value.
     *
     * @param value The non-enum value to coerce.
     * @returns An enum value, or {@link NEVER} if coercion isn't possible.
     */
    EnumShape.prototype._coerce = function (value) {
        var source = this.source;
        if (isArray(value) && value.length === 1 && this.values.includes((value = value[0]))) {
            return value;
        }
        if (!isArray(source) && typeof (value = getCanonicalValueOf(value)) === 'string' && source.hasOwnProperty(value)) {
            return source[value];
        }
        return NEVER;
    };
    return EnumShape;
}(CoercibleShape));
/**
 * Returns unique values of the enum. Source must contain key-value and value-key mapping to be considered a native
 * enum.
 */
function getEnumValues(source) {
    var values = [];
    for (var key in source) {
        var a = source[key];
        var b = source[a];
        var aType = typeof a;
        var bType = typeof b;
        if (((aType !== 'string' || bType !== 'number') && (aType !== 'number' || bType !== 'string')) || b != key) {
            return Object.values(source);
        }
        if (typeof a === 'number') {
            values.push(a);
        }
    }
    return values;
}

var KEY_THIS = 'this';
var KEY_ARGS = 'arguments';
var KEY_RETURN = 'return';
/**
 * The shape of a function.
 *
 * @template ArgsShape The shape of the array of arguments.
 * @template ReturnShape The return value shape, or `null` if unconstrained.
 * @template ThisShape The shape of `this` argument, or `null` if unconstrained.
 * @group Shapes
 */
var FunctionShape = /** @class */ (function (_super) {
    __extends(FunctionShape, _super);
    /**
     * Creates a new {@link FunctionShape} instance.
     *
     * @param argsShape The shape of the array of arguments.
     * @param returnShape The return value shape, or `null` if unconstrained.
     * @param thisShape The shape of `this` argument, or `null` if unconstrained.
     * @param options The issue options or the issue message.
     * @template ArgsShape The shape of the array of arguments.
     * @template ReturnShape The return value shape, or `null` if unconstrained.
     * @template ThisShape The shape of `this` argument, or `null` if unconstrained.
     */
    function FunctionShape(
    /**
     * The shape of the array of arguments.
     */
    argsShape, 
    /**
     * The return value shape, or `null` if unconstrained.
     */
    returnShape, 
    /**
     * The shape of `this` value, or `null` if unconstrained.
     */
    thisShape, options) {
        var _this = _super.call(this) || this;
        _this.argsShape = argsShape;
        _this.returnShape = returnShape;
        _this.thisShape = thisShape;
        /**
         * `true` if input functions are wrapped during parsing to ensure runtime signature type-safety, or `false` otherwise.
         */
        _this.isStrict = false;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_TYPE_FUNCTION, options, TYPE_FUNCTION);
        return _this;
    }
    Object.defineProperty(FunctionShape.prototype, "isAsyncFunction", {
        /**
         * `true` if some shapes that describe the function signature are {@link Shape#isAsync async}, or `false` otherwise.
         */
        get: function () {
            var _a, _b;
            return ((_a = this.returnShape) === null || _a === void 0 ? void 0 : _a.isAsync) || ((_b = this.thisShape) === null || _b === void 0 ? void 0 : _b.isAsync) || this.argsShape.isAsync;
        },
        enumerable: false,
        configurable: true
    });
    /**
     * Constrains the function return value with the given shape.
     *
     * @param shape The return value shape, or `null` if unconstrained.
     * @returns The new function shape.
     * @template S The return value shape.
     */
    FunctionShape.prototype.return = function (shape) {
        return copyOperations(this, new FunctionShape(this.argsShape, shape, this.thisShape));
    };
    /**
     * Constrains the type of `this` inside the function.
     *
     * @param shape The shape of `this` argument, or `null` if unconstrained.
     * @returns The new function shape.
     * @template S The shape of `this` argument.
     */
    FunctionShape.prototype.this = function (shape) {
        return copyOperations(this, new FunctionShape(this.argsShape, this.returnShape, shape));
    };
    /**
     * Enables input function wrapping during parsing to ensure runtime signature type-safety. Wrapper ensures that input
     * function receives arguments and `this` values that conform {@link FunctionShape.argsShape} and
     * {@link FunctionShape.thisShape} respectively, and returns the value that conforms
     * {@link FunctionShape.returnShape}.
     *
     * @param options Options that are used by the wrapper. If omitted then default options are applied: no early-return,
     * no type coercion.
     * @returns The new function shape.
     */
    FunctionShape.prototype.strict = function (options) {
        var shape = this._clone();
        shape.isStrict = true;
        shape._parseOptions = options;
        return shape;
    };
    FunctionShape.prototype.ensure = function (fn, options) {
        if (this.isAsyncFunction) {
            throw new Error(ERR_ASYNC_FUNCTION);
        }
        var _a = this, argsShape = _a.argsShape, returnShape = _a.returnShape, thisShape = _a.thisShape;
        options || (options = this._parseOptions || defaultApplyOptions);
        return function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            var fnValue = fn.apply(thisShape !== null ? getValue(KEY_THIS, thisShape['_apply'](this, options, nextNonce()), this, options) : this, getValue(KEY_ARGS, argsShape['_apply'](args, options, nextNonce()), args, options));
            if (returnShape !== null) {
                return getValue(KEY_RETURN, returnShape['_apply'](fnValue, options, nextNonce()), fnValue, options);
            }
            return fnValue;
        };
    };
    FunctionShape.prototype.ensureAsync = function (fn, options) {
        var _a = this, argsShape = _a.argsShape, returnShape = _a.returnShape, thisShape = _a.thisShape;
        options || (options = this._parseOptions || defaultApplyOptions);
        return function () {
            var _this = this;
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return new Promise(function (resolve) {
                var fnValue;
                if (thisShape !== null) {
                    fnValue = applyShape(thisShape, _this, options, nextNonce(), function (thisResult) {
                        var thisValue = getValue(KEY_THIS, thisResult, _this, options);
                        return applyShape(argsShape, args, options, nextNonce(), function (argsResult) {
                            return fn.apply(thisValue, getValue(KEY_ARGS, argsResult, args, options));
                        });
                    });
                }
                else {
                    fnValue = applyShape(argsShape, args, options, nextNonce(), function (argsResult) {
                        return fn.apply(_this, getValue(KEY_ARGS, argsResult, args, options));
                    });
                }
                if (returnShape !== null) {
                    resolve(applyShape(returnShape, fnValue, options, nextNonce(), function (resultResult) {
                        return getValue(KEY_RETURN, resultResult, fnValue, options);
                    }));
                }
                else {
                    resolve(fnValue);
                }
            });
        };
    };
    FunctionShape.prototype._getInputs = function () {
        return [TYPE_FUNCTION];
    };
    FunctionShape.prototype._apply = function (input, options, nonce) {
        if (typeof input !== 'function') {
            return [this._typeIssueFactory(input, options)];
        }
        var result = this._applyOperations(input, input, options, null);
        if (isArray(result) || !this.isStrict) {
            return result;
        }
        var output = result === null ? input : result.value;
        return ok(this.isAsyncFunction ? this.ensureAsync(output) : this.ensure(output));
    };
    return FunctionShape;
}(Shape));
function getValue(key, result, input, options) {
    if (result === null) {
        return input;
    }
    if (isArray(result)) {
        unshiftIssuesPath(result, key);
        throw new ValidationError(result, getMessage(result, input, options));
    }
    return result.value;
}

/**
 * The shape of the class instance.
 *
 * @template Ctor The class constructor.
 * @group Shapes
 */
var InstanceShape = /** @class */ (function (_super) {
    __extends(InstanceShape, _super);
    /**
     * Creates a new {@link InstanceShape} instance.
     *
     * @param ctor The class constructor.
     * @param options The issue options or the issue message.
     * @template Ctor The class constructor.
     */
    function InstanceShape(ctor, options) {
        var _this = _super.call(this) || this;
        _this.ctor = ctor;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE_INSTANCE_OF, MESSAGE_TYPE_INSTANCE_OF, options, ctor);
        return _this;
    }
    InstanceShape.prototype._getInputs = function () {
        var ctor = this.ctor;
        if (isEqualOrSubclass(ctor, Function)) {
            return [TYPE_FUNCTION];
        }
        if (isEqualOrSubclass(ctor, Promise)) {
            return [TYPE_PROMISE];
        }
        if (isEqualOrSubclass(ctor, Array)) {
            return [TYPE_ARRAY];
        }
        if (isEqualOrSubclass(ctor, Date)) {
            return [TYPE_DATE];
        }
        if (isEqualOrSubclass(ctor, Set)) {
            return [TYPE_SET];
        }
        if (isEqualOrSubclass(ctor, Map)) {
            return [TYPE_MAP];
        }
        return [TYPE_OBJECT];
    };
    InstanceShape.prototype._apply = function (input, options, nonce) {
        if (!(input instanceof this.ctor)) {
            return [this._typeIssueFactory(input, options)];
        }
        return this._applyOperations(input, input, options, null);
    };
    return InstanceShape;
}(Shape));

/**
 * The shape that requires an input to conform all given shapes.
 *
 * @template Shapes The array of shapes that comprise an intersection.
 * @group Shapes
 */
var IntersectionShape = /** @class */ (function (_super) {
    __extends(IntersectionShape, _super);
    /**
     * Creates a new {@link IntersectionShape} instance.
     *
     * @param shapes The array of shapes that comprise an intersection.
     * @param options The intersection constraint options or an issue message.
     * @template Shapes The array of shapes that comprise an intersection.
     */
    function IntersectionShape(
    /**
     * The array of shapes that comprise an intersection.
     */
    shapes, options) {
        var _this = _super.call(this) || this;
        _this.shapes = shapes;
        _this._options = options;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE_INTERSECTION, MESSAGE_TYPE_INTERSECTION, options, undefined);
        return _this;
    }
    IntersectionShape.prototype.at = function (key) {
        var shapes = [];
        for (var _i = 0, _a = this.shapes; _i < _a.length; _i++) {
            var shape = _a[_i];
            var valueShape = shape.at(key);
            if (valueShape !== null) {
                shapes.push(valueShape);
            }
        }
        if (shapes.length === 0) {
            return null;
        }
        if (shapes.length === 1) {
            return shapes[0];
        }
        return new IntersectionShape(shapes);
    };
    IntersectionShape.prototype.deepPartial = function () {
        return new IntersectionShape(this.shapes.map(toDeepPartialShape), this._options);
    };
    IntersectionShape.prototype._isAsync = function () {
        return isAsyncShapes(this.shapes);
    };
    IntersectionShape.prototype._getInputs = function () {
        var inputs = [];
        for (var _i = 0, _a = this.shapes; _i < _a.length; _i++) {
            var shape = _a[_i];
            inputs.push(shape.inputs);
        }
        return distributeTypes(inputs);
    };
    IntersectionShape.prototype._apply = function (input, options, nonce) {
        var shapes = this.shapes;
        var shapesLength = shapes.length;
        var outputs = null;
        var issues = null;
        for (var i = 0; i < shapesLength; ++i) {
            var result = shapes[i]['_apply'](input, options, nonce);
            if (result === null) {
                continue;
            }
            if (isArray(result)) {
                if (options.earlyReturn) {
                    return result;
                }
                issues = concatIssues(issues, result);
                continue;
            }
            var output = result.value;
            if (outputs === null) {
                outputs = [output];
                continue;
            }
            if (!outputs.includes(output)) {
                outputs.push(output);
            }
        }
        if (issues === null) {
            return this._applyIntersection(input, outputs, options);
        }
        return issues;
    };
    IntersectionShape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        return new Promise(function (resolve) {
            var shapes = _this.shapes;
            var shapesLength = shapes.length;
            var outputs = null;
            var issues = null;
            var index = -1;
            var handleResult = function (result) {
                if (result !== null) {
                    if (isArray(result)) {
                        if (options.earlyReturn) {
                            return result;
                        }
                        issues = concatIssues(issues, result);
                    }
                    else {
                        var output = result.value;
                        if (outputs === null) {
                            outputs = [output];
                        }
                        else if (!outputs.includes(output)) {
                            outputs.push(output);
                        }
                    }
                }
                return next();
            };
            var next = function () {
                index++;
                if (index !== shapesLength) {
                    return applyShape(shapes[index], input, options, nonce, handleResult);
                }
                if (issues === null) {
                    return _this._applyIntersection(input, outputs, options);
                }
                return issues;
            };
            resolve(next());
        });
    };
    /**
     * Merge values and apply checks.
     */
    IntersectionShape.prototype._applyIntersection = function (input, outputs, options) {
        var output = input;
        if (outputs !== null) {
            var outputsLength = outputs.length;
            if (outputsLength !== this.shapes.length) {
                outputsLength = outputs.push(input);
            }
            output = outputs[0];
            for (var i = 1; i < outputsLength && output !== NEVER; ++i) {
                output = mergeValues(output, outputs[i]);
            }
            if (output === NEVER) {
                return [this._typeIssueFactory(input, options)];
            }
        }
        return this._applyOperations(input, output, options, null);
    };
    return IntersectionShape;
}(Shape));
/**
 * Merges values into one, or returns {@link NEVER} if values are incompatible.
 */
function mergeValues(a, b) {
    if (isEqual(a, b)) {
        return a;
    }
    var aType = getTypeOf(a);
    var bType = getTypeOf(b);
    var output;
    if (aType !== bType) {
        return NEVER;
    }
    if (aType === TYPE_OBJECT) {
        output = Object.assign({}, a);
        for (var key in b) {
            if (key in output && setObjectProperty(output, key, mergeValues(output[key], b[key])) === NEVER) {
                return NEVER;
            }
            else {
                setObjectProperty(output, key, b[key]);
            }
        }
        return output;
    }
    if (aType === TYPE_ARRAY) {
        if (a.length !== b.length) {
            return NEVER;
        }
        output = a.slice(0);
        for (var i = 0; i < a.length; ++i) {
            if ((output[i] = mergeValues(a[i], b[i])) === NEVER) {
                return NEVER;
            }
        }
        return output;
    }
    if (aType === TYPE_DATE) {
        return a.getTime() === b.getTime() ? a : NEVER;
    }
    return NEVER;
}

/**
 * Lazily loads a shape using the provider callback.
 *
 * @template ProvidedShape The lazy-loaded shape.
 * @template Pointer The value returned when a cyclic reference is detected.
 * @group Shapes
 */
var LazyShape = /** @class */ (function (_super) {
    __extends(LazyShape, _super);
    /**
     * Creates a new {@link LazyShape} instance.
     *
     * @param shapeProvider The provider callback that returns the shape to which {@link LazyShape} delegates input
     * handling. The provider is called only once.
     * @param pointerProvider The provider callback that returns the value that is used instead of a circular
     * reference.
     * @template ProvidedShape The lazy-loaded shape.
     * @template Pointer The value returned when a cyclic reference is detected.
     */
    function LazyShape(
    /**
     * The provider callback that returns the shape to which {@link LazyShape} delegates input handling.
     */
    shapeProvider, 
    /**
     * The provider callback that returns the value that is used instead of a circular reference.
     */
    pointerProvider) {
        var _this = _super.call(this) || this;
        _this.shapeProvider = shapeProvider;
        _this.pointerProvider = pointerProvider;
        /**
         * The map from nonce to an array of inputs seen during parsing.
         */
        _this._stackMap = new Map();
        // 0 = unavailable
        // 1 = recursive resolution
        var shape = 0;
        _this._cachingShapeProvider = function () {
            if (shape !== 1 && (shape !== 0 || ((shape = 1), (shape = shapeProvider())) instanceof Shape)) {
                return shape;
            }
            shape = 0;
            throw new Error(ERR_SHAPE_EXPECTED);
        };
        return _this;
    }
    Object.defineProperty(LazyShape.prototype, "providedShape", {
        /**
         * The lazy-loaded shape.
         */
        get: function () {
            var shape = this._cachingShapeProvider();
            Object.defineProperty(this, 'providedShape', { configurable: true, value: shape });
            return shape;
        },
        enumerable: false,
        configurable: true
    });
    LazyShape.prototype.at = function (key) {
        return this.providedShape.at(key);
    };
    LazyShape.prototype.deepPartial = function () {
        var _cachingShapeProvider = this._cachingShapeProvider;
        return new LazyShape(function () { return toDeepPartialShape(_cachingShapeProvider()); }, identity);
    };
    /**
     * Replace circular references with a placeholder.
     *
     * @param pointer The value that would be used instead of circular references, or a callback that returns such a
     * value.
     * @template Pointer The value returned when a cyclic reference is detected.
     * @returns The clone of the shape.
     */
    LazyShape.prototype.circular = function (pointer) {
        return copyOperations(this, new LazyShape(this.shapeProvider, typeof pointer === 'function' ? pointer : function () { return pointer; }));
    };
    LazyShape.prototype._isAsync = function () {
        return this.providedShape.isAsync;
    };
    LazyShape.prototype._getInputs = function () {
        return this.providedShape.inputs.slice(0);
    };
    LazyShape.prototype._clone = function () {
        var shape = _super.prototype._clone.call(this);
        shape._stackMap = new Map();
        return shape;
    };
    LazyShape.prototype._apply = function (input, options, nonce) {
        var _stackMap = this._stackMap;
        var stack = _stackMap.get(nonce);
        var leading = stack === undefined;
        if (stack === undefined) {
            stack = [input];
            _stackMap.set(nonce, stack);
        }
        else if (stack.includes(input)) {
            var output = void 0;
            try {
                output = this.pointerProvider(input, options);
            }
            catch (error) {
                return captureIssues(error);
            }
            return input === output ? null : ok(output);
        }
        else {
            stack.push(input);
        }
        try {
            return this._handleResult(this.providedShape['_apply'](input, options, nonce), input, options);
        }
        finally {
            if (leading) {
                _stackMap.delete(nonce);
            }
        }
    };
    LazyShape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        var _stackMap = this._stackMap;
        var stack = _stackMap.get(nonce);
        var leading = stack === undefined;
        if (stack === undefined) {
            stack = [input];
            _stackMap.set(nonce, stack);
        }
        else if (stack.includes(input)) {
            return new Promise(function (resolve) {
                var output;
                try {
                    output = _this.pointerProvider(input, options);
                }
                catch (error) {
                    resolve(captureIssues(error));
                    return;
                }
                resolve(input === output ? null : ok(output));
            });
        }
        else {
            stack.push(input);
        }
        return this.providedShape['_applyAsync'](input, options, nonce).then(function (result) {
            if (leading) {
                _stackMap.delete(nonce);
            }
            return _this._handleResult(result, input, options);
        }, function (error) {
            if (leading) {
                _stackMap.delete(nonce);
            }
            throw error;
        });
    };
    LazyShape.prototype._handleResult = function (result, input, options) {
        var output = input;
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
        }
        return this._applyOperations(input, output, options, null);
    };
    return LazyShape;
}(Shape));

/**
 * The shape of a {@link !Map Map} instance.
 *
 * @template KeyShape The key shape.
 * @template ValueShape The value shape.
 * @group Shapes
 */
var MapShape = /** @class */ (function (_super) {
    __extends(MapShape, _super);
    /**
     * Creates a new {@link MapShape} instance.
     *
     * @param keyShape The key shape.
     * @param valueShape The value shape.
     * @param options The issue options or the issue message.
     * @template KeyShape The key shape.
     * @template ValueShape The value shape.
     */
    function MapShape(
    /**
     * The key shape.
     */
    keyShape, 
    /**
     * The value shape.
     */
    valueShape, options) {
        var _this = _super.call(this) || this;
        _this.keyShape = keyShape;
        _this.valueShape = valueShape;
        _this._options = options;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_TYPE_MAP, options, TYPE_MAP);
        return _this;
    }
    MapShape.prototype.at = function (key) {
        return this.valueShape;
    };
    MapShape.prototype.deepPartial = function () {
        var keyShape = toDeepPartialShape(this.keyShape);
        var valueShape = toDeepPartialShape(this.valueShape).optional();
        return new MapShape(keyShape, valueShape, this._options);
    };
    MapShape.prototype._isAsync = function () {
        return this.keyShape.isAsync || this.valueShape.isAsync;
    };
    MapShape.prototype._getInputs = function () {
        if (this.isCoercing) {
            return [TYPE_MAP, TYPE_OBJECT, TYPE_ARRAY];
        }
        else {
            return [TYPE_MAP];
        }
    };
    MapShape.prototype._apply = function (input, options, nonce) {
        var changed = false;
        var entries;
        if (
        // Not a Map
        !(input instanceof Map && (entries = Array.from(input))) &&
            // No coercion or not coercible
            (!(options.coerce || this.isCoercing) || !(changed = (entries = this._coerce(input)) !== NEVER))) {
            return [this._typeIssueFactory(input, options)];
        }
        var _a = this, keyShape = _a.keyShape, valueShape = _a.valueShape, operations = _a.operations;
        var entriesLength = entries.length;
        var issues = null;
        for (var i = 0; i < entriesLength; ++i) {
            var entry = entries[i];
            var key = entry[0], value = entry[1];
            var keyResult = keyShape['_apply'](key, options, nonce);
            if (keyResult !== null) {
                if (isArray(keyResult)) {
                    unshiftIssuesPath(keyResult, key);
                    if (options.earlyReturn) {
                        return keyResult;
                    }
                    issues = concatIssues(issues, keyResult);
                    keyResult = null;
                }
                else {
                    key = keyResult.value;
                }
            }
            var valueResult = valueShape['_apply'](value, options, nonce);
            if (valueResult !== null) {
                if (isArray(valueResult)) {
                    unshiftIssuesPath(valueResult, key);
                    if (options.earlyReturn) {
                        return valueResult;
                    }
                    issues = concatIssues(issues, valueResult);
                    valueResult = null;
                }
                else {
                    value = valueResult.value;
                }
            }
            if ((issues === null || operations.length !== 0) && (keyResult !== null || valueResult !== null)) {
                changed = true;
                entry[0] = key;
                entry[1] = value;
            }
        }
        return this._applyOperations(input, changed ? new Map(entries) : input, options, issues);
    };
    MapShape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        return new Promise(function (resolve) {
            var changed = false;
            var entries;
            if (
            // Not a Map
            !(input instanceof Map && (entries = Array.from(input))) &&
                // No coercion or not coercible
                (!(options.coerce || _this.isCoercing) || !(changed = (entries = _this._coerce(input)) !== NEVER))) {
                resolve([_this._typeIssueFactory(input, options)]);
                return;
            }
            var _a = _this, keyShape = _a.keyShape, valueShape = _a.valueShape, operations = _a.operations;
            var entriesLength = entries.length;
            var issues = null;
            var index = -1;
            var entry;
            var key;
            var value;
            var keyChanged = false;
            var handleKeyResult = function (keyResult) {
                keyChanged = false;
                if (keyResult !== null) {
                    if (isArray(keyResult)) {
                        unshiftIssuesPath(keyResult, key);
                        if (options.earlyReturn) {
                            return keyResult;
                        }
                        issues = concatIssues(issues, keyResult);
                    }
                    else {
                        key = keyResult.value;
                        keyChanged = true;
                    }
                }
                return applyShape(valueShape, value, options, nonce, handleValueResult);
            };
            var handleValueResult = function (valueResult) {
                if (valueResult !== null) {
                    if (isArray(valueResult)) {
                        unshiftIssuesPath(valueResult, key);
                        if (options.earlyReturn) {
                            return valueResult;
                        }
                        issues = concatIssues(issues, valueResult);
                        valueResult = null;
                    }
                    else {
                        value = valueResult.value;
                    }
                }
                if ((issues === null || operations.length !== 0) && (keyChanged || valueResult !== null)) {
                    changed = true;
                    entry[0] = key;
                    entry[1] = value;
                }
                return next();
            };
            var next = function () {
                index++;
                if (index !== entriesLength) {
                    entry = entries[index];
                    key = entry[0];
                    value = entry[1];
                    return applyShape(keyShape, key, options, nonce, handleKeyResult);
                }
                return _this._applyOperations(input, changed ? new Map(entries) : input, options, issues);
            };
            resolve(next());
        });
    };
    /**
     * Coerces a value to an array of {@link !Map Map} entries.
     *
     * @param value A non-{@link !Map Map} value to coerce.
     * @returns An array of entries, or {@link NEVER} if coercion isn't possible.
     */
    MapShape.prototype._coerce = function (value) {
        if (isArray(value)) {
            return value.every(isMapEntry) ? value : NEVER;
        }
        value = getCanonicalValueOf(value);
        if (isIterableObject(value)) {
            value = Array.from(value);
            return value.every(isMapEntry) ? value : NEVER;
        }
        if (isObjectLike(value)) {
            return Object.entries(value);
        }
        return NEVER;
    };
    return MapShape;
}(CoercibleShape));

/**
 * The shape that doesn't match any input.
 *
 * @group Shapes
 */
var NeverShape = /** @class */ (function (_super) {
    __extends(NeverShape, _super);
    /**
     * Creates a new {@link NeverShape} instance.
     *
     * @param options The issue options or the issue message.
     */
    function NeverShape(options) {
        var _this = _super.call(this) || this;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE_NEVER, MESSAGE_TYPE_NEVER, options, undefined);
        return _this;
    }
    NeverShape.prototype._getInputs = function () {
        return [];
    };
    NeverShape.prototype._apply = function (input, options, nonce) {
        return [this._typeIssueFactory(input, options)];
    };
    return NeverShape;
}(Shape));

/**
 * The shape of a number value.
 *
 * @group Shapes
 */
var NumberShape = /** @class */ (function (_super) {
    __extends(NumberShape, _super);
    /**
     * Creates a new {@link NumberShape} instance.
     *
     * @param options The issue options or the issue message.
     */
    function NumberShape(options) {
        var _this = _super.call(this) || this;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_TYPE_NUMBER, options, TYPE_NUMBER);
        return _this;
    }
    NumberShape.prototype.nan = function (defaultValue) {
        return this.replace(NaN, arguments.length === 0 ? NaN : defaultValue);
    };
    NumberShape.prototype._getInputs = function () {
        if (this.isCoercing) {
            return [TYPE_NUMBER, TYPE_OBJECT, TYPE_STRING, TYPE_BOOLEAN, TYPE_ARRAY, TYPE_DATE, null, undefined];
        }
        else {
            return [TYPE_NUMBER];
        }
    };
    NumberShape.prototype._apply = function (input, options, nonce) {
        var output = input;
        if ((typeof output !== 'number' || output !== output) &&
            (!(options.coerce || this.isCoercing) || (output = this._coerce(input)) === NEVER)) {
            return [this._typeIssueFactory(input, options)];
        }
        return this._applyOperations(input, output, options, null);
    };
    /**
     * Coerces a value to a number (not `NaN`).
     *
     * @param value The non-number value to coerce.
     * @returns A number value, or {@link NEVER} if coercion isn't possible.
     */
    NumberShape.prototype._coerce = function (value) {
        if (isArray(value) && value.length === 1 && typeof (value = value[0]) === 'number') {
            return value === value ? value : NEVER;
        }
        if (value === null || value === undefined) {
            return 0;
        }
        value = getCanonicalValueOf(value);
        if ((typeof value === 'string' || typeof value === 'boolean' || typeof value === 'number' || value instanceof Date) &&
            (value = +value) === value) {
            return value;
        }
        return NEVER;
    };
    return NumberShape;
}(CoercibleShape));

/**
 * The shape of an object.
 *
 * @template PropShapes The mapping from a string object key to a corresponding value shape.
 * @template RestShape The shape that constrains values of
 * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures), or `null`
 * if there's no index signature.
 * @group Shapes
 */
var ObjectShape = /** @class */ (function (_super) {
    __extends(ObjectShape, _super);
    /**
     * Creates a new {@link ObjectShape} instance.
     *
     * @param propShapes The mapping from an object key to a corresponding value shape.
     * @param restShape The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures). If `null`
     * then values thea fall under the index signature are unconstrained.
     * @param options The issue options or the issue message.
     * @param keysMode The mode of keys handling.
     * @template PropShapes The mapping from an object key to a corresponding value shape.
     * @template RestShape The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    function ObjectShape(
    /**
     * The mapping from an object key to a corresponding value shape.
     */
    propShapes, 
    /**
     * The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    restShape, options, 
    /**
     * The mode of keys handling.
     */
    keysMode) {
        if (keysMode === void 0) { keysMode = 'preserved'; }
        var _this = _super.call(this) || this;
        _this.propShapes = propShapes;
        _this.restShape = restShape;
        _this.keysMode = keysMode;
        _this.keys = Object.keys(propShapes);
        _this.valueShapes = Object.values(propShapes);
        _this._options = options;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_TYPE_OBJECT, options, TYPE_OBJECT);
        return _this;
    }
    Object.defineProperty(ObjectShape.prototype, "keysShape", {
        /**
         * The enum shape that describes object keys.
         */
        get: function () {
            var keysShape = new EnumShape(this.keys);
            Object.defineProperty(this, 'keysShape', { writable: true, value: keysShape });
            return keysShape;
        },
        enumerable: false,
        configurable: true
    });
    ObjectShape.prototype.at = function (key) {
        return this.propShapes.hasOwnProperty(key) ? this.propShapes[key] : this.restShape;
    };
    ObjectShape.prototype.extend = function (shape) {
        var propsShapes = Object.assign({}, this.propShapes, shape instanceof ObjectShape ? shape.propShapes : shape);
        return new ObjectShape(propsShapes, this.restShape, this._options, this.keysMode);
    };
    /**
     * Returns an object shape that only has properties with listed keys.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * @param keys The array of property keys to pick.
     * @returns The new object shape.
     * @template K The tuple of keys to pick.
     */
    ObjectShape.prototype.pick = function (keys) {
        var propShapes = {};
        for (var key in this.propShapes) {
            if (keys.includes(key)) {
                propShapes[key] = this.propShapes[key];
            }
        }
        return new ObjectShape(propShapes, this.restShape, this._options, this.keysMode);
    };
    /**
     * Returns an object shape that doesn't have the listed keys.
     *
     * **Note:** This method returns a shape without any operations.
     *
     * @param keys The array of property keys to omit.
     * @returns The new object shape.
     * @template K The tuple of keys to omit.
     */
    ObjectShape.prototype.omit = function (keys) {
        var propShapes = {};
        for (var key in this.propShapes) {
            if (!keys.includes(key)) {
                propShapes[key] = this.propShapes[key];
            }
        }
        return new ObjectShape(propShapes, this.restShape, this._options, this.keysMode);
    };
    ObjectShape.prototype.partial = function (keys) {
        var propShapes = {};
        for (var key in this.propShapes) {
            propShapes[key] =
                keys === undefined || keys.includes(key) ? this.propShapes[key].optional() : this.propShapes[key];
        }
        return new ObjectShape(propShapes, this.restShape, this._options, this.keysMode);
    };
    ObjectShape.prototype.deepPartial = function () {
        var propShapes = {};
        for (var key in this.propShapes) {
            propShapes[key] = toDeepPartialShape(this.propShapes[key]).optional();
        }
        var restShape = this.restShape !== null ? toDeepPartialShape(this.restShape).optional() : null;
        return new ObjectShape(propShapes, restShape, this._options, this.keysMode);
    };
    ObjectShape.prototype.required = function (keys) {
        var propShapes = {};
        for (var key in this.propShapes) {
            propShapes[key] =
                keys === undefined || keys.includes(key) ? this.propShapes[key].nonOptional() : this.propShapes[key];
        }
        return copyOperations(this, new ObjectShape(propShapes, this.restShape, this._options, this.keysMode));
    };
    /**
     * Returns an object shape that allows only known keys and has no index signature.
     *
     * @param options The issue options or the issue message.
     * @returns The new object shape.
     */
    ObjectShape.prototype.exact = function (options) {
        var shape = new ObjectShape(this.propShapes, null, this._options, 'exact');
        shape._exactIssueFactory = createIssueFactory(CODE_OBJECT_EXACT, MESSAGE_OBJECT_EXACT, options);
        return copyOperations(this, shape);
    };
    /**
     * Returns an object shape that doesn't have an index signature and all unknown keys are stripped.
     *
     * @returns The new object shape.
     */
    ObjectShape.prototype.strip = function () {
        return copyOperations(this, new ObjectShape(this.propShapes, null, this._options, 'stripped'));
    };
    /**
     * Returns an object shape that has an index signature that doesn't constrain values.
     *
     * @returns The new object shape.
     */
    ObjectShape.prototype.preserve = function () {
        return copyOperations(this, new ObjectShape(this.propShapes, null, this._options));
    };
    /**
     * Returns an object shape that has an index signature that is constrained by the given shape.
     *
     * @param restShape The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures), or `null`
     * if there's no index signature.
     * @returns The new object shape.
     * @template S The index signature shape.
     */
    ObjectShape.prototype.rest = function (restShape) {
        return copyOperations(this, new ObjectShape(this.propShapes, restShape, this._options));
    };
    ObjectShape.prototype._isAsync = function () {
        var _a;
        return ((_a = this.restShape) === null || _a === void 0 ? void 0 : _a.isAsync) || isAsyncShapes(this.valueShapes);
    };
    ObjectShape.prototype._getInputs = function () {
        return [TYPE_OBJECT];
    };
    ObjectShape.prototype._apply = function (input, options, nonce) {
        if (!isObject(input)) {
            return [this._typeIssueFactory(input, options)];
        }
        if (this.keysMode === 'preserved' && this.restShape === null) {
            return this._applyRestUnchecked(input, options, nonce);
        }
        else {
            return this._applyRestChecked(input, options, nonce);
        }
    };
    ObjectShape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        return new Promise(function (resolve) {
            if (!isObject(input)) {
                resolve([_this._typeIssueFactory(input, options)]);
                return;
            }
            var _a = _this, keys = _a.keys, keysMode = _a.keysMode, restShape = _a.restShape, operations = _a.operations, valueShapes = _a.valueShapes;
            var keysLength = keys.length;
            var issues = null;
            var output = input;
            var seenCount = 0;
            var seenBitmask = 0;
            var unknownKeys = null;
            var entries = [];
            for (var key_1 in input) {
                var value = input[key_1];
                var index_1 = keys.indexOf(key_1);
                var valueShape = restShape;
                if (index_1 !== -1) {
                    seenCount++;
                    seenBitmask = toggleBit(seenBitmask, index_1);
                    valueShape = valueShapes[index_1];
                }
                if (valueShape !== null) {
                    entries.push([key_1, value, valueShape]);
                    continue;
                }
                if (keysMode === 'exact') {
                    if (unknownKeys !== null) {
                        unknownKeys.push(key_1);
                        continue;
                    }
                    unknownKeys = [key_1];
                    if (options.earlyReturn) {
                        break;
                    }
                    continue;
                }
                if (input === output && keysMode === 'stripped') {
                    output = cloneDictKeys(input, keys);
                }
            }
            if (unknownKeys !== null) {
                issues = [_this._exactIssueFactory(input, options, unknownKeys)];
                if (options.earlyReturn) {
                    resolve(issues);
                    return;
                }
            }
            if (seenCount !== keysLength) {
                for (var i = 0; i < keysLength; ++i) {
                    if (getBit(seenBitmask, i) === 0) {
                        var key_2 = keys[i];
                        entries.push([key_2, input[key_2], valueShapes[i]]);
                    }
                }
            }
            var entriesLength = entries.length;
            var index = -1;
            var key;
            var handleValueResult = function (result) {
                if (result !== null) {
                    if (isArray(result)) {
                        unshiftIssuesPath(result, key);
                        if (options.earlyReturn) {
                            return result;
                        }
                        issues = concatIssues(issues, result);
                    }
                    else if (issues === null || operations.length !== 0) {
                        if (input === output) {
                            output = cloneDict(input);
                        }
                        setObjectProperty(output, key, result.value);
                    }
                }
                return next();
            };
            var next = function () {
                index++;
                if (index !== entriesLength) {
                    var entry = entries[index];
                    key = entry[0];
                    return applyShape(entry[2], entry[1], options, nonce, handleValueResult);
                }
                return _this._applyOperations(input, output, options, issues);
            };
            resolve(next());
        });
    };
    /**
     * Unknown keys are preserved as is and aren't checked.
     */
    ObjectShape.prototype._applyRestUnchecked = function (input, options, nonce) {
        var _a = this, keys = _a.keys, operations = _a.operations, valueShapes = _a.valueShapes;
        var keysLength = keys.length;
        var issues = null;
        var output = input;
        for (var i = 0; i < keysLength; ++i) {
            var key = keys[i];
            var value = input[key];
            var result = valueShapes[i]['_apply'](value, options, nonce);
            if (result === null) {
                continue;
            }
            if (isArray(result)) {
                unshiftIssuesPath(result, key);
                if (options.earlyReturn) {
                    return result;
                }
                issues = concatIssues(issues, result);
                continue;
            }
            if (issues === null || operations.length !== 0) {
                if (input === output) {
                    output = cloneDict(input);
                }
                setObjectProperty(output, key, result.value);
            }
        }
        return this._applyOperations(input, output, options, issues);
    };
    /**
     * Unknown keys are either parsed with a {@link ObjectShape.restShape}, stripped, or cause an issue.
     */
    ObjectShape.prototype._applyRestChecked = function (input, options, nonce) {
        var _a = this, keys = _a.keys, keysMode = _a.keysMode, restShape = _a.restShape, operations = _a.operations, valueShapes = _a.valueShapes;
        var keysLength = keys.length;
        var issues = null;
        var output = input;
        var seenCount = 0;
        var seenBitmask = 0;
        var unknownKeys = null;
        for (var key in input) {
            var value = input[key];
            var index = keys.indexOf(key);
            var valueShape = restShape;
            // The key is known
            if (index !== -1) {
                seenCount++;
                seenBitmask = toggleBit(seenBitmask, index);
                valueShape = valueShapes[index];
            }
            // The key is known or indexed
            if (valueShape !== null) {
                var result = valueShape['_apply'](value, options, nonce);
                if (result === null) {
                    continue;
                }
                if (isArray(result)) {
                    unshiftIssuesPath(result, key);
                    if (options.earlyReturn) {
                        return result;
                    }
                    issues = concatIssues(issues, result);
                    continue;
                }
                if (issues === null || operations.length !== 0) {
                    if (input === output) {
                        output = restShape === null ? cloneDictKeys(input, keys) : cloneDict(input);
                    }
                    setObjectProperty(output, key, result.value);
                }
                continue;
            }
            // Unknown keys raise an issue
            if (keysMode === 'exact') {
                if (unknownKeys !== null) {
                    unknownKeys.push(key);
                    continue;
                }
                unknownKeys = [key];
                if (options.earlyReturn) {
                    break;
                }
                continue;
            }
            // Unknown keys are stripped
            if (input === output && (issues === null || operations.length !== 0)) {
                output = cloneDictKeys(input, keys);
            }
        }
        // Raise unknown keys issue
        if (unknownKeys !== null) {
            var issue = this._exactIssueFactory(input, options, unknownKeys);
            if (options.earlyReturn) {
                return [issue];
            }
            (issues || (issues = [])).push(issue);
        }
        // Parse absent known keys
        if (seenCount !== keysLength) {
            for (var i = 0; i < keysLength; ++i) {
                if (getBit(seenBitmask, i) === 1) {
                    continue;
                }
                var key = keys[i];
                var value = input[key];
                var result = valueShapes[i]['_apply'](value, options, nonce);
                if (result === null) {
                    continue;
                }
                if (isArray(result)) {
                    unshiftIssuesPath(result, key);
                    if (options.earlyReturn) {
                        return result;
                    }
                    issues = concatIssues(issues, result);
                    continue;
                }
                if (issues === null || operations.length !== 0) {
                    if (input === output) {
                        output = cloneDict(input);
                    }
                    setObjectProperty(output, key, result.value);
                }
            }
        }
        return this._applyOperations(input, output, options, issues);
    };
    return ObjectShape;
}(Shape));

/**
 * The shape of a {@link !Promise Promise} value.
 *
 * @template ValueShape The shape of the resolved value, or `null` if resolved value shouldn't be parsed.
 * @group Shapes
 */
var PromiseShape = /** @class */ (function (_super) {
    __extends(PromiseShape, _super);
    /**
     * Creates a new {@link PromiseShape} instance.
     *
     * @param valueShape The shape of the resolved value, or `null` if resolved value shouldn't be parsed.
     * @param options The issue options or the issue message.
     * @template ValueShape The shape of the resolved value.
     */
    function PromiseShape(valueShape, options) {
        var _this = _super.call(this) || this;
        _this.valueShape = valueShape;
        _this._options = options;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_TYPE_PROMISE, options, TYPE_PROMISE);
        return _this;
    }
    PromiseShape.prototype.deepPartial = function () {
        var valueShape = this.valueShape !== null ? toDeepPartialShape(this.valueShape).optional() : null;
        return new PromiseShape(valueShape, this._options);
    };
    PromiseShape.prototype._isAsync = function () {
        return this.valueShape !== null;
    };
    PromiseShape.prototype._getInputs = function () {
        if (!this.isCoercing) {
            return [TYPE_PROMISE];
        }
        if (this.valueShape === null) {
            return [TYPE_UNKNOWN];
        }
        return this.valueShape.inputs.concat(TYPE_PROMISE);
    };
    PromiseShape.prototype._apply = function (input, options, nonce) {
        var output = input;
        if (!(input instanceof Promise) &&
            (!(options.coerce || this.isCoercing) || (output = this._coerce(input)) === NEVER)) {
            return [this._typeIssueFactory(input, options)];
        }
        return this._applyOperations(input, output, options, null);
    };
    PromiseShape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        var output = input;
        if (!(input instanceof Promise) &&
            (!(options.coerce || this.isCoercing) || (output = this._coerce(input)) === NEVER)) {
            return Promise.resolve([this._typeIssueFactory(input, options)]);
        }
        return output.then(function (value) {
            return applyShape(_this.valueShape, value, options, nonce, function (result) {
                var issues = null;
                if (result !== null) {
                    if (isArray(result)) {
                        if (options.earlyReturn || _this.operations.length === 0) {
                            return result;
                        }
                        issues = result;
                    }
                    else {
                        output = Promise.resolve(result.value);
                    }
                }
                return _this._applyOperations(input, output, options, issues);
            });
        });
    };
    PromiseShape.prototype._coerce = function (value) {
        return Promise.resolve(value);
    };
    return PromiseShape;
}(CoercibleShape));

/**
 * The shape that describes an object with string keys and values that conform the given shape.
 *
 * @template KeyShape The key shape.
 * @template ValueShape The value shape.
 * @group Shapes
 */
var RecordShape = /** @class */ (function (_super) {
    __extends(RecordShape, _super);
    /**
     * Creates a new {@link RecordShape} instance.
     *
     * @param keyShape The key shape, or `null` if keys should be preserved intact.
     * @param valueShape The value shape.
     * @param options The issue options or the issue message.
     * @template KeyShape The key shape.
     * @template ValueShape The value shape.
     */
    function RecordShape(
    /**
     * The key shape or `null` if keys are preserved intact.
     */
    keyShape, 
    /**
     * The value shape.
     */
    valueShape, options) {
        var _this = _super.call(this) || this;
        _this.keyShape = keyShape;
        _this.valueShape = valueShape;
        _this._options = options;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_TYPE_OBJECT, options, TYPE_OBJECT);
        return _this;
    }
    RecordShape.prototype.at = function (key) {
        return typeof key === 'string' || typeof key === 'number' ? this.valueShape : null;
    };
    RecordShape.prototype.deepPartial = function () {
        return new RecordShape(this.keyShape, toDeepPartialShape(this.valueShape).optional(), this._options);
    };
    RecordShape.prototype._isAsync = function () {
        var _a;
        return ((_a = this.keyShape) === null || _a === void 0 ? void 0 : _a.isAsync) || this.valueShape.isAsync;
    };
    RecordShape.prototype._getInputs = function () {
        return [TYPE_OBJECT];
    };
    RecordShape.prototype._apply = function (input, options, nonce) {
        if (!isObject(input)) {
            return [this._typeIssueFactory(input, options)];
        }
        var _a = this, keyShape = _a.keyShape, valueShape = _a.valueShape, operations = _a.operations;
        var output = input;
        var issues = null;
        var index = -1;
        for (var key in input) {
            var value = input[key];
            var keyResult = null;
            index++;
            if (keyShape !== null) {
                keyResult = keyShape['_apply'](key, options, nonce);
                if (keyResult !== null) {
                    if (isArray(keyResult)) {
                        unshiftIssuesPath(keyResult, key);
                        if (options.earlyReturn) {
                            return keyResult;
                        }
                        issues = concatIssues(issues, keyResult);
                        keyResult = null;
                    }
                    else {
                        key = keyResult.value;
                    }
                }
            }
            var valueResult = valueShape['_apply'](value, options, nonce);
            if (valueResult !== null) {
                if (isArray(valueResult)) {
                    unshiftIssuesPath(valueResult, key);
                    if (options.earlyReturn) {
                        return valueResult;
                    }
                    issues = concatIssues(issues, valueResult);
                    valueResult = null;
                }
                else {
                    value = valueResult.value;
                }
            }
            if ((issues === null || operations.length !== 0) && (keyResult !== null || valueResult !== null)) {
                if (input === output) {
                    output = cloneDictHead(input, index);
                }
                setObjectProperty(output, key, value);
            }
        }
        return this._applyOperations(input, output, options, issues);
    };
    RecordShape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        return new Promise(function (resolve) {
            if (!isObject(input)) {
                resolve([_this._typeIssueFactory(input, options)]);
                return;
            }
            var _a = _this, keyShape = _a.keyShape, valueShape = _a.valueShape, operations = _a.operations;
            var keys = Object.keys(input);
            var keysLength = keys.length;
            var output = input;
            var issues = null;
            var index = -1;
            var key;
            var value;
            var keyChanged = false;
            var handleKeyResult = function (keyResult) {
                keyChanged = false;
                if (keyResult !== null) {
                    if (isArray(keyResult)) {
                        unshiftIssuesPath(keyResult, key);
                        if (options.earlyReturn) {
                            return keyResult;
                        }
                        issues = concatIssues(issues, keyResult);
                    }
                    else {
                        key = keyResult.value;
                        keyChanged = true;
                    }
                }
                return applyShape(valueShape, value, options, nonce, handleValueResult);
            };
            var handleValueResult = function (valueResult) {
                if (valueResult !== null) {
                    if (isArray(valueResult)) {
                        unshiftIssuesPath(valueResult, key);
                        if (options.earlyReturn) {
                            return valueResult;
                        }
                        issues = concatIssues(issues, valueResult);
                        valueResult = null;
                    }
                    else {
                        value = valueResult.value;
                    }
                }
                if ((issues === null || operations.length !== 0) && (keyChanged || valueResult !== null)) {
                    if (input === output) {
                        output = cloneDictHead(input, index);
                    }
                    setObjectProperty(output, key, value);
                }
                return next();
            };
            var next = function () {
                index++;
                if (index !== keysLength) {
                    key = keys[index];
                    value = input[key];
                    if (keyShape !== null) {
                        return applyShape(keyShape, key, options, nonce, handleKeyResult);
                    }
                    else {
                        return valueShape['_applyAsync'](value, options, nonce).then(handleValueResult);
                    }
                }
                return _this._applyOperations(input, output, options, issues);
            };
            resolve(next());
        });
    };
    return RecordShape;
}(Shape));

/**
 * The shape of a {@link !Set Set} instance.
 *
 * @template ValueShape The value shape.
 * @group Shapes
 */
var SetShape = /** @class */ (function (_super) {
    __extends(SetShape, _super);
    /**
     * Creates a new {@link SetShape} instance.
     *
     * @param valueShape The value shape.
     * @param options The issue options or the issue message.
     * @template ValueShape The value shape.
     */
    function SetShape(
    /**
     * The value shape.
     */
    valueShape, options) {
        var _this = _super.call(this) || this;
        _this.valueShape = valueShape;
        _this._options = options;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_TYPE_SET, options, TYPE_SET);
        return _this;
    }
    SetShape.prototype.at = function (key) {
        return toArrayIndex(key) === -1 ? null : this.valueShape;
    };
    SetShape.prototype.deepPartial = function () {
        return new SetShape(toDeepPartialShape(this.valueShape).optional(), this._options);
    };
    SetShape.prototype._isAsync = function () {
        return this.valueShape.isAsync;
    };
    SetShape.prototype._getInputs = function () {
        if (this.isCoercing) {
            return this.valueShape.inputs.concat(TYPE_SET, TYPE_OBJECT, TYPE_ARRAY);
        }
        else {
            return [TYPE_SET];
        }
    };
    SetShape.prototype._apply = function (input, options, nonce) {
        var changed = false;
        var values;
        var issues = null;
        if (
        // Not a Set
        !(input instanceof Set && (values = Array.from(input))) &&
            // No coercion or not coercible
            (!(options.coerce || this.isCoercing) || !(changed = (values = this._coerce(input)) !== NEVER))) {
            return [this._typeIssueFactory(input, options)];
        }
        var valueShape = this.valueShape;
        var valuesLength = values.length;
        for (var i = 0; i < valuesLength; ++i) {
            var value = values[i];
            var result = valueShape['_apply'](value, options, nonce);
            if (result === null) {
                continue;
            }
            if (isArray(result)) {
                unshiftIssuesPath(result, i);
                if (options.earlyReturn) {
                    return result;
                }
                issues = concatIssues(issues, result);
                continue;
            }
            changed = true;
            values[i] = result.value;
        }
        return this._applyOperations(input, changed ? new Set(values) : input, options, issues);
    };
    SetShape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        return new Promise(function (resolve) {
            var changed = false;
            var values;
            if (
            // Not a Set
            !(input instanceof Set && (values = Array.from(input))) &&
                // No coercion or not coercible
                (!(options.coerce || _this.isCoercing) || !(changed = (values = _this._coerce(input)) !== NEVER))) {
                resolve([_this._typeIssueFactory(input, options)]);
                return;
            }
            var valueShape = _this.valueShape;
            var valuesLength = values.length;
            var issues = null;
            var index = -1;
            var handleResult = function (result) {
                if (result !== null) {
                    if (isArray(result)) {
                        unshiftIssuesPath(result, index);
                        if (options.earlyReturn) {
                            return result;
                        }
                        issues = concatIssues(issues, result);
                    }
                    else {
                        changed = true;
                        values[index] = result.value;
                    }
                }
                return next();
            };
            var next = function () {
                index++;
                if (index !== valuesLength) {
                    return valueShape['_applyAsync'](values[index], options, nonce).then(handleResult);
                }
                return _this._applyOperations(input, changed ? new Set(values) : input, options, issues);
            };
            resolve(next());
        });
    };
    /**
     * Coerces a value to an array of {@link !Set Set} values.
     *
     * @param value The non-{@link !Set Set} value to coerce.
     * @returns An array, or {@link NEVER} if coercion isn't possible.
     */
    SetShape.prototype._coerce = function (value) {
        value = getCanonicalValueOf(value);
        if (isArray(value)) {
            return value;
        }
        if (isIterableObject(value)) {
            return Array.from(value);
        }
        return [value];
    };
    return SetShape;
}(CoercibleShape));

/**
 * The shape of a string value.
 *
 * @group Shapes
 */
var StringShape = /** @class */ (function (_super) {
    __extends(StringShape, _super);
    /**
     * Creates a new {@link StringShape} instance.
     *
     * @param options The issue options or the issue message.
     */
    function StringShape(options) {
        var _this = _super.call(this) || this;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_TYPE_STRING, options, TYPE_STRING);
        return _this;
    }
    StringShape.prototype._getInputs = function () {
        if (this.isCoercing) {
            return [TYPE_STRING, TYPE_OBJECT, TYPE_NUMBER, TYPE_BOOLEAN, TYPE_BIGINT, TYPE_ARRAY, null, undefined];
        }
        else {
            return [TYPE_STRING];
        }
    };
    StringShape.prototype._apply = function (input, options, nonce) {
        var output = input;
        if (typeof output !== 'string' &&
            (!(options.coerce || this.isCoercing) || (output = this._coerce(input)) === NEVER)) {
            return [this._typeIssueFactory(input, options)];
        }
        return this._applyOperations(input, output, options, null);
    };
    /**
     * Coerces a value to a string.
     *
     * @param value The non-string value to coerce.
     * @returns A string value, or {@link NEVER} if coercion isn't possible.
     */
    StringShape.prototype._coerce = function (value) {
        if (isArray(value) && value.length === 1 && typeof (value = value[0]) === 'string') {
            return value;
        }
        if (value === null || value === undefined) {
            return '';
        }
        value = getCanonicalValueOf(value);
        if (typeof value === 'string') {
            return value;
        }
        if (Number.isFinite(value) || typeof value === 'boolean' || typeof value === 'bigint') {
            return '' + value;
        }
        if (isValidDate(value)) {
            return value.toISOString();
        }
        return NEVER;
    };
    return StringShape;
}(CoercibleShape));

/**
 * The shape of an arbitrary symbol value.
 *
 * @group Shapes
 */
var SymbolShape = /** @class */ (function (_super) {
    __extends(SymbolShape, _super);
    /**
     * Creates a new {@link SymbolShape} instance.
     *
     * @param options The issue options or the issue message.
     */
    function SymbolShape(options) {
        var _this = _super.call(this) || this;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_TYPE_SYMBOL, options, TYPE_SYMBOL);
        return _this;
    }
    SymbolShape.prototype._getInputs = function () {
        return [TYPE_SYMBOL];
    };
    SymbolShape.prototype._apply = function (input, options, nonce) {
        if (typeof input !== 'symbol') {
            return [this._typeIssueFactory(input, options)];
        }
        return this._applyOperations(input, input, options, null);
    };
    return SymbolShape;
}(Shape));

/**
 * The shape that requires an input to conform at least one of shapes.
 *
 * @template Shapes The array of shapes that comprise a union.
 * @group Shapes
 */
var UnionShape = /** @class */ (function (_super) {
    __extends(UnionShape, _super);
    /**
     * Creates a new {@link UnionShape} instance.
     *
     * @param shapes The array of shapes that comprise a union.
     * @param options The union constraint options or an issue message.
     * @template Shapes The array of shapes that comprise a union.
     */
    function UnionShape(
    /**
     * The array of shapes that comprise a union.
     */
    shapes, options) {
        var _this = _super.call(this) || this;
        _this.shapes = shapes;
        _this._options = options;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE_UNION, MESSAGE_TYPE_UNION, options);
        return _this;
    }
    Object.defineProperty(UnionShape.prototype, "_lookup", {
        get: function () {
            var shapes = this.shapes.filter(unique);
            var lookup = createLookupByDiscriminator(shapes) || createLookupByType(shapes);
            Object.defineProperty(this, '_lookup', { configurable: true, value: lookup });
            return lookup;
        },
        enumerable: false,
        configurable: true
    });
    UnionShape.prototype.at = function (key) {
        var shapes = [];
        for (var _i = 0, _a = this.shapes; _i < _a.length; _i++) {
            var shape = _a[_i];
            var valueShape = shape.at(key);
            if (valueShape !== null) {
                shapes.push(valueShape);
            }
        }
        if (shapes.length === 0) {
            return null;
        }
        if (shapes.length === 1) {
            return shapes[0];
        }
        return new UnionShape(shapes);
    };
    UnionShape.prototype.deepPartial = function () {
        return new UnionShape(this.shapes.map(toDeepPartialShape), this._options);
    };
    UnionShape.prototype._isAsync = function () {
        return isAsyncShapes(this.shapes);
    };
    UnionShape.prototype._getInputs = function () {
        var inputs = [];
        for (var _i = 0, _a = this.shapes; _i < _a.length; _i++) {
            var shape = _a[_i];
            inputs.push.apply(inputs, shape.inputs);
        }
        return inputs;
    };
    UnionShape.prototype._apply = function (input, options, nonce) {
        var result = null;
        var output = input;
        var issues = null;
        var issueGroups = null;
        var index = 0;
        var shapes = this._lookup(input);
        var shapesLength = shapes.length;
        while (index < shapesLength) {
            result = shapes[index]['_apply'](input, options, nonce);
            if (result === null) {
                break;
            }
            if (!isArray(result)) {
                output = result.value;
                break;
            }
            if (issueGroups === null) {
                issueGroups = [result];
            }
            else {
                issueGroups.push(result);
            }
            issues = result;
            index++;
        }
        if (index === shapesLength) {
            if (shapesLength === 1) {
                return issues;
            }
            return [this._typeIssueFactory(input, options, { inputs: this.inputs, issueGroups: issueGroups })];
        }
        return this._applyOperations(input, output, options, null);
    };
    UnionShape.prototype._applyAsync = function (input, options, nonce) {
        var _this = this;
        return new Promise(function (resolve) {
            var shapes = _this._lookup(input);
            var shapesLength = shapes.length;
            var issues = null;
            var issueGroups = null;
            var index = -1;
            var handleResult = function (result) {
                var output = input;
                if (result !== null) {
                    if (isArray(result)) {
                        if (issueGroups === null) {
                            issueGroups = [result];
                        }
                        else {
                            issueGroups.push(result);
                        }
                        issues = result;
                        return next();
                    }
                    else {
                        output = result.value;
                    }
                }
                return _this._applyOperations(input, output, options, null);
            };
            var next = function () {
                index++;
                if (index !== shapesLength) {
                    return applyShape(shapes[index], input, options, nonce, handleResult);
                }
                if (shapesLength === 1) {
                    return issues;
                }
                return [_this._typeIssueFactory(input, options, { inputs: _this.inputs, issueGroups: issueGroups })];
            };
            resolve(next());
        });
    };
    return UnionShape;
}(Shape));
/**
 * Creates a lookup that finds a shape using an input value type.
 */
function createLookupByType(shapes) {
    var emptyArray = [];
    var buckets = {
        object: emptyArray,
        array: emptyArray,
        function: emptyArray,
        string: emptyArray,
        symbol: emptyArray,
        number: emptyArray,
        bigint: emptyArray,
        boolean: emptyArray,
        date: emptyArray,
        promise: emptyArray,
        set: emptyArray,
        map: emptyArray,
        null: emptyArray,
        undefined: emptyArray,
    };
    var bucketNames = Object.keys(buckets);
    for (var _i = 0, shapes_1 = shapes; _i < shapes_1.length; _i++) {
        var shape = shapes_1[_i];
        var names = shape.inputs[0] === TYPE_UNKNOWN ? bucketNames : shape.inputs.map(function (input) { return getTypeOf(input).name; }).filter(unique);
        for (var _a = 0, names_1 = names; _a < names_1.length; _a++) {
            var name_1 = names_1[_a];
            buckets[name_1] = buckets[name_1].concat(shape);
        }
    }
    return function (input) { return buckets[getTypeOf(input).name]; };
}
/**
 * Creates a lookup that uses a discriminator property, or returns `null` if discriminator property cannot be detected.
 */
function createLookupByDiscriminator(shapes) {
    var discriminator = shapes.every(isObjectShape) ? getDiscriminator(shapes) : null;
    if (discriminator === null) {
        return null;
    }
    var key = discriminator.key, valueGroups = discriminator.valueGroups;
    var shapesLength = shapes.length;
    var shapeGroups = [];
    var emptyArray = [];
    var monoValues = [];
    for (var i = 0; i < shapesLength; ++i) {
        shapeGroups.push([shapes[i]]);
        var valueGroup = valueGroups[i];
        if (monoValues === null || valueGroup.length !== 1 || valueGroup[0] !== valueGroup[0]) {
            monoValues = null;
        }
        else {
            monoValues.push(valueGroup[0]);
        }
    }
    if (monoValues !== null) {
        return function (input) {
            var index;
            if (!isObject(input) || (index = monoValues.indexOf(input[key])) === -1) {
                return emptyArray;
            }
            return shapeGroups[index];
        };
    }
    return function (input) {
        if (!isObject(input)) {
            return emptyArray;
        }
        var value = input[key];
        for (var i = 0; i < shapesLength; ++i) {
            if (valueGroups[i].includes(value)) {
                return shapeGroups[i];
            }
        }
        return emptyArray;
    };
}
/**
 * Returns a discriminator property description. Discriminator conforms the following rules:
 *
 * - has a key that is common for all provided object shapes;
 * - its values are discrete;
 * - its values uniquely identify each shape.
 */
function getDiscriminator(shapes) {
    if (shapes.length < 2) {
        // Discriminator may exist only among multiple objects
        return null;
    }
    var keys = shapes[0].keys;
    var valueGroups = [];
    var valueSet = new Set();
    nextKey: for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
        var key = keys_1[_i];
        for (var i = 0; i < shapes.length; ++i) {
            var shape = shapes[i];
            if (!shape.keys.includes(key)) {
                // Key doesn't exist on every shape
                continue nextKey;
            }
            var inputs = shape.propShapes[key].inputs;
            if (inputs.length === 0 || inputs.some(isType)) {
                // Values aren't discrete
                continue nextKey;
            }
            valueGroups[i] = inputs;
        }
        valueSet.clear();
        for (var i = 0, valueCount = 0; i < valueGroups.length; ++i) {
            for (var _a = 0, _b = valueGroups[i]; _a < _b.length; _a++) {
                var value = _b[_a];
                if (++valueCount !== valueSet.add(value).size) {
                    // Values don't uniquely identify each shape
                    continue nextKey;
                }
            }
        }
        return { key: key, valueGroups: valueGroups };
    }
    return null;
}
function isObjectShape(shape) {
    return shape instanceof ObjectShape;
}

function any(cb, options) {
    var shape = new Shape();
    return cb === null || cb === undefined ? shape : shape.refine(cb, options);
}

function array(shape, options) {
    if (shape instanceof Shape) {
        return new ArrayShape([], shape, options);
    }
    else {
        return new ArrayShape([], new Shape(), shape);
    }
}

/**
 * Creates the bigint shape.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
function bigint(options) {
    return new BigIntShape(options);
}

/**
 * Creates the boolean shape.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
function boolean(options) {
    return new BooleanShape(options);
}

/**
 * Creates the constant value shape.
 *
 * @param value The value to which the input must be strictly equal.
 * @param options The issue options or the issue message.
 * @template Value The expected value.
 * @group DSL
 */
function const_(value, options) {
    return new ConstShape(value, options);
}

/**
 * Creates the {@link !Date Date} shape.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
function date(options) {
    return new DateShape(options);
}

function enum_(source, options) {
    return new EnumShape(source, options);
}

function function_(argShapes, options) {
    if (isArray(argShapes)) {
        argShapes = new ArrayShape(argShapes, null);
    }
    if (!(argShapes instanceof Shape)) {
        options = argShapes;
        argShapes = new ArrayShape([], null);
    }
    return new FunctionShape(argShapes, null, null, options);
}

/**
 * Creates the class instance shape.
 *
 * @param ctor The instance constructor.
 * @param options The issue options or the issue message.
 * @template Ctor The instance constructor.
 * @group DSL
 */
function instanceOf(ctor, options) {
    return new InstanceShape(ctor, options);
}

/**
 * Creates an intersection shape that tries to parse the input with all provided shapes and merge parsing results.
 *
 * @param shapes The array of shapes.
 * @param options The issue options or the issue message.
 * @template Shapes The tuple of intersected shapes.
 * @group DSL
 */
function intersection(shapes, options) {
    return new IntersectionShape(shapes, options);
}

/**
 * Creates the shape that resolves the underlying shape on-demand.
 *
 * @param shapeProvider The provider that returns the resolved shape.
 * @template ProvidedShape The provided shape.
 * @group DSL
 */
function lazy(shapeProvider) {
    return new LazyShape(shapeProvider, identity);
}

/**
 * Creates the {@link !Map Map} instance shape.
 *
 * @param keyShape The key shape.
 * @param valueShape The value shape.
 * @param options The issue options or the issue message.
 * @template KeyShape The key shape.
 * @template ValueShape The value shape.
 * @group DSL
 */
function map(keyShape, valueShape, options) {
    return new MapShape(keyShape, valueShape, options);
}

/**
 * Creates the shape that requires an input to be equal to `NaN`.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
function nan(options) {
    return new ConstShape(NaN, options);
}

/**
 * Creates the shape that always raises an issue.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
function never(options) {
    return new NeverShape(options);
}

/**
 * Creates the shape that only allows values that don't conform the shape.
 *
 * @param shape The shape to which the output must not conform.
 * @param options The issue options or the issue message.
 * @template ExcludedShape The shape to which the output must not conform.
 * @group DSL
 */
function not(shape, options) {
    return new Shape().not(shape, options);
}

/**
 * Creates the shape that requires an input to be equal to `null`.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
function null_(options) {
    return new ConstShape(null, options);
}

/**
 * Creates the number shape.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
function number(options) {
    return new NumberShape(options);
}

/**
 * Creates the object shape.
 *
 * @param shapes The mapping from an object key to a corresponding shape.
 * @param options The issue options or the issue message.
 * @template PropShapes The mapping from a string object key to a corresponding value shape.
 * @group DSL
 */
function object(shapes, options) {
    return new ObjectShape(shapes, null, options);
}

function promise(shape, options) {
    if (shape instanceof Shape) {
        return new PromiseShape(shape, options);
    }
    else {
        return new PromiseShape(null, shape);
    }
}

function record(keyShape, valueShape, options) {
    if (valueShape instanceof Shape) {
        return new RecordShape(keyShape, valueShape, options);
    }
    else {
        return new RecordShape(null, keyShape, options);
    }
}

/**
 * Creates the {@link !Set Set} instance shape.
 *
 * @param shape The value shape
 * @param options The issue options or the issue message.
 * @template ValueShape The value shape.
 * @group DSL
 */
function set(shape, options) {
    return new SetShape(shape, options);
}

/**
 * Creates the string shape.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
function string(options) {
    return new StringShape(options);
}

/**
 * Creates the symbol shape.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
function symbol(options) {
    return new SymbolShape(options);
}

/**
 * Creates the shape that synchronously converts the input value.
 *
 * @param cb The callback that converts the input value. Throw a {@link ValidationError} to notify that the conversion
 * cannot be successfully completed.
 * @template Value The input value.
 * @template ConvertedValue The output value.
 * @group DSL
 */
function convert(
/**
 * @param value The input value.
 * @param options Parsing options.
 */
cb) {
    return new ConvertShape(cb);
}
/**
 * Creates the shape that asynchronously converts the input value.
 *
 * @param cb The callback that converts the input value asynchronously. The returned promise can be rejected with a
 * {@link ValidationError} to notify that the conversion cannot be successfully completed.
 * @template Value The input value.
 * @template ConvertedValue The output value.
 * @group DSL
 */
function convertAsync(
/**
 * @param value The input value.
 * @param options Parsing options.
 */
cb) {
    return new ConvertShape(cb, true);
}

function tuple(shapes, restShape, options) {
    if (restShape === null || restShape === undefined || restShape instanceof Shape) {
        return new ArrayShape(shapes, restShape || null, options);
    }
    else {
        return new ArrayShape(shapes, null, restShape);
    }
}

/**
 * Creates the shape that requires an input to be equal to `undefined`.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
function undefined_(options) {
    return new ConstShape(undefined, options);
}

/**
 * Creates a union shape that tries to parse the input with one of the provided shapes.
 *
 * @param shapes The array of shapes to try.
 * @param options The issue options or the issue message.
 * @template Shapes The tuple of united shapes.
 * @group DSL
 */
function union(shapes, options) {
    return new UnionShape(shapes, options);
}

/**
 * Creates the unconstrained shape with unknown value.
 *
 * @group DSL
 */
function unknown() {
    return new Shape();
}

/**
 * Creates a shape that requires an input to be `undefined` at runtime and typed as `void`.
 *
 * @param options The issue options or the issue message.
 * @group DSL
 */
function void_(options) {
    return new ConstShape(undefined, options);
}

export { ArrayShape, BigIntShape, BooleanShape, CatchShape, CoercibleShape, ConstShape, ConvertShape, DateShape, DenyShape, EnumShape, ExcludeShape, FunctionShape, InstanceShape, IntersectionShape, LazyShape, MapShape, NEVER, NeverShape, NumberShape, ObjectShape, PipeShape, PromiseShape, RecordShape, ReplaceShape, SetShape, Shape, StringShape, SymbolShape, Type, UnionShape, ValidationError, intersection as and, any, array, bigint, boolean as bool, boolean, const_ as const, convert, convertAsync, date, enum_ as enum, function_ as fn, function_ as function, instanceOf, intersection, lazy, map, nan, never, not, null_ as null, number, object, union as or, promise, record, set, string, symbol, tuple, undefined_ as undefined, union, unknown, void_ as void };
