import { AnyShape, IssueOptions, Message, ArrayShape } from '../core';

/**
 * The plugin that enhances {@link core!ArrayShape ArrayShape} with additional methods.
 *
 * ```ts
 * import { ArrayShape } from 'doubter/core';
 * import enableArrayEssentials from 'doubter/plugin/array-essentials';
 *
 * enableArrayEssentials(ArrayShape.prototype);
 * ```
 *
 * @module plugin/array-essentials
 */

declare module '../core' {
    interface ArrayShape<HeadShapes extends readonly AnyShape[], RestShape extends AnyShape | null> {
        /**
         * Constrains the array length.
         *
         * @param length The minimum array length.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/array-essentials! plugin/array-essentials}
         */
        length(length: number, options?: IssueOptions | Message): this;
        /**
         * Constrains the minimum array length.
         *
         * @param length The minimum array length.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/array-essentials! plugin/array-essentials}
         */
        min(length: number, options?: IssueOptions | Message): this;
        /**
         * Constrains the maximum array length.
         *
         * @param length The maximum array length.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/array-essentials! plugin/array-essentials}
         */
        max(length: number, options?: IssueOptions | Message): this;
        /**
         * Constrains the array length to be at least one element.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/array-essentials! plugin/array-essentials}
         */
        nonEmpty(options?: IssueOptions | Message): this;
        /**
         * Requires an array to contain at least one element that conforms the given shape.
         *
         * @param value The shape of the required element or its literal value. If a shape is provided, then it _must_
         * support {@link Shape#isAsync the synchronous parsing}.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/array-essentials! plugin/array-essentials}
         */
        includes(value: AnyShape | any, options?: IssueOptions | Message): this;
    }
}
/**
 * Enhances {@link core!ArrayShape ArrayShape} with additional methods.
 */
declare function enableArrayEssentials(prototype: ArrayShape<any, any>): void;

export { enableArrayEssentials as default };
