import { Shape } from '../core.mjs';
import { createIssueFactory } from '../utils.mjs';

var ERR_SYNC_REQUIRED = 'Shape cannot be used because it is async.';
var CODE_ARRAY_INCLUDES = 'array.includes';
var CODE_ARRAY_MIN = 'array.min';
var CODE_ARRAY_MAX = 'array.max';
var MESSAGE_ARRAY_INCLUDES = 'Must include a value';
var MESSAGE_ARRAY_MAX = 'Must have the maximum length of %s';
var MESSAGE_ARRAY_MIN = 'Must have the minimum length of %s';

/**
 * The plugin that enhances {@link core!ArrayShape ArrayShape} with additional methods.
 *
 * ```ts
 * import { ArrayShape } from 'doubter/core';
 * import enableArrayEssentials from 'doubter/plugin/array-essentials';
 *
 * enableArrayEssentials(ArrayShape.prototype);
 * ```
 *
 * @module plugin/array-essentials
 */
/**
 * Enhances {@link core!ArrayShape ArrayShape} with additional methods.
 */
function enableArrayEssentials(prototype) {
    prototype.length = function (length, options) {
        return this.min(length, options).max(length, options);
    };
    prototype.min = function (length, options) {
        var issueFactory = createIssueFactory(CODE_ARRAY_MIN, MESSAGE_ARRAY_MIN, options, length);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output.length < length) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_ARRAY_MIN, param: length });
    };
    prototype.max = function (length, options) {
        var issueFactory = createIssueFactory(CODE_ARRAY_MAX, MESSAGE_ARRAY_MAX, options, length);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output.length > length) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_ARRAY_MAX, param: length });
    };
    prototype.nonEmpty = function (options) {
        return this.min(1, options);
    };
    prototype.includes = function (value, options) {
        var issueFactory = createIssueFactory(CODE_ARRAY_INCLUDES, MESSAGE_ARRAY_INCLUDES, options, value);
        var lookup;
        if (value instanceof Shape) {
            if (value.isAsync) {
                throw new Error(ERR_SYNC_REQUIRED);
            }
            lookup = function (output, options) {
                for (var _i = 0, output_1 = output; _i < output_1.length; _i++) {
                    var outputValue = output_1[_i];
                    if (value.try(outputValue, options).ok) {
                        return true;
                    }
                }
                return false;
            };
        }
        else {
            lookup = function (output, options) { return output.includes(value); };
        }
        return this.use(function (next) { return function (input, output, options, issues) {
            if (!lookup(output, options)) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_ARRAY_INCLUDES, param: value });
    };
}

export { enableArrayEssentials as default };
