import { IssueOptions, Message, BigIntShape } from '../core';

/**
 * The plugin that enhances {@link core!BigIntShape BigIntShape} with additional methods.
 *
 * ```ts
 * import { BigIntShape } from 'doubter/core';
 * import enableBigIntEssentials from 'doubter/plugin/bigint-essentials';
 *
 * enableBigIntEssentials(BigIntShape.prototype);
 * ```
 *
 * @module plugin/bigint-essentials
 */

declare module '../core' {
    interface BigIntShape {
        /**
         * Constrains the bigint to be greater than zero.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/bigint-essentials! plugin/bigint-essentials}
         */
        positive(options?: IssueOptions | Message): this;
        /**
         * Constrains the bigint to be less than zero.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/bigint-essentials! plugin/bigint-essentials}
         */
        negative(options?: IssueOptions | Message): this;
        /**
         * Constrains the bigint to be less or equal to zero.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/bigint-essentials! plugin/bigint-essentials}
         */
        nonPositive(options?: IssueOptions | Message): this;
        /**
         * Constrains the bigint to be greater or equal to zero.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/bigint-essentials! plugin/bigint-essentials}
         */
        nonNegative(options?: IssueOptions | Message): this;
        /**
         * Constrains the bigint to be greater than or equal to the value.
         *
         * @param value The inclusive minimum value.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/bigint-essentials! plugin/bigint-essentials}
         */
        min(value: bigint | number, options?: IssueOptions | Message): this;
        /**
         * Constrains the number to be less than or equal to the value.
         *
         * @param value The inclusive maximum value.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/bigint-essentials! plugin/bigint-essentials}
         */
        max(value: bigint | number, options?: IssueOptions | Message): this;
    }
}
/**
 * Enhances {@link core!BigIntShape BigIntShape} with additional methods.
 */
declare function enableBigIntEssentials(prototype: BigIntShape): void;

export { enableBigIntEssentials as default };
