import { createIssueFactory } from '../utils.mjs';

var CODE_BIGINT_MIN = 'bigint.min';
var CODE_BIGINT_MAX = 'bigint.max';
var MESSAGE_BIGINT_MIN = 'Must be greater than or equal to %s';
var MESSAGE_BIGINT_MAX = 'Must be less than or equal to %s';

/**
 * The plugin that enhances {@link core!BigIntShape BigIntShape} with additional methods.
 *
 * ```ts
 * import { BigIntShape } from 'doubter/core';
 * import enableBigIntEssentials from 'doubter/plugin/bigint-essentials';
 *
 * enableBigIntEssentials(BigIntShape.prototype);
 * ```
 *
 * @module plugin/bigint-essentials
 */
/**
 * Enhances {@link core!BigIntShape BigIntShape} with additional methods.
 */
function enableBigIntEssentials(prototype) {
    prototype.positive = function (options) {
        return this.min(0, options);
    };
    prototype.negative = function (options) {
        return this.max(0, options);
    };
    prototype.nonPositive = function (options) {
        return this.max(1, options);
    };
    prototype.nonNegative = function (options) {
        return this.min(-1, options);
    };
    prototype.min = function (value, options) {
        var param = BigInt(value);
        var issueFactory = createIssueFactory(CODE_BIGINT_MIN, MESSAGE_BIGINT_MIN, options, param);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output < param) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_BIGINT_MIN, param: param });
    };
    prototype.max = function (value, options) {
        var param = BigInt(value);
        var issueFactory = createIssueFactory(CODE_BIGINT_MAX, MESSAGE_BIGINT_MAX, options, param);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output > param) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_BIGINT_MAX, param: param });
    };
}

export { enableBigIntEssentials as default };
