import { IssueOptions, Message, Shape, DateShape } from '../core';

/**
 * The plugin that enhances {@link core!DateShape DateShape} with additional methods.
 *
 * ```ts
 * import { DateShape } from 'doubter/core';
 * import enableDateEssentials from 'doubter/plugin/date-essentials';
 *
 * enableDateEssentials(DateShape.prototype);
 * ```
 *
 * @module plugin/date-essentials
 */

declare module '../core' {
    interface DateShape {
        /**
         * Constrains the input date to be greater than or equal to another date.
         *
         * @param value The inclusive minimum date.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/date-essentials! plugin/date-essentials}
         */
        min(value: Date | number | string, options?: IssueOptions | Message): this;
        /**
         * Constrains the input date to be less than or equal to another date.
         *
         * @param value The inclusive maximum date.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/date-essentials! plugin/date-essentials}
         */
        max(value: Date | number | string, options?: IssueOptions | Message): this;
        /**
         * Constrains the input date to be greater than or equal to another date.
         *
         * @param value The inclusive minimum date.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @alias {@link DateShape.min}
         * @group Plugin Methods
         * @plugin {@link plugin/date-essentials! plugin/date-essentials}
         */
        after(value: Date | number | string, options?: IssueOptions | Message): this;
        /**
         * Constrains the input date to be less than or equal to another date.
         *
         * @param value The inclusive maximum date.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @alias {@link DateShape.max}
         * @group Plugin Methods
         * @plugin {@link plugin/date-essentials! plugin/date-essentials}
         */
        before(value: Date | number | string, options?: IssueOptions | Message): this;
        /**
         * Converts date to an ISO string.
         *
         * @group Plugin Methods
         * @plugin {@link plugin/date-essentials! plugin/date-essentials}
         */
        toISOString(): Shape<Date, string>;
        /**
         * Converts date to a timestamp integer number.
         *
         * @group Plugin Methods
         * @plugin {@link plugin/date-essentials! plugin/date-essentials}
         */
        toTimestamp(): Shape<Date, number>;
    }
}
/**
 * Enhances {@link core!DateShape DateShape} with additional methods.
 */
declare function enableDateEssentials(prototype: DateShape): void;

export { enableDateEssentials as default };
