import { createIssueFactory } from '../utils.mjs';

var CODE_DATE_MIN = 'date.min';
var CODE_DATE_MAX = 'date.max';
var MESSAGE_DATE_MIN = 'Must be after %s';
var MESSAGE_DATE_MAX = 'Must be before %s';

/**
 * The plugin that enhances {@link core!DateShape DateShape} with additional methods.
 *
 * ```ts
 * import { DateShape } from 'doubter/core';
 * import enableDateEssentials from 'doubter/plugin/date-essentials';
 *
 * enableDateEssentials(DateShape.prototype);
 * ```
 *
 * @module plugin/date-essentials
 */
/**
 * Enhances {@link core!DateShape DateShape} with additional methods.
 */
function enableDateEssentials(prototype) {
    prototype.min = function (value, options) {
        var param = new Date(value);
        var timestamp = param.getTime();
        var issueFactory = createIssueFactory(CODE_DATE_MIN, MESSAGE_DATE_MIN, options, param);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output.getTime() < timestamp) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_DATE_MIN, param: param });
    };
    prototype.max = function (value, options) {
        var param = new Date(value);
        var timestamp = param.getTime();
        var issueFactory = createIssueFactory(CODE_DATE_MAX, MESSAGE_DATE_MAX, options, param);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output.getTime() > timestamp) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_DATE_MAX, param: param });
    };
    prototype.after = prototype.min;
    prototype.before = prototype.max;
    prototype.toISOString = function () {
        return this.convert(toISOString);
    };
    prototype.toTimestamp = function () {
        return this.convert(toTimestamp);
    };
}
function toISOString(date) {
    return date.toISOString();
}
function toTimestamp(date) {
    return date.getTime();
}

export { enableDateEssentials as default };
