import { IssueOptions, Message, NumberShape } from '../core';

/**
 * The plugin that enhances {@link core!NumberShape NumberShape} with additional methods.
 *
 * ```ts
 * import { NumberShape } from 'doubter/core';
 * import enableNumberEssentials from 'doubter/plugin/number-essentials';
 *
 * enableNumberEssentials(NumberShape.prototype);
 * ```
 *
 * @module plugin/number-essentials
 */

interface MultipleOfOptions extends IssueOptions {
    /**
     * By default, {@link core!NumberShape#multipleOf NumberShape.multipleOf} uses
     * [the modulo operator](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Remainder) which
     * may produce unexpected results when used with floating point numbers. This happens because of
     * [the way numbers are represented by IEEE 754](https://docs.oracle.com/cd/E19957-01/806-3568/ncg_goldberg.html).
     *
     * For example, for small dividers such as 0.01 the result of the division is usually not integer (even when it should
     * be integer). If you need to use fractional dividers set this option to some positive integer to have `multipleOf`
     * validated using this formula:
     *
     * ```
     * Math.abs(Math.round(value / divisor) - value / divisor) < Math.pow(10, -precision)
     * ```
     */
    precision?: number;
}
declare module '../core' {
    interface NumberShape {
        /**
         * Constrains the number to be a finite number.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/number-essentials! plugin/number-essentials}
         */
        finite(options?: IssueOptions | Message): this;
        /**
         * Constrains the number to be an integer.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/number-essentials! plugin/number-essentials}
         */
        int(options?: IssueOptions | Message): this;
        /**
         * Constrains the number to be greater than zero.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/number-essentials! plugin/number-essentials}
         */
        positive(options?: IssueOptions | Message): this;
        /**
         * Constrains the number to be less than zero.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/number-essentials! plugin/number-essentials}
         */
        negative(options?: IssueOptions | Message): this;
        /**
         * Constrains the number to be less or equal to zero.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/number-essentials! plugin/number-essentials}
         */
        nonPositive(options?: IssueOptions | Message): this;
        /**
         * Constrains the number to be greater or equal to zero.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/number-essentials! plugin/number-essentials}
         */
        nonNegative(options?: IssueOptions | Message): this;
        /**
         * Constrains the number to be greater than the value.
         *
         * @param value The exclusive minimum value.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/number-essentials! plugin/number-essentials}
         */
        gt(value: number, options?: IssueOptions | Message): this;
        /**
         * Constrains the number to be less than the value.
         *
         * @param value The exclusive maximum value.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/number-essentials! plugin/number-essentials}
         */
        lt(value: number, options?: IssueOptions | Message): this;
        /**
         * Constrains the number to be greater than or equal to the value.
         *
         * @param value The inclusive minimum value.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/number-essentials! plugin/number-essentials}
         */
        gte(value: number, options?: IssueOptions | Message): this;
        /**
         * Constrains the number to be less than or equal to the value.
         *
         * @param value The inclusive maximum value.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/number-essentials! plugin/number-essentials}
         */
        lte(value: number, options?: IssueOptions | Message): this;
        /**
         * Constrains the number to be greater than or equal to the value.
         *
         * @param value The inclusive minimum value.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @alias {@link NumberShape.gte}
         * @group Plugin Methods
         * @plugin {@link plugin/number-essentials! plugin/number-essentials}
         */
        min(value: number, options?: IssueOptions | Message): this;
        /**
         * Constrains the number to be less than or equal to the value.
         *
         * @param value The inclusive maximum value.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @alias {@link NumberShape.lte}
         * @group Plugin Methods
         * @plugin {@link plugin/number-essentials! plugin/number-essentials}
         */
        max(value: number, options?: IssueOptions | Message): this;
        /**
         * Constrains the number to be a multiple of the divisor.
         *
         * @param divisor The positive number by which the input should be divisible without a remainder.
         * @param options The check options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/number-essentials! plugin/number-essentials}
         */
        multipleOf(divisor: number, options?: MultipleOfOptions | Message): this;
        /**
         * Number must be between `Number.MIN_SAFE_INTEGER` and `Number.MAX_SAFE_INTEGER`.
         *
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/number-essentials! plugin/number-essentials}
         */
        safe(options?: IssueOptions | Message): this;
    }
}
/**
 * Enhances {@link core!NumberShape NumberShape} with additional methods.
 */
declare function enableNumberEssentials(prototype: NumberShape): void;

export { type MultipleOfOptions, enableNumberEssentials as default };
