'use strict';

var utils = require('../utils.js');

var CODE_NUMBER_FINITE = 'number.finite';
var CODE_NUMBER_INT = 'number.int';
var CODE_NUMBER_GT = 'number.gt';
var CODE_NUMBER_GTE = 'number.gte';
var CODE_NUMBER_LT = 'number.lt';
var CODE_NUMBER_LTE = 'number.lte';
var CODE_NUMBER_MULTIPLE_OF = 'number.multipleOf';
var MESSAGE_NUMBER_FINITE = 'Must be a finite number';
var MESSAGE_NUMBER_INT = 'Must be an integer';
var MESSAGE_NUMBER_GT = 'Must be greater than %s';
var MESSAGE_NUMBER_GTE = 'Must be greater than or equal to %s';
var MESSAGE_NUMBER_LT = 'Must be less than %s';
var MESSAGE_NUMBER_LTE = 'Must be less than or equal to %s';
var MESSAGE_NUMBER_MULTIPLE_OF = 'Must be a multiple of %s';

/**
 * The plugin that enhances {@link core!NumberShape NumberShape} with additional methods.
 *
 * ```ts
 * import { NumberShape } from 'doubter/core';
 * import enableNumberEssentials from 'doubter/plugin/number-essentials';
 *
 * enableNumberEssentials(NumberShape.prototype);
 * ```
 *
 * @module plugin/number-essentials
 */
/**
 * Enhances {@link core!NumberShape NumberShape} with additional methods.
 */
function enableNumberEssentials(prototype) {
    prototype.finite = function (options) {
        var isFinite = Number.isFinite;
        var issueFactory = utils.createIssueFactory(CODE_NUMBER_FINITE, MESSAGE_NUMBER_FINITE, options, undefined);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (!isFinite(output)) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_NUMBER_FINITE });
    };
    prototype.int = function (options) {
        var isInteger = Number.isInteger;
        var issueFactory = utils.createIssueFactory(CODE_NUMBER_INT, MESSAGE_NUMBER_INT, options, undefined);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (!isInteger(output)) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_NUMBER_INT });
    };
    prototype.positive = function (options) {
        return this.gt(0, options);
    };
    prototype.negative = function (options) {
        return this.lt(0, options);
    };
    prototype.nonPositive = function (options) {
        return this.lte(0, options);
    };
    prototype.nonNegative = function (options) {
        return this.gte(0, options);
    };
    prototype.gt = function (value, options) {
        var issueFactory = utils.createIssueFactory(CODE_NUMBER_GT, MESSAGE_NUMBER_GT, options, value);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output <= value) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_NUMBER_GT, param: value });
    };
    prototype.lt = function (value, options) {
        var issueFactory = utils.createIssueFactory(CODE_NUMBER_LT, MESSAGE_NUMBER_LT, options, value);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output >= value) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_NUMBER_LT, param: value });
    };
    prototype.gte = function (value, options) {
        var issueFactory = utils.createIssueFactory(CODE_NUMBER_GTE, MESSAGE_NUMBER_GTE, options, value);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output < value) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_NUMBER_GTE, param: value });
    };
    prototype.lte = function (value, options) {
        var issueFactory = utils.createIssueFactory(CODE_NUMBER_LTE, MESSAGE_NUMBER_LTE, options, value);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output > value) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_NUMBER_LTE, param: value });
    };
    prototype.min = prototype.gte;
    prototype.max = prototype.lte;
    prototype.multipleOf = function (divisor, options) {
        var abs = Math.abs, round = Math.round;
        var precision = utils.extractOptions(options).precision;
        var epsilon = precision !== undefined ? Math.pow(10, -precision) : -1;
        var issueFactory = utils.createIssueFactory(CODE_NUMBER_MULTIPLE_OF, MESSAGE_NUMBER_MULTIPLE_OF, options, divisor);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (epsilon !== -1 ? abs(round(output / divisor) - output / divisor) > epsilon : output % divisor !== 0) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_NUMBER_MULTIPLE_OF, param: divisor });
    };
    prototype.safe = function (options) {
        return this.min(Number.MIN_SAFE_INTEGER, options).max(Number.MAX_SAFE_INTEGER, options);
    };
}

module.exports = enableNumberEssentials;
