import { createIssueFactory } from '../utils.mjs';

var CODE_OBJECT_ALL_KEYS = 'object.allKeys';
var CODE_OBJECT_NOT_ALL_KEYS = 'object.notAllKeys';
var CODE_OBJECT_OR_KEYS = 'object.orKeys';
var CODE_OBJECT_XOR_KEYS = 'object.xorKeys';
var CODE_OBJECT_OXOR_KEYS = 'object.oxorKeys';
var CODE_OBJECT_PLAIN = 'object.plain';
var MESSAGE_OBJECT_ALL_KEYS = 'Must contain all or no keys: %s';
var MESSAGE_OBJECT_NOT_ALL_KEYS = 'Must contain not all or no keys: %s';
var MESSAGE_OBJECT_OR_KEYS = 'Must contain at least one key: %s';
var MESSAGE_OBJECT_XOR_KEYS = 'Must contain exactly one key: %s';
var MESSAGE_OBJECT_OXOR_KEYS = 'Must contain one or no keys: %s';
var MESSAGE_OBJECT_PLAIN = 'Must be a plain object';

/**
 * The plugin that enhances {@link core!ObjectShape ObjectShape} with additional methods.
 *
 * ```ts
 * import { ObjectShape } from 'doubter/core';
 * import enableObjectEssentials from 'doubter/plugin/object-essentials';
 *
 * enableObjectEssentials(ObjectShape.prototype);
 * ```
 *
 * @module plugin/object-essentials
 */
/**
 * Enhances {@link core!ObjectShape ObjectShape} with additional methods.
 */
function enableObjectEssentials(prototype) {
    prototype.plain = function (options) {
        var getPrototypeOf = Object.getPrototypeOf;
        var issueFactory = createIssueFactory(CODE_OBJECT_PLAIN, MESSAGE_OBJECT_PLAIN, options, undefined);
        return this.use(function (next) { return function (input, output, options, issues) {
            var prototype = getPrototypeOf(output);
            if (prototype !== null && prototype.constructor !== Object) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_OBJECT_PLAIN });
    };
    prototype.allKeys = function (keys, options) {
        var issueFactory = createIssueFactory(CODE_OBJECT_ALL_KEYS, MESSAGE_OBJECT_ALL_KEYS, options, keys);
        return this.use(function (next) { return function (input, output, options, issues) {
            var keyCount = getKeyCount(output, keys, keys.length);
            if (keyCount > 0 && keyCount < keys.length) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_OBJECT_ALL_KEYS, param: keys });
    };
    prototype.notAllKeys = function (keys, options) {
        var issueFactory = createIssueFactory(CODE_OBJECT_NOT_ALL_KEYS, MESSAGE_OBJECT_NOT_ALL_KEYS, options, keys);
        return this.use(function (next) { return function (input, output, options, issues) {
            var keyCount = getKeyCount(output, keys, keys.length);
            if (keyCount > 0 && keyCount <= keys.length) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_OBJECT_NOT_ALL_KEYS, param: keys });
    };
    prototype.orKeys = function (keys, options) {
        var issueFactory = createIssueFactory(CODE_OBJECT_OR_KEYS, MESSAGE_OBJECT_OR_KEYS, options, keys);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (getKeyCount(output, keys, 1) === 0) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_OBJECT_OR_KEYS, param: keys });
    };
    prototype.xorKeys = function (keys, options) {
        var issueFactory = createIssueFactory(CODE_OBJECT_XOR_KEYS, MESSAGE_OBJECT_XOR_KEYS, options, keys);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (getKeyCount(output, keys, 2) !== 1) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_OBJECT_XOR_KEYS, param: keys });
    };
    prototype.oxorKeys = function (keys, options) {
        var issueFactory = createIssueFactory(CODE_OBJECT_OXOR_KEYS, MESSAGE_OBJECT_OXOR_KEYS, options, keys);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (getKeyCount(output, keys, 2) > 1) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_OBJECT_OXOR_KEYS, param: keys });
    };
}
function getKeyCount(output, keys, maxCount) {
    var count = 0;
    for (var i = 0; i < keys.length && count < maxCount; ++i) {
        if (output[keys[i]] !== undefined) {
            ++count;
        }
    }
    return count;
}

export { enableObjectEssentials as default };
