'use strict';

var utils = require('../utils.js');

var CODE_SET_MIN = 'set.min';
var CODE_SET_MAX = 'set.max';
var MESSAGE_SET_MIN = 'Must have the minimum size of %s';
var MESSAGE_SET_MAX = 'Must have the maximum size of %s';

/**
 * The plugin that enhances {@link core!SetShape SetShape} with additional methods.
 *
 * ```ts
 * import { SetShape } from 'doubter/core';
 * import enableSetEssentials from 'doubter/plugin/set-essentials';
 *
 * enableSetEssentials(SetShape.prototype);
 * ```
 *
 * @module plugin/set-essentials
 */
/**
 * Enhances {@link core!SetShape SetShape} with additional methods.
 */
function enableSetEssentials(prototype) {
    prototype.size = function (size, options) {
        return this.min(size, options).max(size, options);
    };
    prototype.min = function (size, options) {
        var issueFactory = utils.createIssueFactory(CODE_SET_MIN, MESSAGE_SET_MIN, options, size);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output.size < size) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_SET_MIN, param: size });
    };
    prototype.max = function (size, options) {
        var issueFactory = utils.createIssueFactory(CODE_SET_MAX, MESSAGE_SET_MAX, options, size);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output.size > size) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_SET_MAX, param: size });
    };
    prototype.nonEmpty = function (options) {
        return this.min(1, options);
    };
}

module.exports = enableSetEssentials;
