import { IssueOptions, Message, StringShape } from '../core';

/**
 * The plugin that enhances {@link core!StringShape StringShape} with additional methods.
 *
 * ```ts
 * import { StringShape } from 'doubter/core';
 * import enableStringEssentials from 'doubter/plugin/string-essentials';
 *
 * enableStringEssentials(StringShape.prototype);
 * ```
 *
 * @module plugin/string-essentials
 */

declare module '../core' {
    interface StringShape {
        /**
         * The shortcut to apply both {@link StringShape.min} and {@link StringShape.max} constraints.
         *
         * @param length The exact length a string must have.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/string-essentials! plugin/string-essentials}
         */
        length(length: number, options?: IssueOptions | Message): this;
        /**
         * Constrains the string length to be greater than or equal to the length.
         *
         * @param length The minimum string length.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/string-essentials! plugin/string-essentials}
         */
        min(length: number, options?: IssueOptions | Message): this;
        /**
         * Constrains the string length to be less than or equal to the length.
         *
         * @param length The maximum string length.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/string-essentials! plugin/string-essentials}
         */
        max(length: number, options?: IssueOptions | Message): this;
        /**
         * Constrains the string to match a regexp.
         *
         * @param re The regular expression that the sting must conform.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/string-essentials! plugin/string-essentials}
         */
        regex(re: RegExp, options?: IssueOptions | Message): this;
        /**
         * Checks that the string includes a substring.
         *
         * @param value The substring to look for.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/string-essentials! plugin/string-essentials}
         */
        includes(value: string, options?: IssueOptions | Message): this;
        /**
         * Checks that the string starts with a substring.
         *
         * @param value The substring to look for.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/string-essentials! plugin/string-essentials}
         */
        startsWith(value: string, options?: IssueOptions | Message): this;
        /**
         * Checks that the string ends with a substring.
         *
         * @param value The substring to look for.
         * @param options The issue options or the issue message.
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/string-essentials! plugin/string-essentials}
         */
        endsWith(value: string, options?: IssueOptions | Message): this;
        /**
         * Checks that the string doesn't consist of whitespace characters only.
         *
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/string-essentials! plugin/string-essentials}
         */
        nonBlank(options?: IssueOptions | Message): this;
        /**
         * Checks that the string has at least one character.
         *
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/string-essentials! plugin/string-essentials}
         */
        nonEmpty(options?: IssueOptions | Message): this;
        /**
         * Trims the output string.
         *
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/string-essentials! plugin/string-essentials}
         */
        trim(): this;
        /**
         * Converts string to lower case.
         *
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/string-essentials! plugin/string-essentials}
         */
        toLowerCase(): this;
        /**
         * Converts string to upper case.
         *
         * @returns The clone of the shape.
         * @group Plugin Methods
         * @plugin {@link plugin/string-essentials! plugin/string-essentials}
         */
        toUpperCase(): this;
    }
}
/**
 * Enhances {@link core!StringShape StringShape} with additional methods.
 */
declare function enableStringEssentials(prototype: StringShape): void;

export { enableStringEssentials as default };
