'use strict';

var utils = require('../utils.js');

var CODE_STRING_NON_BLANK = 'string.nonBlank';
var CODE_STRING_MIN = 'string.min';
var CODE_STRING_MAX = 'string.max';
var CODE_STRING_REGEX = 'string.regex';
var CODE_STRING_INCLUDES = 'string.includes';
var CODE_STRING_STARTS_WITH = 'string.startsWith';
var CODE_STRING_ENDS_WITH = 'string.endsWith';
var MESSAGE_STRING_NON_BLANK = 'Must not be blank';
var MESSAGE_STRING_MIN = 'Must have the minimum length of %s';
var MESSAGE_STRING_MAX = 'Must have the maximum length of %s';
var MESSAGE_STRING_REGEX = 'Must match the pattern %s';
var MESSAGE_STRING_INCLUDES = 'Must include: %s';
var MESSAGE_STRING_STARTS_WITH = 'Must start with: %s';
var MESSAGE_STRING_ENDS_WITH = 'Must end with: %s';

/**
 * The plugin that enhances {@link core!StringShape StringShape} with additional methods.
 *
 * ```ts
 * import { StringShape } from 'doubter/core';
 * import enableStringEssentials from 'doubter/plugin/string-essentials';
 *
 * enableStringEssentials(StringShape.prototype);
 * ```
 *
 * @module plugin/string-essentials
 */
/**
 * Enhances {@link core!StringShape StringShape} with additional methods.
 */
function enableStringEssentials(prototype) {
    prototype.length = function (length, options) {
        return this.min(length, options).max(length, options);
    };
    prototype.min = function (length, options) {
        var issueFactory = utils.createIssueFactory(CODE_STRING_MIN, MESSAGE_STRING_MIN, options, length);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output.length < length) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_STRING_MIN, param: length });
    };
    prototype.max = function (length, options) {
        var issueFactory = utils.createIssueFactory(CODE_STRING_MAX, MESSAGE_STRING_MAX, options, length);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output.length > length) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_STRING_MAX, param: length });
    };
    prototype.regex = function (re, options) {
        var issueFactory = utils.createIssueFactory(CODE_STRING_REGEX, MESSAGE_STRING_REGEX, options, re);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (!re.test(output)) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_STRING_REGEX, param: re });
    };
    prototype.includes = function (value, options) {
        var issueFactory = utils.createIssueFactory(CODE_STRING_INCLUDES, MESSAGE_STRING_INCLUDES, options, value);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output.indexOf(value) === -1) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_STRING_INCLUDES, param: value });
    };
    prototype.startsWith = function (value, options) {
        var issueFactory = utils.createIssueFactory(CODE_STRING_STARTS_WITH, MESSAGE_STRING_STARTS_WITH, options, value);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (!output.startsWith(value)) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_STRING_STARTS_WITH, param: value });
    };
    prototype.endsWith = function (value, options) {
        var issueFactory = utils.createIssueFactory(CODE_STRING_ENDS_WITH, MESSAGE_STRING_ENDS_WITH, options, value);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (!output.endsWith(value)) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_STRING_ENDS_WITH, param: value });
    };
    prototype.nonBlank = function (options) {
        var issueFactory = utils.createIssueFactory(CODE_STRING_NON_BLANK, MESSAGE_STRING_NON_BLANK, options, undefined);
        return this.use(function (next) { return function (input, output, options, issues) {
            if (output.trim().length === 0) {
                (issues || (issues = [])).push(issueFactory(output, options));
                if (options.earlyReturn) {
                    return issues;
                }
            }
            return next(input, output, options, issues);
        }; }, { type: CODE_STRING_NON_BLANK });
    };
    prototype.nonEmpty = function (options) {
        return this.min(1, options);
    };
    prototype.trim = function () {
        return this.use(function (next) { return function (input, output, options, issues) { return next(input, output.trim(), options, issues); }; }, {
            type: 'string.trim',
        });
    };
    prototype.toLowerCase = function () {
        return this.use(function (next) { return function (input, output, options, issues) { return next(input, output.toLowerCase(), options, issues); }; }, {
            type: 'string.toLowerCase',
        });
    };
    prototype.toUpperCase = function () {
        return this.use(function (next) { return function (input, output, options, issues) { return next(input, output.toUpperCase(), options, issues); }; }, {
            type: 'string.toUpperCase',
        });
    };
}

module.exports = enableStringEssentials;
